/*
 * Authenticated User Community
 * Copyright (c) 2000 David C. Moore
 *
 * You may distribute and/or use for any purpose modified or unmodified
 * copies of this software if you preserve the copyright notice above.
 */
 
/**
 * @file auc.c
 * 
 * The main program that governs AUC execution
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <auc/cgic.h>
#include <auc/auc.h>
#include <auc/database.h>
#include <auc/pages.h>
#include <auc/conf.h>
#include <auc/email.h>
#include <auc/files.h>
#include <auc/calendar.h>
#include <auc/course.h>
#include <auc/directory.h>
#include <auc/discussion.h>
#include <auc/coursecal.h>
#include <auc/plugin.h>
#include <auc/error.h>
#include <auc/install.h>
#include <auc/registeracct.h>
#include <auc/html.h>

#include <locale.h>

int upgrade_global(char *oldversion);
void do_upgrade(void);
int init_global(void);

MYSQL *db;

char *aucOrgName = "AUC Installation";
char *aucSiteName = "AUC";
char *aucWebmaster = "webmaster@host.com";
char *aucGraphDir = "/auc/g";
char *aucGraphDirLocal = "/home/httpd/html/auc/g";
char *aucVarRoot = "/var/auc";
char *aucEtcRoot = aucEtcDirDefault;
char *aucDataDir = aucDataDirDefault;
char *aucUserDir = ".auc";
char *aucDBHost = "localhost";
char *aucDBName = "";
char *aucDBUser = "";
char *aucDBPasswd = "";
char *aucModExt = ".so";
char *aucHeaderBGColor = "EEEEF8";
char *aucHeaderTextColor = "660000";
char *aucTrimColor = "990000";
char *aucLinkColor = "CC3333";
char *aucVLinkColor = "996666";
unsigned long aucMaxIdleSecs = 3000;
unsigned long aucMaxUsernameLen = 8;
unsigned long aucMaxGroupLen = 16;
unsigned long aucCurrSemester = 1;

int aucAuthenticated = 0;
int aucAllowLocal = 1;
int aucDBConnectFailed = 0;
time_t aucTime;
int aucNodeColumns = 3;
int aucFileExpert = 0;
aucAuthCachePtr aucUserCache = NULL;
int aucChPasswdStatus = 0;
int aucFirstRun = 0;


char **aucRoles;
char *roles_default[] = {
	"system",
	"staff",
	"student",
	"visitor",
	"parent",
	"guest",
	NULL
};



/**
 * AUC's "main" function.
 *
 * Although main() itself exists in cgic.c, it just acts as a
 * wrapper where CGI queries get parsed.  This function initializes AUC, performs the
 * desired request, and terminates AUC.
 *
 * @param argc Number of arguments.  Ignored since this is a CGI program.
 * @param argv List of arguments.  Ignored.
 *
 * @return Always 0
 */

int
cgiMain(int argc, char **argv)
{
	char *forcelang = NULL;
	char *lang = (char *) getenv("HTTP_ACCEPT_LANGUAGE");

	if (aucConfParse())
		aucFirstRun = 1;

	aucConfToken("child::ForceLang", &forcelang);
	
	if ((forcelang && !strcmp(forcelang, "yes")) || !lang)
		aucConfToken("child::DefaultLang", &lang);
	
	if (lang)
		setenv("LANGUAGE", lang, 0);

	setlocale(LC_ALL, "en_US");
	bindtextdomain(PACKAGE, LOCALEDIR);
	textdomain(PACKAGE);
	bind_textdomain_codeset(PACKAGE, "UTF-8");

	umask(077);
	db = NULL;

	aucTime = time(NULL);
	srand(aucTime);
	aucAuthenticated = 0;

	init_global();
	aucEmailInit();
	aucFilesInit();
	aucCalendarInit();
	aucCourseInit();
	aucDirInit();
	aucDiscInit();
	aucCourseCalInit();
	aucPluginInitAll(aucDataDir);
	do_upgrade();

	aucUserCache = aucAuthNewCache();

	if (!aucPagesRun(cgiOut, cgiPathInfo))
		aucErrorSaveQueue();

	aucConfSave();
	aucPluginCleanupAll(aucDataDir);
	if (aucAuthenticated)
		aucAuthWriteCache(aucUserCache, aucUserCache->uname);
	aucAuthFreeCache(aucUserCache);
	aucDBDisconnect();

//   aucFreeResources();

	return 0;
}

int
init_global(void)
{
	char *buf;

	aucNodelettInit();
	aucPagesInit();

	aucConfToken("child::OrganizationName", &aucOrgName);
	aucConfToken("child::ServerName", &aucSiteName);
	aucConfToken("child::Webmaster", &aucWebmaster);
	aucConfToken("child::Files/GraphicsUrl", &aucGraphDir);
	aucConfToken("child::Files/GraphicsDirectory", &aucGraphDirLocal);
	aucConfToken("child::Files/VarDirectory", &aucVarRoot);
	aucConfToken("child::Files/EtcDirectory", &aucEtcRoot);
	aucConfToken("child::Files/DataDirectory", &aucDataDir);

	aucConfToken("child::Database/SQLServer/attribute::host", &aucDBHost);
	if (!strcmp(aucDBHost, "localhost"))
		aucDBHost = NULL;
	aucConfToken("child::Database/SQLServer/attribute::name", &aucDBName);
	aucConfToken("child::Database/SQLServer/attribute::username", &aucDBUser);
	aucConfToken("child::Database/SQLServer/attribute::password", &aucDBPasswd);

	aucConfToken("child::Colors/HeaderBGColor", &aucHeaderBGColor);
	aucConfToken("child::Colors/HeaderTextColor", &aucHeaderTextColor);
	aucConfToken("child::Colors/TrimColor", &aucTrimColor);
	aucConfToken("child::Colors/LinkColor", &aucLinkColor);
	aucConfToken("child::Colors/VLinkColor", &aucVLinkColor);

	aucConfToken("child::Plugins/attribute::sharedlibextension", &aucModExt);

	if (!aucConfToken("child::Authentication/ExpirationTime", &buf))
		aucMaxIdleSecs = atol(buf);
	if (!aucConfToken("child::Authentication/MaxUsernameLen", &buf))
		aucMaxUsernameLen = atol(buf);
	if (!aucConfToken("child::Authentication/AllowLocal", &buf))
		if (!strcasecmp(buf, "no"))
			aucAllowLocal = 0;

	if (aucConfTokens("child::Roles/Role/attribute::name", &aucRoles))
		aucRoles = roles_default;

	aucNodelettRegister("site-setup",
						_("Administration Tools"),
						aucSiteSetupNodeFrame, 2, 80, 100,
						_("Administrative tools for the AUC site"), "sysop", 0);


	aucPagesRegister(NULL, "logincheck", PAGE_FILE, aucPageLoginCheck, "guest", 0);
	aucPagesRegister(NULL, "logoff", PAGE_FILE, aucPageLogoff, "guest", 0);
	aucPagesRegister(NULL, "about", PAGE_FILE, aucPageAbout, NULL, 0);
	aucPagesRegister(NULL, "home", PAGE_FILE, aucPageStart, "guest", 0);
	aucPagesRegister(NULL, "prefs", PAGE_FILE, aucPagePrefs, "guest", 0);
	aucPagesRegister(NULL, "account-manager", PAGE_FILE, aucPageAccountManager, "sysop", 0);
	aucPagesRegister(NULL, "account-edit", PAGE_FILE, aucPageAddAccount, "sysop", 0);
	aucPagesRegister(NULL, "account-delete", PAGE_FILE, aucPageDeleteAccount, "sysop", 0);
	aucPagesRegister(NULL, "account-bulk-add", PAGE_FILE, aucPageBulkAddAccount, "sysop", 0);
	aucPagesRegister(NULL, "set-passwd", PAGE_FILE, aucPageSetPasswd, "sysop", 0);
	aucPagesRegister(NULL, "group-manager", PAGE_FILE, aucPageGroupManager, "sysop", 0);
	aucPagesRegister(NULL, "group-edit", PAGE_FILE, aucPageAddGroup, "sysop", 0);
	aucPagesRegister(NULL, "group-delete", PAGE_FILE, aucPageDeleteGroup, "sysop", 0);
	aucPagesRegister(NULL, "group-deluser", PAGE_FILE, aucPageDeleteUserGroup, "sysop", 0);
	aucPagesRegister(NULL, "register", PAGE_FILE, aucPageRegisterAcct, NULL, 0);
	aucPagesRegister(NULL, "acct-pending", PAGE_FILE, aucPagePendingAcct, "sysop", 0);
	aucPagesRegister(NULL, "acct-approve", PAGE_FILE, aucPageApproveAcct, "sysop", 0);
	aucPagesRegister(NULL, "help", PAGE_DIRFUNC, aucHtmlHelpPage, NULL, 0);
	aucPagesRegister(NULL, "firstrun", PAGE_FILE, aucPageFirstRun, NULL, 0);
	aucPagesRegister(NULL, "site-prefs", PAGE_FILE, aucPageSitePrefs, "sysop", 0);

	aucDBConnect();

	aucDBRegister("userinfo", "CREATE TABLE %s (uname VARCHAR(64) DEFAULT '' NOT NULL, \
         fname VARCHAR(32) NOT NULL, lname VARCHAR(32) NOT NULL, \
         mname VARCHAR(32) NOT NULL, id VARCHAR(16), \
         grp VARCHAR(16), class YEAR, \
         permissions VARCHAR(16), lastlogin DATETIME, PRIMARY KEY (uname))");
	aucDBRegister("passwd", "CREATE TABLE %s (uname VARCHAR(64) DEFAULT '' NOT NULL, \
         passwd CHAR(16) DEFAULT '' NOT NULL, PRIMARY KEY (uname))");
	aucDBRegister("groups", "CREATE TABLE %s (grp VARCHAR(16) DEFAULT '' NOT NULL, \
         PRIMARY KEY (grp))");
	aucDBRegister("groupmembers", "CREATE TABLE %s (grp VARCHAR(16) DEFAULT '' NOT NULL, \
         uname VARCHAR(64) DEFAULT '' NOT NULL, \
         PRIMARY KEY (grp,uname))");
	aucDBRegister("prefs", "CREATE TABLE %s (uname VARCHAR(64) DEFAULT '' NOT NULL, \
         title VARCHAR(32) DEFAULT '' NOT NULL, \
         fnamecfg VARCHAR(32) DEFAULT '' NOT NULL, \
         mnamecfg VARCHAR(32) DEFAULT '' NOT NULL, \
         cols CHAR(3) DEFAULT '3' NOT NULL, \
         fileexpert VARCHAR(32) DEFAULT 'Regular' NOT NULL, \
         commanderror TEXT, \
         PRIMARY KEY (uname))");
	aucDBRegister("registrations", "CREATE TABLE %s (uname VARCHAR(64) DEFAULT '' NOT NULL, \
         pid BIGINT UNSIGNED DEFAULT 0 NOT NULL AUTO_INCREMENT, \
         fname VARCHAR(32), lname VARCHAR(32), mname VARCHAR(32), \
         type VARCHAR(16), email VARCHAR(64), streetadd VARCHAR(64), \
         city VARCHAR(64), stateprov VARCHAR(30), zip VARCHAR(10), \
         phone VARCHAR(12), students TEXT, KEY autoinc (pid), \
         PRIMARY KEY (pid))");


	return 0;
}

/**
 * Updates AUC's databases when an upgrade is detected.
 *
 * This function is called whenever AUC detects that its version number has changed.
 * One of its important functions is to perform any necessary ALTER queries on the
 * SQL database if the fields of any tables have changed from version to version.
 *
 * @return Always 0
 */

int
upgrade_global(char *oldversion)
{
	MYSQL_RES *res, *res2;
	MYSQL_ROW row, row2;
	char oldid[64];
	char newid[64];

	if (!db || aucDBConnectFailed)
		return 0;

	if (strstr(oldversion, ".")) {
		/* Upgrade databases from 0.6.x to 0.7.0 */
		if (strcmp(oldversion, "0.7.0") < 0) {
			char buf[256], buf2[256];
			aucError("Migrating databases to version 0.7.0");
			if (aucDBQueryNumRows("SHOW COLUMNS FROM directory LIKE \"state\"") == 1) {
				aucDBQuery(db, "ALTER TABLE directory CHANGE state stateprov varchar(30) NOT NULL");
				aucDBQuery(db,
							"ALTER TABLE registrations CHANGE state stateprov VARCHAR(30) NOT NULL");
			}
			if (aucDBQueryNumRows("SHOW COLUMNS FROM scheds LIKE \"p5B\"") == 1) {
				aucDBQuery(db, "ALTER TABLE scheds CHANGE p8 p9 varchar(16) DEFAULT ''");
				aucDBQuery(db, "ALTER TABLE scheds CHANGE p7 p8 varchar(16) DEFAULT ''");
				aucDBQuery(db, "ALTER TABLE scheds CHANGE p6 p7 varchar(16) DEFAULT ''");
				aucDBQuery(db, "ALTER TABLE scheds CHANGE p5B p6 varchar(16) DEFAULT ''");
				aucDBQuery(db, "ALTER TABLE scheds CHANGE p5A p5 varchar(16) DEFAULT ''");
				aucDBQuery(db, "UPDATE classes SET period='9' WHERE period='8'");
				aucDBQuery(db, "UPDATE classes SET period='8' WHERE period='7'");
				aucDBQuery(db, "UPDATE classes SET period='7' WHERE period='6'");
				aucDBQuery(db, "UPDATE classes SET period='6' WHERE period='5B'");
				aucDBQuery(db, "UPDATE classes SET period='5' WHERE period='5A'");
			}
			aucDBQuery(db, "ALTER TABLE userinfo CHANGE uname uname VARCHAR(64) DEFAULT '' NOT NULL");
			aucDBQuery(db, "ALTER TABLE userinfo ADD lastlogin DATETIME");
			aucDBQuery(db, "ALTER TABLE passwd CHANGE uname uname VARCHAR(64) DEFAULT '' NOT NULL");
			aucDBQuery(db,
						"ALTER TABLE groupmembers CHANGE uname uname VARCHAR(64) DEFAULT '' NOT NULL");
			aucDBQuery(db, "ALTER TABLE prefs CHANGE uname uname VARCHAR(64) DEFAULT '' NOT NULL");
			aucDBQuery(db,
						"ALTER TABLE registrations CHANGE uname uname VARCHAR(64) DEFAULT '' NOT NULL");
			aucDBQuery(db,
						"ALTER TABLE classes CHANGE teacher teacher VARCHAR(64) DEFAULT '' NOT NULL");
			aucDBQuery(db, "ALTER TABLE classaides CHANGE aide aide VARCHAR(64) DEFAULT '' NOT NULL");
			aucDBQuery(db, "ALTER TABLE scheds CHANGE uname uname VARCHAR(64) DEFAULT '' NOT NULL");
			aucDBQuery(db, "ALTER TABLE directory CHANGE uname uname VARCHAR(64) DEFAULT '' NOT NULL");
			aucDBQuery(db,
						"ALTER TABLE peoplelinks CHANGE uname uname VARCHAR(64) DEFAULT '' NOT NULL");
			aucDBQuery(db,
						"ALTER TABLE peoplelinks CHANGE friend friend VARCHAR(64) DEFAULT '' NOT NULL");
			aucDBQuery(db, "ALTER TABLE comments CHANGE uname uname VARCHAR(64)");

			snprintf(buf, sizeof(buf), "%s/class", aucVarRoot);
			snprintf(buf2, sizeof(buf2), "%s/course", aucVarRoot);
			rename(buf, buf2);

			aucDBQuery(db, "SELECT * FROM classes");
			res = mysql_store_result(db);
			while ((row = mysql_fetch_row(res))) {
				snprintf(oldid, sizeof(oldid), "%.6s%s", row[0], row[2]);
				snprintf(newid, sizeof(newid), "%s@%s", row[0], row[6]);
				aucDBQuery(db, "UPDATE classaides SET idstr=\"%s\" WHERE idstr=\"%s\"", newid, oldid);
				aucDBQuery(db, "UPDATE classannc SET idstr=\"%s\" WHERE idstr=\"%s\"", newid, oldid);
				aucDBQuery(db, "UPDATE classsched SET idstr=\"%s\" WHERE idstr=\"%s\"", newid, oldid);
				aucDBQuery(db, "SELECT discid FROM discussions WHERE type=\"class\" AND \
					subtype=\"%s\"", oldid);
				res2 = mysql_store_result(db);
				while ((row2 = mysql_fetch_row(res2)))
					aucDBQuery(db, "INSERT INTO coursedisc (idstr, discid) VALUES (\"%s\", \"%s\")",
							   newid, row2[0]);
				mysql_free_result(res2);

				snprintf(buf, sizeof(buf), "%s/course/%s", aucVarRoot, oldid);
				snprintf(buf2, sizeof(buf2), "%s/course/%s", aucVarRoot, newid);
				rename(buf, buf2);
			}
			mysql_free_result(res);
		}

		/* Upgrade from 0.7.0 to 0.7.1 */
		if (strcmp(oldversion, "0.7.1") < 0) {
			aucError("Migrating databases to version 0.7.1");
			aucDBQuery(db, "ALTER IGNORE TABLE classes DROP aliases, ADD classtime \
				VARCHAR(64), ADD description TEXT, CHANGE room room VARCHAR(32), \
				DROP PRIMARY KEY, ADD PRIMARY KEY (id,semester)");
			aucDBQuery(db, "ALTER IGNORE TABLE classes CHANGE classtime classtime VARCHAR(64) \
				NOT NULL, CHANGE description description TEXT NOT NULL, CHANGE room room \
				VARCHAR(32) NOT NULL, CHANGE title title VARCHAR(64) NOT NULL, CHANGE \
				department department VARCHAR(32) NOT NULL, CHANGE period period CHAR(2) \
				NOT NULL");
		}
		
		/* Ensure that we get all database upgrades after this point */
		strcpy(oldversion, "0");
	}


	if (strcmp(oldversion, "000704a") < 0) {
		aucError("Migrating databases to version 000704a");
		aucDBQuery(db, "UPDATE classaides SET aide=LCASE(aide)");
		aucDBQuery(db, "UPDATE classes SET teacher=LCASE(teacher)");
		aucDBQuery(db, "UPDATE comments SET uname=LCASE(uname)");
		aucDBQuery(db, "UPDATE directory SET uname=LCASE(uname)");
		aucDBQuery(db, "UPDATE groupmembers SET uname=LCASE(uname)");
		aucDBQuery(db, "UPDATE passwd SET uname=LCASE(uname)");
		aucDBQuery(db, "UPDATE peoplelinks SET uname=LCASE(uname)");
		aucDBQuery(db, "UPDATE peoplelinks SET friend=LCASE(friend)");
		aucDBQuery(db, "UPDATE prefs SET uname=LCASE(uname)");
		aucDBQuery(db, "UPDATE registrations SET uname=LCASE(uname)");
		aucDBQuery(db, "UPDATE scheds SET uname=LCASE(uname)");
		aucDBQuery(db, "UPDATE userinfo SET uname=LCASE(uname)");
	}

	aucError("AUC database is now upgraded to version %s", DB_VERSION);
	
	/* Write the current version number to the configuration file. */
	aucConfUpdateToken("/child::auc", "version", DB_VERSION);
	aucConfSetChanged();
	return 0;
}

void
do_upgrade(void)
{
	char *version = strdup(DB_VERSION);

	aucConfToken("/child::auc/attribute::version", &version);

	/* if the version is blank, they probably copied aucconf.xml.sample to aucconf.xml
	 * by hand, so we need to run the setup tool. */
	if (!strlen(version)) {
		aucFirstRun = 1;
		free(version);
		return;
	}

	if (strcmp(version, DB_VERSION))
		upgrade_global(version);
	free(version);
}

