/*
 * Authenticated User Community
 * Copyright (c) 2000 David C. Moore
 *
 * You may distribute and/or use for any purpose modified or unmodified
 * copies of this software if you preserve the copyright notice above.
 *
 * conf.c
 *
 * Generic utility functions for parsing and writing the config file
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <unistd.h>
#include <sys/types.h>

#include <libxml/parser.h>
#include <libxml/xpath.h>
#include <libxml/HTMLparser.h>

#include <auc/auc.h>
#include <auc/util.h>
#include "conf.h"

static xmlDocPtr configdoc = NULL;
static xmlXPathContextPtr sitectx = NULL;
static xmlXPathObjectPtr siteobj = NULL;
static int configchanged = 0;

xmlDocPtr
aucConfGetDoc(void)
{
	return configdoc;
}

void
aucConfSetChanged(void)
{
	configchanged = 1;
}

int
aucConfParse(void)
{
	int result = 0;
	char buf[128];

	xmlKeepBlanksDefault(0);
	snprintf(buf, sizeof(buf), "%s/aucconf.xml", aucEtcRoot);
	
	aucSuRoot();
	if (!(configdoc = xmlParseFile(buf))) {
		aucSuUser();
		return -1;
	}
	aucSuUser();
	
	return result;
}

int
aucConfSave(void)
{
	char buf[128];

	if (configchanged) {
		snprintf(buf, sizeof(buf), "%s/aucconf.xml", aucEtcRoot);
		
		aucSuRoot();
		xmlSaveFile(buf, configdoc);
		aucSuUser();
	}
	return 0;
}

int
storetext(xmlNodePtr node, char **str)
{
	int result = 0;
	char *tmp;

	if (!node)
		return -1;

	switch (node->type) {
	case XML_ELEMENT_NODE:
		if ((tmp = xmlNodeGetContent(node))) {
/*			inlen = strlen(tmp);
			outlen = 2*inlen+10;
			*str = malloc(sizeof(char)*outlen);
			UTF8ToHtml(*str, &outlen, tmp, &inlen);
			*(*str+outlen) = '\0'; */
			*str = strdup(xmlNodeGetContent(node));
		} else
			*str = strdup("");
		break;
	case XML_ATTRIBUTE_NODE:
		if (((xmlAttrPtr) node)->children && ((xmlAttrPtr) node)->children->content)
			*str = strdup(((xmlAttrPtr) node)->children->content);
		else
			*str = strdup("");
		break;
	case XML_TEXT_NODE:
		if (node->content)
			*str = strdup(node->content);
		else
			*str = strdup("");
		break;
	default:
		return -1;
		break;
	}

	return result;
}

int
aucConfUpdateToken(char *expr, char *attrname, char *str)
{
	xmlXPathObjectPtr xpathobj;
	int result = 0;

	if (!configdoc)
		return -1;

	sitectx = xmlXPathNewContext(configdoc);
	siteobj = xmlXPathEval("/child::auc/Site", sitectx);

	sitectx->node = siteobj->nodesetval->nodeTab[0];
	xpathobj = xmlXPathEval(expr, sitectx);

	if (xpathobj && (xpathobj->type == 1) && (xpathobj->nodesetval->nodeNr > 0)) {
		if (attrname)
			xmlSetProp(xpathobj->nodesetval->nodeTab[0], attrname, str);
		else
			xmlNodeSetContent(xpathobj->nodesetval->nodeTab[0], str);
		configchanged = 1;
	} else
		result = -1;

	xmlXPathFreeObject(xpathobj);
	xmlXPathFreeContext(sitectx);

	return result;
}

int
aucConfToken(char *expr, char **str)
{
	xmlXPathObjectPtr xpathobj;
	int result = 0;

	if (!configdoc)
		return -1;

	sitectx = xmlXPathNewContext(configdoc);
	siteobj = xmlXPathEval("/child::auc/Site", sitectx);
	
	sitectx->node = siteobj->nodesetval->nodeTab[0];
	xpathobj = xmlXPathEval(expr, sitectx);

	if (xpathobj && (xpathobj->type == 1) && (xpathobj->nodesetval->nodeNr > 0))
		result = storetext(xpathobj->nodesetval->nodeTab[0], str);
	else
		result = -1;

	xmlXPathFreeObject(xpathobj);
	xmlXPathFreeContext(sitectx);

	return result;
}

int
aucConfTokenExist(char *expr)
{
	xmlXPathObjectPtr xpathobj;
	int result = 0;

	if (!configdoc)
		return 0;

	sitectx = xmlXPathNewContext(configdoc);
	siteobj = xmlXPathEval("/child::auc/Site", sitectx);

	sitectx->node = siteobj->nodesetval->nodeTab[0];
	xpathobj = xmlXPathEval(expr, sitectx);

	if (xpathobj && (xpathobj->type == 1) && (xpathobj->nodesetval->nodeNr > 0))
		result = 1;
	else
		result = 0;

	xmlXPathFreeObject(xpathobj);
	xmlXPathFreeContext(sitectx);

	return result;
}

int
aucConfTokens(char *expr, char ***str)
{
	xmlXPathObjectPtr xpathobj;
	int result = 0, i;

	if (!configdoc)
		return -1;

	sitectx = xmlXPathNewContext(configdoc);
	siteobj = xmlXPathEval("/child::auc/Site", sitectx);

	sitectx->node = siteobj->nodesetval->nodeTab[0];
	xpathobj = xmlXPathEval(expr, sitectx);

	if (xpathobj && (xpathobj->type == 1) && (xpathobj->nodesetval->nodeNr > 0)) {
		*str = (char **) malloc(sizeof(char *));
		(*str)[0] = NULL;
		for (i = 0; i < xpathobj->nodesetval->nodeNr; i++) {
			*str = (char **) realloc(*str, (i + 2) * sizeof(char *));
			storetext(xpathobj->nodesetval->nodeTab[i], &((*str)[i]));
			(*str)[i + 1] = NULL;
		}
	} else
		result = -1;

	xmlXPathFreeObject(xpathobj);
	xmlXPathFreeContext(sitectx);

	return result;
}
