/*
 * Authenticated User Community
 * Copyright (c) 2000 David C. Moore
 *
 * You may distribute and/or use for any purpose modified or unmodified
 * copies of this software if you preserve the copyright notice above.
 *
 * coursecal.c
 *
 * Routines for the display and modification of course calendars
 * in each iClassroom
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <malloc.h>

#include <auc/cgic.h>
#include <auc/auc.h>
#include <auc/course.h>
#include <auc/calendar.h>
#include <auc/coursenode.h>
#include <auc/util.h>
#include <auc/pages.h>
#include <auc/error.h>
#include <auc/html.h>

#include "coursecal.h"

char *assntypes[] = {
	N_("Test"),
	N_("Quiz"),
	N_("Long Term"),
	N_("HW Due"),
	N_("Objective"),
	N_("Other"),
	NULL
};

int printitems(FILE *, aucCourseInfoPtr clinfo, unsigned long, int, int, int);
char *typequery(unsigned long);
void calendar_node(FILE * output, aucCourseNodeInfoPtr node, aucCourseInfoPtr clinfo);

int
aucCourseCalInit(void)
{
	aucPagesRegister(NULL, "course-cal", PAGE_FILE, aucCourseCalView, "guest", 0);
	aucPagesRegister(NULL, "coursecal-edit", PAGE_FILE, aucCourseCalAdd, "normal", 0);
	aucPagesRegister(NULL, "coursecal-del", PAGE_FILE, aucCourseCalDel, "normal", 0);

	aucCourseNodeRegister("calendar",
						  _("Course Calendar"),
						  calendar_node,
						  5, _("A schedule of upcoming assignments and events for the course."), 0);

	return 0;
}

void
calendar_node(FILE * output, aucCourseNodeInfoPtr node, aucCourseInfoPtr clinfo)
{
	fprintf(output, "<div align=right><small><b><a href=\"../course-cal?course=%s\">%s</a>, \n",
			aucLinkText(clinfo->uniquestr), _("Complete Calendar"));
	fprintf(output, "<a href=\"../course-cal?type=log&course=%s\">%s</a></b></small></div>\n",
			aucLinkText(clinfo->uniquestr), _("Course Log"));
	aucCourseCalPreview(output, clinfo);
	if (clinfo->admin)
		fprintf(output,
				"<small><b><a href=\"../coursecal-edit?course=%s\">%s</a></b></small><br>\n",
				aucLinkText(clinfo->uniquestr), _("Add New Item"));
}

int
aucCourseCalView(FILE * output, char *extrapath)
{
	char course[32], buf[512];
	aucCourseInfo clinfo;
	int log = 0;

	cgiHeaderContentType("text/html");

	if (cgiFormSuccess != cgiFormString("course", course, sizeof(course)))
		return -1;
	if (aucCourseGetInfo(&clinfo, course, 0))
		return -1;

	if (cgiFormSuccess == cgiFormString("type", buf, sizeof(buf)))
		if (!strcmp(buf, "log"))
			log = 1;

	aucCourseCompactHeader(output, &clinfo, log ? _("Course Logbook") : _("Course Calendar"));

	if (log)
		fprintf(output, "<p>%s</p>\n\n",
				_("This is the log of all activies and assignments\n"
				  "that have already occurred in this course."));
	else
		fprintf(output, "<p>%s</p>\n\n",
				_("This is a list of upcoming assignments, tests/quizzes,\n"
				  "and activities coming up in this course."));

	printitems(output, &clinfo, 0, 0, 1, log);

	aucHtmlSlimFoot();
	aucCourseFreeInfo(&clinfo);

	return 0;

}

int
aucCourseCalPreview(FILE * output, aucCourseInfoPtr clinfo)
{
	fprintf(output, "<table cellspacing=1 cellpadding=0 border=0 width=100%%>\n");
	fprintf(output, "<tr bgcolor=#CCCCCC valign=top><td width=50%%>\n");
	fprintf(output, "<b><small>%s</small></b>\n", _("Next 5 Days"));
	fprintf(output, "</td><td width=50%%>\n");
	fprintf(output, "<b><small>%s</small></b>\n", _("Tests/Quizzes"));
	fprintf(output, "</td></tr>\n\n");

	fprintf(output, "<tr valign=top><td width=50%% rowspan=3>\n");
	printitems(output, clinfo, ASSN_TEST | ASSN_QUIZ | ASSN_LONGTERM | ASSN_HWORK, 7, 0, 0);
	fprintf(output, "</td><td width=50%%>\n");
	printitems(output, clinfo, ASSN_TEST | ASSN_QUIZ, -2, 0, 0);
	fprintf(output, "</td></tr>\n\n");

	fprintf(output, "<tr bgcolor=#CCCCCC valign=top><td width=50%%>\n");
	fprintf(output, "<b><small>%s</small></b>\n", _("Long Term Work"));
	fprintf(output, "</td></tr>\n\n");

	fprintf(output, "<tr valign=top><td width=50%%>\n");
	printitems(output, clinfo, ASSN_LONGTERM, -2, 0, 0);
	fprintf(output, "</td></tr>\n\n");

	fprintf(output, "</table>\n\n");
	return 0;
}

int
printitems(FILE * output, aucCourseInfoPtr clinfo, unsigned long flags,
		   int daycount, int verbose, int log)
{
	MYSQL_RES *res;
	MYSQL_ROW row;
	char buf[512], lastday[32];
	int dcount = 0;

	strcpy(lastday, "");
	if (verbose)
		fprintf(output, "<p>\n");
	else
		fprintf(output, "<small>\n");
	if (log)
		snprintf(buf, sizeof(buf), "SELECT name, comment, DATE_FORMAT(assndate, \"%s\"), \
         assntype, sid, assndate, block FROM classsched WHERE idstr=\"%s\" AND (curdate() >= assndate) \
         AND (assndate > DATE_SUB(curdate(), INTERVAL %d day)) %s ORDER BY assndate DESC, sid ASC", "%a, %b %d", clinfo->uniquestr, (daycount > 0) ? daycount : 45, typequery(flags));
	else
		snprintf(buf, sizeof(buf), "SELECT name, comment, DATE_FORMAT(assndate, \"%s\"), \
         assntype, sid, assndate, block FROM classsched WHERE idstr=\"%s\" AND (curdate() <= assndate) \
         AND (assndate < DATE_ADD(curdate(), INTERVAL %d DAY)) %s ORDER BY assndate ASC, sid ASC", "%a, %b %d", clinfo->uniquestr, (daycount > 0) ? daycount : 200, typequery(flags));

	if (mysql_query(db, buf)) {
		aucError(mysql_error(db));
	} else {
		res = mysql_store_result(db);
		if (verbose) {
			fprintf(output, "<ul>\n");
			if (clinfo->admin && log)
				fprintf(output, "<li><a href=\"coursecal-edit?course=%s&type=log\">%s</a></li>\n",
						aucLinkText(clinfo->uniquestr), _("Add Log Entry"));
			else if (clinfo->admin)
				fprintf(output, "<li><a href=\"coursecal-edit?course=%s\">%s</a></li>\n",
						aucLinkText(clinfo->uniquestr), _("Add New Calendar Item"));
		}
		while ((row = mysql_fetch_row(res))) {
			if ((daycount < 0) && (dcount == -daycount))
				break;
			if (strcmp(row[2], lastday)) {
				if (verbose)
					fprintf(output, "</ul>\n");
				fprintf(output, "%s<u><b>%s:</b></u>\n\n",
						(!verbose && dcount) ? "<br>" : "", row[2]);
				strcpy(lastday, row[2]);
				dcount++;
				if (verbose)
					fprintf(output, "<ul>\n");
			}
			if (!strcmp(row[3], _("Comments"))) {
				fprintf(output, "<li><dl><dt><b>%s:</b>\n", _("Comments"));
				if (verbose && clinfo->admin)
					fprintf(output, "- <a href=\"coursecal-edit?course=%s&sid=%s\">%s</a>, "
							"<a href=\"coursecal-del?course=%s&sid=%s\">%s</a>\n",
							aucLinkText(clinfo->uniquestr), row[4], _("Edit"),
							aucLinkText(clinfo->uniquestr), row[4], _("Del"));
				fprintf(output, "</dt>\n");
				fprintf(output, "<dd>\n");
				aucPrintText(row[1], output, TEXT_BSPACES);
				fprintf(output, "</dd></dl>\n\n");
			} else {
				fprintf(output, "%s %s: %s\n", verbose ? "<li><dl><dt>" : "<br><b>o</b>", row[3],
						row[0]);
				if (verbose && clinfo->admin)
					fprintf(output,
							"- <a href=\"coursecal-edit?course=%s&sid=%s\">%s</a>, "
							"<a href=\"coursecal-del?course=%s&sid=%s\">%s</a>\n",
							aucLinkText(clinfo->uniquestr), row[4], _("Edit"),
							aucLinkText(clinfo->uniquestr), row[4], _("Del"));
				if (verbose && strlen(row[1])) {
					fprintf(output, "</dt><dd>\n");
					aucPrintText(row[1], output, TEXT_BSPACES);
					fprintf(output, "</dd></dl>\n\n");
				} else if (verbose)
					fprintf(output, "</dt></dl>\n\n");
			}
		}
		if (verbose)
			fprintf(output, "</ul>\n");
		mysql_free_result(res);
	}
	if (!dcount)
		fprintf(output, "%s\n", _("None"));
	if (verbose)
		fprintf(output, "</p>\n\n");
	else
		fprintf(output, "</small>\n\n");
	return 0;
}

char *
typequery(unsigned long types)
{
	char *ret;
	int count = 0;
	int i;
	char buf[256];

	ret = (char *) strdup("");
	for (i = 0; assntypes[i]; i++)
		if (types & (0x1 << i)) {
			if (count) {
				ret = (char *) realloc(ret, strlen(ret) + 4);
				strcat(ret, "OR ");
			} else {
				ret = (char *) realloc(ret, strlen(ret) + 7);
				strcat(ret, "AND ( ");
			}
			ret = (char *) realloc(ret, strlen(ret) + strlen(assntypes[i]) + 13);
			snprintf(buf, sizeof(buf), "assntype=\"%s\" ", assntypes[i]);
			strcat(ret, buf);
			count++;
		}
	if (count) {
		ret = (char *) realloc(ret, strlen(ret) + 2);
		strcat(ret, ")");
	}
	return ret;
}

int
aucCourseCalDel(FILE * output, char *extrapath)
{
	char buf[512], course[32], sid[20];
	aucCourseInfo clinfo;

	strcpy(sid, "");
	if (cgiFormSuccess != cgiFormString("course", course, sizeof(course)))
		return -1;
	if (aucCourseGetInfo(&clinfo, course, 0))
		return -1;

	cgiFormString("sid", sid, sizeof(sid));
	if (!clinfo.admin)
		return -2;
	if (strlen(sid)) {
		snprintf(buf, sizeof(buf), "DELETE FROM classsched WHERE idstr=\"%s\" AND sid=\"%s\"",
				 clinfo.uniquestr, sid);
		if (mysql_query(db, buf))
			aucError(mysql_error(db));
	}
	snprintf(buf, sizeof(buf), "course/%s", aucLinkText(clinfo.uniquestr));
	cgiHeaderLocation(buf);
	aucCourseFreeInfo(&clinfo);
	return 0;
}

int
aucCourseCalAdd(FILE * output, char *extrapath)
{
	MYSQL_RES *res = NULL;
	MYSQL_ROW row = NULL;
	int x, num, edit = 0, nogood = 0, comments = 0;
	char buffer[8192], buffer2[256], course[32], sid[16], *fields[32];
	aucCourseInfo clinfo;

	for (x = 0; x < 32; x++)
		fields[x] = NULL;


	if (cgiFormSuccess != cgiFormString("course", course, sizeof(course)))
		return -1;
	if (aucCourseGetInfo(&clinfo, course, 0))
		return -1;
	if (!clinfo.admin)
		return -2;

	if (!strcmp(cgiRequestMethod, "GET")) {
		if (cgiFormSuccess == cgiFormString("sid", sid, sizeof(sid))) {
			snprintf(buffer, sizeof(buffer),
					 "SELECT * FROM classsched WHERE idstr=\"%s\" AND sid=%s", clinfo.uniquestr,
					 sid);
			if (mysql_query(db, buffer))
				aucError(mysql_error(db));
			else {
				res = mysql_store_result(db);
				if ((row = mysql_fetch_row(res))) {
					edit = 1;
					for (x = 0, num = mysql_num_fields(res); x < num; x++)
						fields[x] = row[x];
					fields[7] = NULL;
				}
			}
		}
		if (cgiFormSuccess == cgiFormString("type", buffer, sizeof(buffer)))
			if (!strcmp(buffer, "log"))
				comments = 1;
		if (fields[5] && !strcmp(fields[5], _("Comments")))
			comments = 1;
		if (cgiFormSuccess == cgiFormString("date", buffer, sizeof(buffer)))
			fields[4] = (char *) strdup(buffer);
	} else {
		time_t instime = time(NULL);
		int deleted = 0;

		aucHtmlReadForm("course sid name comment assndate assntype",
						   fields, STR_ESCAPE | STR_ENCBLOCK);
		if (fields[5] && !strcmp(fields[5], _("Comments")))
			comments = 1;

		snprintf(buffer2, sizeof(buffer2), "DELETE FROM classsched WHERE idstr=\"%s\" \
         AND sid=%s", fields[0], fields[1]);

		snprintf(buffer, sizeof(buffer),
				 "INSERT INTO classsched (idstr, name, comment, assndate, assntype, block, instime) \
         VALUES (\"%s\", \"%s\", \"%s\", \"%s\", \"%s\", \"%s\", %ld)",
				 fields[0], fields[2], fields[3], fields[4], fields[5], "", (long int) instime);

		if (strlen(fields[0]) && strlen(fields[2])
			&& (strcmp(fields[5], _("Comments")) || strlen(fields[3]))) {
			if (strlen(fields[1]) && !deleted && mysql_query(db, buffer2))
				aucError(mysql_error(db));
			deleted = 1;
			if (mysql_query(db, buffer))
				aucError(mysql_error(db));
		} else
			nogood = 1;
		if (!nogood) {
			if (!deleted && strlen(fields[1]) && mysql_query(db, buffer2))
				aucError(mysql_error(db));
			snprintf(buffer, sizeof(buffer), "course/%s", aucLinkText(fields[0]));
			cgiHeaderLocation(buffer);
			return 0;
		}
	}

	cgiHeaderContentType("text/html");

	if (edit) {
		aucHtmlSlimHead("%s", _("Edit Item"));
	} else {
		aucHtmlSlimHead("%s", _("Add New Calendar Item"));
	}

	if (nogood)
		fprintf(output, "<p><b>%s:</b> %s</p>\n\n", _("Error"),
				_("You must specify an item name."));

	aucDrawTableForm(output, "POST", "coursecal-edit");
	aucDrawHidden(output, "course", fields[0] ? fields[0] : clinfo.uniquestr);
	aucDrawHidden(output, "sid", fields[1]);

	if (comments) {
		aucDrawHidden(output, "assntype", "Comments");
		aucDrawHidden(output, "name", "log");
	} else {
		aucDrawDropdown(output, "assntype", assntypes, assntypes, fields[5], _("Type of Item"));
		aucDrawTextbox(output, "name", 50, 64, fields[2], _("Description"));
	}
	aucDrawTextarea(output, "comment", 3, 50, fields[3], _("Comments"));

	aucDrawDateSchool(output, "assndate", fields[4], aucCurrSemester, _("Date"));

	if (edit)
		aucDrawSubmit(output, _("Save Item"));
	else
		aucDrawSubmit(output, _("Add Item"));
	aucDrawTableFormEnd(output);
	aucHtmlSlimFoot();

	if (edit)
		mysql_free_result(res);

	if (strcmp(cgiRequestMethod, "GET"))
		for (x = 0; fields[x]; x++)
			free(fields[x]);

	aucCourseFreeInfo(&clinfo);
	return 0;
}
