/*
 * Authenticated User Community
 * Copyright (c) 2000 David C. Moore
 *
 * You may distribute and/or use for any purpose modified or unmodified
 * copies of this software if you preserve the copyright notice above.
 *
 * database.c
 *
 * Routines for general control of the SQL database
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>

#include <auc/cgic.h>
#include <auc/auc.h>
#include <auc/html.h>
#include <auc/error.h>

#include "database.h"

int
aucDBConnect(void)
{
	if (!aucDBUser || !strlen(aucDBUser) || !aucDBName || !strlen(aucDBName)) {
		aucDBConnectFailed = 1;
		return -1;
	}

	db = mysql_init(db);

	if (!mysql_real_connect(db, aucDBHost, aucDBUser, aucDBPasswd, aucDBName, 0, NULL, 0)) {
		aucError(mysql_error(db));
		aucError("ERROR: Failed to connect to MySQL database.");
		aucDBConnectFailed = 1;
		return -1;
	}
	return 0;
}

int
aucDBDisconnect(void)
{
	if (db)
		mysql_close(db);
	return 0;
}

int
aucDBRegister(char *name, char *insertquery)
{
	MYSQL_RES *res = NULL;
	char buffer[512];

	if (!db || aucDBConnectFailed)
		return 0;

	res = mysql_list_tables(db, name);
	if (res && (mysql_num_rows(res) != 1)) {
		snprintf(buffer, sizeof(buffer), insertquery, name);
		if (mysql_query(db, buffer)) {
			aucError(mysql_error(db));
			snprintf(buffer, sizeof(buffer), "Failed to create table \"%s\"!", name);
			aucError(buffer);
		} else {
			snprintf(buffer, sizeof(buffer), "Table \"%s\" created successfully", name);
			aucError(buffer);
		}
	}
	mysql_free_result(res);
	return 0;
}

unsigned long
aucDBNumRows(char *table, char *where)
{
	MYSQL_RES *res;
	MYSQL_ROW row;
	char buffer[1024];
	unsigned long num = 0;

	if (where)
		snprintf(buffer, sizeof(buffer), "SELECT COUNT(*) FROM %s WHERE %s", table, where);
	else
		snprintf(buffer, sizeof(buffer), "SELECT COUNT(*) FROM %s", table);

	if (mysql_query(db, buffer))
		aucError(mysql_error(db));
	else {
		res = mysql_store_result(db);
		if ((row = mysql_fetch_row(res)))
			num = strtoul(row[0], NULL, 10);
		mysql_free_result(res);
	}

	return num;
}

unsigned long
aucDBQueryNumRows(char *query)
{
	MYSQL_RES *res;
	unsigned long num = 0;

	if (mysql_query(db, query))
		aucError(mysql_error(db));
	else {
		res = mysql_store_result(db);
		num = mysql_num_rows(res);
		mysql_free_result(res);
	}

	return num;
}

int
aucDBQuery(MYSQL * indb, char *fquery, ...)
{
	va_list ap;
	char buf[4096];
	int result;

	va_start(ap, fquery);
	vsnprintf(buf, sizeof(buf), fquery, ap);
	va_end(ap);
	result = mysql_query(indb, buf);
	if (result)
		aucError("MySQL: %s", mysql_error(indb));
	return result;
}

char *
aucDBProcessFormCreate(void)
{
	char *values[8];
	MYSQL *tempdb;

	aucHtmlReadForm("rootpasswd username password host name", values, 0);
	if (!strlen(values[1]))
		return _("A SQL username was not entered.");
	if (!strlen(values[2]))
		return _("A SQL password was not entered (blank passwords are disallowed).");
	if (!strlen(values[3]))
		return _("A SQL hostname was not entered.");
	if (!strlen(values[4]))
		return _("A SQL database name was not entered.");

	tempdb = mysql_init(NULL);
	if (!mysql_real_connect(tempdb, values[3], "root", values[0], NULL, 0, NULL, 0))
		return mysql_error(tempdb);

	aucDBQuery(tempdb, "CREATE DATABASE %s", values[4]);
	aucDBQuery(tempdb, "GRANT ALL PRIVILEGES ON %s.* TO %s@%s IDENTIFIED BY '%s'",
			   values[4], values[1], (!strcmp(values[3], "localhost")) ? "localhost" : cgiRemoteHost,
			   values[2]);
	mysql_close(tempdb);

	return NULL;
}

int
aucDBStoreResults(aucVarField * vars, MYSQL_RES * res, MYSQL_ROW row)
{
	MYSQL_FIELD *fields;
	int num_fields, i, x = 0;

	for (x = 0; vars[x].name; x++);

	num_fields = mysql_num_fields(res);
	fields = mysql_fetch_fields(res);
	for (i = 0; i < num_fields; i++, x++) {
		vars[x].name = strdup(fields[i].name);
		if (row[i])
			vars[x].value = strdup(row[i]);
		else
			vars[x].value = NULL;
	}
	vars[x].name = NULL;
	return 0;
}

int
aucDBEscapeResults(aucVarField * vars)
{
	int x;
	char *tmp;

	for (x = 0; vars[x].name; x++) {
		if (vars[x].value) {
			tmp = malloc(strlen(vars[x].value) * 2 + 1);
			mysql_escape_string(tmp, vars[x].value, strlen(vars[x].value));
			vars[x].value = realloc(vars[x].value, strlen(tmp) + 1);
			sprintf(vars[x].value, "%s", tmp);
			free(tmp);
		} else
			vars[x].value = strdup("");
	}
	return 0;
}

int
aucDBInsertQuery(char *str, int len, char *table, aucVarField * vars, ...)
{
	va_list ap;
	int chrs = 0, started = 0;
	char *tmp;

	snprintf(str, len, "INSERT INTO %s (", table);
	chrs = strlen(str);

	va_start(ap, vars);
	while ((tmp = va_arg(ap, char *)))
	{
		if (started) {
			if ((chrs + strlen(", ")) >= len)
				return -1;
			strcat(str, ", ");
			chrs += strlen(", ");
		}
		if ((chrs + strlen(tmp)) >= len)
			return -1;
		strcat(str, tmp);
		chrs += strlen(tmp);
		started = 1;
	}
	va_end(ap);

	if ((chrs + strlen(") VALUES (")) >= len)
		return -1;
	strcat(str, ") VALUES (");
	chrs += strlen(") VALUES (");

	started = 0;
	va_start(ap, vars);
	while ((tmp = va_arg(ap, char *)))
	{
		char *tok;
		if (started) {
			if ((chrs + strlen(", ")) >= len)
				return -1;
			strcat(str, ", ");
			chrs += strlen(", ");
		}
		if (!(tok = aucDBGetToken(vars, tmp)))
			tok = "";
		if ((chrs + strlen("\"")) >= len)
			return -1;
		strcat(str, "\"");
		chrs += strlen("\"");
		if ((chrs + strlen(tok)) >= len)
			return -1;
		strcat(str, tok);
		chrs += strlen(tok);
		if ((chrs + strlen("\"")) >= len)
			return -1;
		strcat(str, "\"");
		chrs += strlen("\"");
		started = 1;
	}
	va_end(ap);

	if ((chrs + strlen(")")) >= len)
		return -1;
	strcat(str, ")");
	chrs += strlen(")");
	return 0;
}

char *
aucDBGetToken(aucVarField * vars, char *name)
{
	int x;

	for (x = 0; vars[x].name; x++) {
		if (!strcmp(vars[x].name, name))
			return vars[x].value;
	}
	return NULL;
}
