/*
 * Authenticated User Community
 * Copyright (c) 2000 David C. Moore
 *
 * You may distribute and/or use for any purpose modified or unmodified
 * copies of this software if you preserve the copyright notice above.
 *
 * email.c
 *
 * Low-level email manipulation routines for the web-based email
 * client
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <fcntl.h>
#include <errno.h>
#include <netdb.h>

#include <auc/cgic.h>
#include <auc/auc.h>
#include <auc/sock.h>
#include <auc/pages.h>
#include <auc/util.h>
#include <auc/error.h>
#include <auc/html.h>
#include <auc/conf.h>


#include "email.h"

struct options
{
	char personal[64], sigfile[64], domain[32], fcc[32], *hostlist[5];
	int newwindow, usingimap, pagelength;
};

#define PR_PINE   1
#define PR_YESNO  2
#define PR_SIG    3
#define PR_CGI    4
#define PR_IMAPAUTH   5
#define PR_SIZE   6
#define PR_FORWARD   7

struct pineopt
{
	char *setting, *value;
	int path, level, type;
	char *caption, *desc;
}
prefs[] =
{
	{
		"personal-name", 0, 0, 1, PR_PINE, N_("Personal Name"),
			N_("Appears in the \"From:\" field of outgoing messages.")}
	, {
		"forward-email", "", 0, 1, PR_FORWARD, N_("Email Forwarding"),
			N_
			("Forwards locally received email (not IMAP) to the specified address.  Leave blank for no forwarding.")}
	, {
		"user-domain", 0, 0, 1, PR_PINE, N_("User Domain"),
			N_("The domain name used in the return address of outgoing mail.")}
/*	, {
		"inbox-path", 0, 1, 1, PR_PINE, N_("Inbox Path"),
			N_("The location of the mbox file to be used as the system inbox.")} */
	, {
		"folder-collections", 0, 0, 1, PR_PINE, N_("Folder Collections"),
			N_("The path to additional mail folders.")}
	, {
		"default-fcc", "sent-mail", 0, 1, PR_PINE, N_("Default FCC"),
			N_("Outgoing mail is copied to this folder.")}
	, {
		"page-length", "20", 0, 0, PR_SIZE, N_("Max. Index Page Length"),
			N_("The maximum number of messages to display on one page of a mailbox's message list.")}
	, {
		"signature-file", ".signature", 1, 1, PR_PINE, N_("Signature File"),
			N_("File appended to end of outgoing mail.")}
	, {
		"new-window-on-compose", N_("no"), 0, 0, PR_YESNO, N_("New Window On Compose"),
			N_("Spawns a new browser window for message composition.")}
	, {
		"edit-sig", "", 0, 0, PR_SIG, N_("Edit Signature File"),
			N_("This text appears automatically at the end of composed email.")}
	, {
		"imap-server", 0, 0, 0, PR_PINE, N_("IMAP Server"),
			N_
			("The hostname of a mail server that supports the IMAP protocol.  This server will be used for all mail functions.")}
	, {
		"imap-user", 0, 0, 0, PR_IMAPAUTH, N_("IMAP User"),
			N_
			("If you are using an IMAP server, type your username, password, and email address on the server.")}
	, {
	"imap-name", 0, 0, 2, PR_PINE, N_("IMAP Name"), ""}
	, {
	"imap-passwd", 0, 0, 2, PR_PINE, N_("IMAP Password"), ""}
	, {
	"advanced-ops", N_("advanced"), 0, 0, PR_CGI, N_("Edit Advanced Options"), ""}
	, {
	NULL}
};


struct options user;

void getdefaultopt(char *, size_t, char *);
void getcustomopt(char *, size_t, char *);
int getpinekey(char *, size_t, char *, char *);
void getmailopt(char *, size_t, char *);
void dropcase(char *);


/*
char * aucEmailClGetAddress(char * uname)
{
   char buf[256], *temp;
   
   getdefaultopt(buf, sizeof(buf), "user-domain");
   
   if (!getpwnam(uname))
      return NULL;
   
   temp = (char *) malloc((strlen(uname)+strlen(buf)+2)*sizeof(char));
   sprintf(temp, "%s@%s", uname, buf);
   return temp;
}

*/

void
aucComposeHref(char *email, FILE * output)
{
	if (!aucAuthenticated) {
		fprintf(output, "<a href=\"mailto:%s\">", email);
		return;
	}

	aucEmailClParseOptions();

	if (!user.usingimap && (aucMailSourceType == LOCALSPOOL) && !aucUserCache->local)
		fprintf(output, "<a href=\"mailto:%s\">", email);
	else
		fprintf(output, "<a %s href=\"%s/email/compose?to=%s\">",
				user.newwindow ? "target=\"_blank\"" : "", aucPagesRoot, email);
}

void
aucLinkComposeMessage(FILE * output, char *email, char *fullname, char *text)
{
	char buf[256];

	aucEmailClParseOptions();
	if (fullname && strlen(fullname))
		snprintf(buf, sizeof(buf), "%s <%s>", fullname, email);
	else
		snprintf(buf, sizeof(buf), "%s", email);

	if (!user.usingimap && (aucMailSourceType == LOCALSPOOL) && !aucUserCache->local)
		fprintf(output, "<a href=\"mailto:%s\">%s</a>\n", aucLinkText(buf), text);
	else
		fprintf(output, "<a %s href=\"%s/email/compose?to=%s\">%s</a>\n",
				user.newwindow ? "target=\"_blank\"" : "", aucPagesRoot, aucLinkText(buf), text);
}

int
aucEmailClMoveTo(char *dest, char *sequence)
{
	char *args[] = {
		dest,
		sequence,
		NULL
	};
	int sock, retval;

	if ((sock = aucSockConnect("aucemail", aucUserCache->uname)) == -1)
		return -1;

	aucSockSendCommand(sock, "aucemail", aucUserCache->authstring, EMAIL_MOVE_TO, args);
	retval = aucSockRecvReply(sock);
	close(sock);

	return retval;
}

int
aucEmailClParseOptions()
{
	char buf[256];

	getmailopt(user.personal, sizeof(user.personal), "personal-name");
	getmailopt(user.domain, sizeof(user.domain), "user-domain");
	getmailopt(user.fcc, sizeof(user.fcc), "default-fcc");
	getmailopt(user.sigfile, sizeof(user.sigfile), "signature-file");
//   getmailopt(user.imapemail, sizeof(user.imapemail), "imap-email");
	getmailopt(buf, sizeof(buf), "page-length");
	user.pagelength = strtol(buf, NULL, 10);

	getmailopt(buf, sizeof(buf), "new-window-on-compose");
	dropcase(buf);
	if (*buf == 'y')
		user.newwindow = 1;
	else
		user.newwindow = 0;

	getmailopt(buf, sizeof(buf), "imap-server");
	if (strcmp(buf, _("None")))
		user.usingimap = 1;
	else
		user.usingimap = 0;

	return 1;
}

int mail_debug = 0;

int
aucEmailClInit(char *mailbox, int debug)
{
	char buf[1024];
	char *args[11];
	int sock, retval, x;

	mail_debug = debug;

	aucEmailClParseOptions();

	getmailopt(buf, sizeof(buf), "imap-server");
	if (strcmp(buf, _("None")))
		args[0] = strdup(buf);
	else {
		if ((aucMailSourceType == LOCALSPOOL) && !aucUserCache->local)
			return -2;
		args[0] = strdup("");
	}

	getmailopt(buf, sizeof(buf), "inbox-path");
	args[1] = strdup(buf);

	getmailopt(buf, sizeof(buf), "folder-collections");
	if (buf[strlen(buf) - 2] == '[')
		buf[strlen(buf) - 2] = '\0';
	args[2] = strdup(buf);

	args[3] = aucSMTPHost;
	args[4] = mailbox;
	args[5] = aucMailTrashFolder;
	getmailopt(buf, sizeof(buf), "imap-name");
	args[6] = strdup(buf);
	getmailopt(buf, sizeof(buf), "imap-passwd");
	args[7] = strdup(buf);
	args[8] = aucUserCache->homedir;
	snprintf(buf, sizeof(buf), "%d", debug);
	args[9] = strdup(buf);
	args[10] = NULL;

	if ((sock = aucSockConnect("aucemail", aucUserCache->uname)) == -1) {
		aucEmailClStartAgent(aucUserCache->uname, aucUserCache->authstring);
		for (x = 0; x < 20; x++) {
			sock = aucSockConnect("aucemail", aucUserCache->uname);
			if (sock != -1)
				break;
			usleep(250000);
		}
	}

	if (sock == -1)
		return -1;

	if (aucSockSendCommand(sock, "aucemail", aucUserCache->authstring, EMAIL_INIT, args) == -1)
		retval = -1;
	else
		retval = aucSockRecvReply(sock);

	close(sock);

	free(args[0]);
	free(args[1]);
	free(args[2]);
	free(args[6]);
	free(args[7]);
	free(args[9]);

	aucEmailClSendOptions();


	return retval;

}

int
aucEmailClStartAgent(char *uname, char *authstring)
{
	int x;
	pid_t child;

	aucSuRoot();
	aucCheckDir("/tmp/aucemail", 01777, 0, 0);
	aucSuUser();


	child = fork();
	if (child == -1) {
		aucError("ERROR: aucemail fork() failed: %s", strerror(errno));
		return -1;
	} else if (child == 0) {
		char buf[128];
		close(0);
		close(1);

		/* give up all root privileges */
		if (getuid() != geteuid())
			setuid(geteuid());

		snprintf(buf, sizeof(buf), "%s/aucemail", aucDataDir);
		execl(buf, "aucemail", aucUserCache->uname, NULL);
		aucError("ERROR: aucemail exec() failed: %s", strerror(errno));
		exit(0);
	}


	for (x = 0; x < 20; x++)
		authstring[x] = (int) ((92.0 * rand()) / RAND_MAX) + 33;
	authstring[20] = '\0';
	return 0;
}

int
aucEmailClSendOptions(void)
{
	char *args[] = {
		user.personal,
		user.domain,
		user.fcc,
		user.sigfile,
		"",
		NULL,
		NULL,
		aucMailPostponedFolder,
		aucUserCache->uname,
		aucUserCache->homedir,
		aucUserDir,
		NULL,
		NULL,
		NULL,
		aucGraphDir,
		aucMailSavedFolder,
		aucPagesRoot,
		aucEtcRoot,
		NULL
	};
	char buf[64];
	int sock, retval;

	getmailopt(buf, sizeof(buf), "page-length");
	args[5] = strdup(buf);
	getmailopt(buf, sizeof(buf), "new-window-on-compose");
	args[6] = strdup(buf);
	snprintf(buf, sizeof(buf), "%u", aucUserCache->uid);
	args[11] = strdup(buf);
	snprintf(buf, sizeof(buf), "%u", aucUserCache->uid);
	args[12] = strdup(buf);
	snprintf(buf, sizeof(buf), "%u", aucUserCache->local);
	args[13] = strdup(buf);

	if ((sock = aucSockConnect("aucemail", aucUserCache->uname)) == -1)
		return -1;

	aucSockSendCommand(sock, "aucemail", aucUserCache->authstring, EMAIL_OPTIONS, args);
	retval = aucSockRecvReply(sock);
	close(sock);

	free(args[5]);
	free(args[6]);
	free(args[11]);
	free(args[12]);
	free(args[13]);

	return retval;
}

int
aucEmailClShutdown(void)
{
	int sock;

	if ((sock = aucSockConnect("aucemail", aucUserCache->uname)) == -1)
		return -1;

	aucSockSendCommand(sock, "aucemail", aucUserCache->authstring, EMAIL_CLOSE, NULL);
	close(sock);

	strcpy(aucUserCache->authstring, "");

	return 0;
}

int
aucEmailClClose(void)
{
	return 0;
}

int
aucEmailClFirstNew(FILE * output, char *extrapath)
{
	int firstnew, sock;
	char buf[1024];

	if (!aucEmailClInit("INBOX", mail_debug)) {
		if ((sock = aucSockConnect("aucemail", aucUserCache->uname)) == -1) {
			cgiHeaderLocation("view");
			return 0;
		}
		aucSockSendCommand(sock, "aucemail", aucUserCache->authstring, EMAIL_FIRST_NEW, NULL);
		firstnew = aucSockRecvReply(sock);
		close(sock);
		snprintf(buf, sizeof(buf), "view?mailbox=INBOX&UID=%d", firstnew);
		cgiHeaderLocation(buf);
	} else
		cgiHeaderLocation("view");

	return 0;
}

int
aucFastMailstat(char *path, int *num, int *new)
{
	int statusfound, mail, oldmail;
	char string[1024];
	FILE *mailfp;

	if (!(mailfp = fopen(path, "r"))) {
		return -1;
	}
	statusfound = mail = oldmail = 0;
	while (fgets(string, 11, mailfp) != NULL) {
		if (strncmp(string, "From ", 5) == 0) {
			mail++;
			statusfound = 0;
		} else if (!statusfound && strncmp(string, "X-IMAP: ", 8) == 0) {
			mail--;
			statusfound = 1;
		} else if (!statusfound && strncmp(string, "Status: RO", 10) == 0) {
			oldmail++;
			statusfound = 1;
		}
		if (string[strlen(string) - 1] != '\n')
			while (fgetc(mailfp) != '\n') {
				if (feof(mailfp))
					break;
			}
	}
	*num = mail;
	*new = mail - oldmail;
	return 0;
}

int
aucEmailClMailstat(char *mailbox, int *num, int *new)
{
	int sock, n1, n2;
	char *args[] = {
		mailbox,
		NULL
	};
	if ((sock = aucSockConnect("aucemail", aucUserCache->uname)) == -1)
		return -1;

	aucSockSendCommand(sock, "aucemail", aucUserCache->authstring, EMAIL_MAILSTAT, args);
	n1 = aucSockRecvReply(sock);
	n2 = aucSockRecvReply(sock);
	close(sock);

	if ((n1 == -1) || (n2 == -1))
		return -1;
	else {
		*num = n1;
		*new = n2;
	}

	return 0;
}

void
aucEmailClPrintNodeFrame(FILE * output, aucNodelettInfoPtr nodelett, char *arg)
{
	char buf[1024];
	int num = 0, new = 0, mustconfigimap = 0;

	fprintf(output,
			"<table width=100%% height=%d cellspacing=0 border=0 cellpadding=4 bgcolor=#FFFFFF>\n",
			nodelett->height);
	fprintf(output, "<tr bgcolor=#000000><td height=25><font color=#FFFFFF><b>%s</b></font></td>\n",
			_("Inbox"));
	fprintf(output, "<td align=right><font color=#FFFFFF><b>\n");


	if (!aucEmailClMailstat("INBOX", &num, &new))
		fprintf(output, "%s: %d %s: %d", _("Mail"), num, _("New"), new);
	else {
		getmailopt(buf, sizeof(buf), "imap-server");
		if (strcmp(buf, _("None")))
			fprintf(output, "%s", _("Closed"));
		else if ((aucMailSourceType == LOCALSPOOL) && !aucUserCache->local) {
			fprintf(output, _("None Available"));
			mustconfigimap = 1;
		} else {
			getmailopt(buf, sizeof(buf), "inbox-path");
			if (aucFastMailstat(buf, &num, &new))
				fprintf(output, "%s", _("No Mailbox"));
			else
				fprintf(output, "%s: %d %s: %d", _("Mail"), num, _("New"), new);
		}
	}

	aucEmailClParseOptions();

	fprintf(output, "</b></font></td></tr><tr valign=top><td colspan=2 height=%d>\n",
			nodelett->height - 25);
	if (!mustconfigimap) {
		fprintf(output, "<a href=\"email/view\">%s</a><br>\n", _("INBOX"));
		if (new)
			fprintf(output, "<b><a href=\"email/firstnew\">%s</a></b><br>\n",
					_("Goto First New Message"));
		fprintf(output, "<a %s href=\"email/compose\">%s</a><br>\n",
				user.newwindow ? "target=\"_blank\"" : "", _("Compose Message"));
		fprintf(output, "<a href=\"email/folders\">%s</a><br>\n", _("Folder List"));
	}
	fprintf(output, "<a href=\"email/options\">%s</a><br>\n", _("Preferences"));

	if (mustconfigimap) {
		aucGetFullPath("config-imap-mess.html", aucDataDir, buf, sizeof(buf));
		aucHtmlEchoFile(output, buf, 0);
	}

	fprintf(output, "</td></tr></table>\n");

}

/*
int aucQNNumMesgs()
{
   return (mailinfo.num);
}

int aucQNNumNewMesgs()
{
   if (mailinfo.new != -1)
      return (mailinfo.new);
   else
      return (0);
}
*/

int
aucEmailClSetAnswered(FILE * outfile, unsigned long uid)
{
	char *args[2];
	int sock, retval;
	char buf[16];

	snprintf(buf, sizeof(buf), "%lu", uid);
	args[0] = buf;
	args[1] = NULL;

	if ((sock = aucSockConnect("aucemail", aucUserCache->uname)) == -1)
		return -1;

	aucSockSendCommand(sock, "aucemail", aucUserCache->authstring, EMAIL_SET_ANSWERED, args);
	retval = aucSockRecvReply(sock);
	close(sock);

	return retval;
}

int
aucEmailClRemoveFolder(FILE * output, char *folder)
{
	char *args[2];
	int sock, retval;

	args[0] = folder;
	args[1] = NULL;

	if ((sock = aucSockConnect("aucemail", aucUserCache->uname)) == -1)
		return -1;

	aucSockSendCommand(sock, "aucemail", aucUserCache->authstring, EMAIL_REMOVE_FOLDER, args);
	retval = aucSockRecvReply(sock);
	close(sock);

	return retval;
}

int
aucEmailClNewFolder(FILE * output, char *folder)
{
	char *args[2];
	int sock, retval;

	args[0] = folder;
	args[1] = NULL;

	if ((sock = aucSockConnect("aucemail", aucUserCache->uname)) == -1)
		return -1;

	aucSockSendCommand(sock, "aucemail", aucUserCache->authstring, EMAIL_NEW_FOLDER, args);
	retval = aucSockRecvReply(sock);
	close(sock);

	return retval;
}

int
aucEmailClEmptyTrash(FILE * output)
{
	int sock, retval;

	if ((sock = aucSockConnect("aucemail", aucUserCache->uname)) == -1)
		return -1;

	aucSockSendCommand(sock, "aucemail", aucUserCache->authstring, EMAIL_EMPTY_TRASH, NULL);
	retval = aucSockRecvReply(sock);
	close(sock);

	return retval;
}

int
aucEmailClQuickInject(FILE * output, char *from_head, char *from_in, char *to_in, char *cc_in,
					  char *subj_in, char *text_in)
{
	FILE *mailout;
	char buf[128];

	snprintf(buf, sizeof(buf), "/usr/lib/sendmail -t -f%s", from_head);
	if (!(mailout = popen(buf, "w"))) {
		fprintf(output, "<b>%s:</b> %s /usr/lib/sendmail", _("Error"), _("Could not execute"));
		return -1;
	}

	fprintf(mailout, "From: %s\n", from_in);
	fprintf(mailout, "To: %s\n", to_in);
	if (cc_in && strlen(cc_in))
		fprintf(mailout, "CC: %s\n", cc_in);
	fprintf(mailout, "Subject: %s\n", subj_in);
	fprintf(mailout, "\n");
	fprintf(mailout, "%s\n", text_in);
	pclose(mailout);
	fprintf(output, "<i>%s</i>\n", _("Success"));
	return 0;
}

int
aucEmailClInject(FILE * output, char *to_in, char *cc_in, char *subj_in, char *text_in,
				 char *inreplyto_in, unsigned long func_flags)
{
	char *args[] = {
		to_in,
		cc_in,
		subj_in,
		text_in,
		inreplyto_in,
		NULL,
		NULL
	};
	char buf[16];
	int sock;

	snprintf(buf, sizeof(buf), "%lu", func_flags);
	args[5] = buf;

	if (!user.usingimap && (aucMailSourceType == LOCALSPOOL) && !aucUserCache->local) {
		fprintf(output, "<p><big><b>%s:</b> %s <a href=\"options\">%s</a></big></p>\n\n",
				_("Error"),
				_("Email functions cannot be used unless an IMAP server is configured in your"),
				_("preferences"));
		return -1;
	}

	if ((sock = aucSockConnect("aucemail", aucUserCache->uname)) == -1) {
		fprintf(output, "<p><big><b>%s:</b> %s: %s</big></p>\n\n",
				_("Error"), _("Could not connect to mail helper agent"), strerror(errno));
		return -1;
	}

	aucSockSendCommand(sock, "aucemail", aucUserCache->authstring, EMAIL_INJECT, args);
	aucSockRecvEcho(sock, output);
	close(sock);

	return 0;
}

int
aucEmailClCompose(FILE * output, char *to, long messid, int func_flags)
{
	char *args[4];
	char buf[16];
	int sock;

	if (!user.usingimap && (aucMailSourceType == LOCALSPOOL) && !aucUserCache->local) {
		fprintf(output, "<p><big><b>%s:</b> %s <a href=\"options\">%s</a></big></p>\n\n",
				_("Error"),
				_("Email functions cannot be used unless an IMAP server is configured in your"),
				_("preferences"));
		return -1;
	}

	if ((sock = aucSockConnect("aucemail", aucUserCache->uname)) == -1) {
		fprintf(output, "<p><big><b>%s:</b> %s: %s</big></p>\n",
				_("Error"), _("Could not connect to mail helper agent"), strerror(errno));
		return -1;
	}

	args[0] = to;
	snprintf(buf, sizeof(buf), "%ld", messid);
	args[1] = strdup(buf);
	snprintf(buf, sizeof(buf), "%d", func_flags);
	args[2] = strdup(buf);
	args[3] = NULL;

	aucSockSendCommand(sock, "aucemail", aucUserCache->authstring, EMAIL_COMPOSE, args);
	aucSockRecvEcho(sock, output);
	close(sock);

	free(args[1]);
	free(args[2]);

	return 0;

}

int
aucEmailClPrintMessage(FILE * output, unsigned long uid, unsigned long func_flags)
{
	char *args[4];
	char buf[16];
	int sock;

	if ((sock = aucSockConnect("aucemail", aucUserCache->uname)) == -1) {
		fprintf(output, "<p><big><b>%s:</b> %s: %s</big></p>\n",
				_("Error"), _("Could not connect to mail helper agent"), strerror(errno));
		return -1;
	}

	snprintf(buf, sizeof(buf), "%lu", uid);
	args[0] = strdup(buf);
	snprintf(buf, sizeof(buf), "%lu", func_flags);
	args[1] = strdup(buf);
	args[2] = NULL;

	aucSockSendCommand(sock, "aucemail", aucUserCache->authstring, EMAIL_PRINT_MESSAGE, args);
	aucSockRecvEcho(sock, output);
	close(sock);

	free(args[0]);
	free(args[1]);

	return 0;
}

int
aucEmailClHeader(FILE * output, int pageno, int func_flags)
{
	char *args[2];
	char buf[16];
	int sock;

//   aucEmailClParseOptions();

	if ((sock = aucSockConnect("aucemail", aucUserCache->uname)) == -1) {
		fprintf(output, "<p><big><b>%s:</b> %s: %s</big></p>\n",
				_("Error"), _("Could not connect to mail helper agent"), strerror(errno));
		return -1;
	}

	snprintf(buf, sizeof(buf), "%d", func_flags);
	args[0] = strdup(buf);
	snprintf(buf, sizeof(buf), "%d", pageno);
	args[1] = strdup(buf);
	args[2] = NULL;

	aucSockSendCommand(sock, "aucemail", aucUserCache->authstring, EMAIL_HEADER, args);
	aucSockRecvEcho(sock, output);
	close(sock);

	free(args[0]);
	free(args[1]);

	return 0;
}

int
aucEmailClFooter(FILE * output, int func_flags)
{
	char *args[2];
	char buf[16];
	int sock;

	if ((sock = aucSockConnect("aucemail", aucUserCache->uname)) == -1) {
		fprintf(output, "<p><big><b>%s:</b> %s: %s</big></p>\n",
				_("Error"), _("Could not connect to mail helper agent"), strerror(errno));
		return -1;
	}

	snprintf(buf, sizeof(buf), "%d", func_flags);
	args[0] = buf;
	args[1] = NULL;

	aucSockSendCommand(sock, "aucemail", aucUserCache->authstring, EMAIL_FOOTER, args);
	aucSockRecvEcho(sock, output);
	close(sock);

	return 0;
}

int
aucEmailClViewFolders(FILE * output)
{
	int sock;

	if (!user.usingimap && (aucMailSourceType == LOCALSPOOL) && !aucUserCache->local) {
		fprintf(output, "<p><big><b>%s:</b> %s <a href=\"options\">%s</a></big></p>\n\n",
				_("Error"),
				_("Email functions cannot be used unless an IMAP server is configured in your"),
				_("preferences"));
		return -1;
	}

	if ((sock = aucSockConnect("aucemail", aucUserCache->uname)) == -1) {
		fprintf(output, "<p><big><b>%s:</b> %s: %s</big></p>\n",
				_("Error"), _("Could not connect to mail helper agent"), strerror(errno));
		return -1;
	}

	aucSockSendCommand(sock, "aucemail", aucUserCache->authstring, EMAIL_VIEW_FOLDERS, NULL);
	aucSockRecvEcho(sock, output);
	close(sock);

	return 0;
}

int
aucEmailClPrintList(FILE * output, int page, int func_flags)
{
	char *args[4];
	char buf[16];
	int sock;

	if (!user.usingimap && (aucMailSourceType == LOCALSPOOL) && !aucUserCache->local) {
		fprintf(output, "<p><big><b>%s:</b> %s <a href=\"options\">%s</a></big></p>\n\n",
				_("Error"),
				_("Email functions cannot be used unless an IMAP server is configured in your"),
				_("preferences"));
		return -1;
	}

	if ((sock = aucSockConnect("aucemail", aucUserCache->uname)) == -1) {
		fprintf(output, "<p><big><b>%s:</b> %s: %s</big></p>\n",
				_("Error"), _("Could not connect to mail helper agent"), strerror(errno));
		return -1;
	}

	snprintf(buf, sizeof(buf), "%d", page);
	args[0] = strdup(buf);
	snprintf(buf, sizeof(buf), "%d", func_flags);
	args[1] = strdup(buf);
	args[2] = NULL;

	aucSockSendCommand(sock, "aucemail", aucUserCache->authstring, EMAIL_PRINT_LIST, args);
	aucSockRecvEcho(sock, output);
	close(sock);

	free(args[0]);
	free(args[1]);

	return 0;
}

void
dropcase(char *str)
{
	char *a;

	for (a = str; *a; a++)
		if ((*a <= 'Z') && (*a >= 'A'))
			*a -= 'A' - 'a';
}

void
aucEmailClGetOptionNames(char **options, int max)
{
	int x;
	char **a;

	for (x = 0, a = options; prefs[x].setting; x++, a++)
		*a = prefs[x].setting;
	*a = prefs[x].setting;
}

void
aucEmailClFreePinerc(struct pinerckey *pinerc)
{
	struct pinerckey *a;

	for (a = pinerc; a->setting; a++)
		if (a->custom)
			free(a->value);
}

int
aucEmailClWritePinerc(struct pinerckey *pinerc)
{
	char tmpfile[256], pinercfile[256], buf[256];
	FILE *rcfile, *newrc;
	struct pinerckey *a;
	int matched, fd;

/*	snprintf(buf, sizeof(buf), "auc-%sXXXXXX", aucUserCache->uname);
	if (!mktemp(buf))
		return 0; */
	snprintf(tmpfile, sizeof(tmpfile), "%s/auc-%sXXXXXX", aucUserCache->homedir, aucUserCache->uname);
	fd = mkstemp(tmpfile);
	if (!(newrc = fdopen(fd, "w"))) {
		return 0;
	}
	snprintf(pinercfile, sizeof(pinercfile), "%s/%s", aucUserCache->homedir, aucMailPineUserConfig);
	if ((rcfile = fopen(pinercfile, "r"))) {
		while (fgets(buf, sizeof(buf), rcfile)) {
			matched = 0;
			if ((*buf == '\n') || (*buf == ' ') || (*buf == '#') || (*buf == '\t')) {
				fputs(buf, newrc);
				continue;
			}
			for (a = pinerc; a->setting; a++)
				if (!(a->set) && !strncmp(a->setting, buf, strlen(a->setting))) {
					matched = 1;
					if (a->custom)
						fprintf(newrc, "%s=%s\n", a->setting, a->value);
					else
						fprintf(newrc, "%s=\n", a->setting);
					a->set = 1;
					break;
				}
			if (!matched)
				fputs(buf, newrc);
		}
		fclose(rcfile);
	}

	for (a = pinerc; a->setting; a++)
		if (!(a->set)) {
			if (a->custom)
				fprintf(newrc, "\n%s=%s\n", a->setting, a->value);
			else
				fprintf(newrc, "\n%s=\n", a->setting);
			a->set = 1;
		}

	fclose(newrc);

	unlink(pinercfile);
	link(tmpfile, pinercfile);
	unlink(tmpfile);
	return 1;
}

int
aucEmailClWriteSigFile(char *sig)
{
	FILE *sigfile;
	char path[256];

	getmailopt(path, sizeof(path), "signature-file");
	if ((sigfile = fopen(path, "w"))) {
		fputs(sig, sigfile);
		fclose(sigfile);
	} else
		return 0;

	return 1;
}

int
aucEmailClWriteForwardFile(char *email)
{
	FILE *outfile;
	char path[256];

	snprintf(path, sizeof(path), "%s/.forward", aucUserCache->homedir);
	if (!strlen(email))
		unlink(path);
	else {
		if ((outfile = fopen(path, "w"))) {
			fprintf(outfile, "%s\n", email);
			fclose(outfile);
		}
	}
	snprintf(path, sizeof(path), "%s/.qmailrc", aucUserCache->homedir);
	if (!strlen(email))
		unlink(path);
	else {
		if ((outfile = fopen(path, "w"))) {
			fprintf(outfile, "%s\n", email);
			fclose(outfile);
		}
	}

	return 1;
}

int
aucEmailClShowOptions(FILE * output, int level)
{
	int x, y, cust;
	char defaultopt[256], customopt[256], realopt[256];

	fprintf(output, "<form action=\"options\" method=post>\n");
	fprintf(output, "<table border=0>\n");
	if (level == 1)
		fprintf(output, "<input type=hidden name=type value=basic>\n");

	for (x = 0; prefs[x].setting; x++) {
		if (prefs[x].level != level)
			continue;
		if ((aucMailSourceType == IMAPSPOOL) && !strncmp(prefs[x].setting, "imap", 4))
			continue;
		if (!aucMailAdvancedPrefs && !aucAuthPermsof("sysop") && (prefs[x].type == PR_CGI))
			continue;
		fprintf(output,
				"<tr valign=top><td width=40%%><b>%s:</b><br><font size=-1>%s</a></td><td width=60%%>\n",
				_(prefs[x].caption), _(prefs[x].desc));
		switch (prefs[x].type) {
		case PR_PINE:
			getdefaultopt(defaultopt, sizeof(defaultopt), prefs[x].setting);
			getcustomopt(customopt, sizeof(customopt), prefs[x].setting);
			cust = 0;
			if (strlen(customopt))
				cust = 1;
			fprintf(output, "<input type=radio name=\"%s-custom\" value=no %s> %s<br>\n",
					prefs[x].setting, !cust ? "Checked" : "", defaultopt);
			fprintf(output, "<input type=radio name=\"%s-custom\" value=yes %s>\n",
					prefs[x].setting, cust ? "Checked" : "");
			fprintf(output, "<input type=text name=\"%s-val\" value=\"%s\" size=30>\n",
					prefs[x].setting, customopt);
			break;
		case PR_YESNO:
			getdefaultopt(defaultopt, sizeof(defaultopt), prefs[x].setting);
			getcustomopt(customopt, sizeof(customopt), prefs[x].setting);
			getmailopt(realopt, sizeof(realopt), prefs[x].setting);
			dropcase(defaultopt);
			dropcase(customopt);
			dropcase(realopt);
			cust = 0;
			if (strlen(customopt))
				cust = 1;
			fprintf(output, "<input type=radio name=\"%s-custom\" value=%s %s>%s<br>\n",
					prefs[x].setting, (*defaultopt == 'n') ? "yes" : "no",
					(*realopt == 'y') ? "Checked" : "", _("Yes"));
			fprintf(output, "<input type=radio name=\"%s-custom\" value=%s %s>%s\n",
					prefs[x].setting, (*defaultopt == 'n') ? "no" : "yes",
					(*realopt == 'n') ? "Checked" : "", _("No"));
			fprintf(output, "<input type=hidden name=\"%s-val\" value=\"%s\">\n",
					prefs[x].setting, (*defaultopt == 'n') ? "yes" : "no");
			break;
		case PR_SIG:
			fprintf(output, "<textarea name=sigtext cols=50 rows=4>");
			getmailopt(realopt, sizeof(realopt), "signature-file");
			aucHtmlEchoFile(output, realopt, 1);
			fprintf(output, "</textarea>\n");
			break;
		case PR_FORWARD:
			fprintf(output, "<input type=text name=forwardemail value=\"");
			snprintf(realopt, sizeof(realopt), "%s/.forward", aucUserCache->homedir);
			aucHtmlEchoFile(output, realopt, 1);
			fprintf(output, "\" size=30>\n");
			break;
		case PR_CGI:
			fprintf(output, "<input type=submit name=type value=%s>\n", _(prefs[x].value));
			break;
		case PR_IMAPAUTH:
			getcustomopt(customopt, sizeof(customopt), "imap-name");
			fprintf(output, "%s:\n", _("Username"));
			fprintf(output, "<input type=hidden name=\"imap-name-custom\" value=yes>\n");
			fprintf(output, "<input type=text name=\"imap-name-val\" value=\"%s\" size=16>\n",
					customopt);
			fprintf(output, "<br>\n");
			getcustomopt(customopt, sizeof(customopt), "imap-passwd");
			fprintf(output, "%s:\n", _("Password"));
			fprintf(output, "<input type=hidden name=\"imap-passwd-custom\" value=yes>\n");
			fprintf(output, "<input type=password name=\"imap-passwd-val\" value=\"%s\" size=16>\n",
					customopt);
			fprintf(output, "<br>\n");
			break;
		case PR_SIZE:
			getmailopt(realopt, sizeof(realopt), prefs[x].setting);
			fprintf(output, "<input type=hidden name=%s-custom value=yes>\n", prefs[x].setting);
			fprintf(output, "<select name=%s-val>\n", prefs[x].setting);
			for (y = 0; y < 60; y += 10)
				if (y == 0)
					fprintf(output, "<option value=0 %s>%s</option>\n",
							(atoi(realopt) == y) ? "selected" : "", _("No Limit"));
				else
					fprintf(output, "<option %s>%d</option>\n",
							(atoi(realopt) == y) ? "selected" : "", y);
			fprintf(output, "</select>\n\n");
			break;
		}
		fprintf(output, "</td></tr>\n");
		fprintf(output, "<tr><td colspan=2><hr></td></tr>\n");
	}

	fprintf(output, "</table>\n");
	fprintf(output, "<input type=submit value=\"%s\"></form>\n", _("Save Preferences"));
	return (1);
}


void
getdefaultopt(char *str, size_t len, char *setting)
{
	int x;
	char buffer[256], *buf, *a;
	struct hostent *host;

	for (x = 0; prefs[x].setting && strcmp(prefs[x].setting, setting); x++);
	if (prefs[x].value)
		snprintf(str, len, "%s", _(prefs[x].value));
	else {
		if (!strcmp(setting, "personal-name"))
			snprintf(str, len, "%s", aucUserCache->fullname);
		else if (!strcmp(setting, "imap-name"))
			snprintf(str, len, "%s", aucUserCache->uname);
		else if (!strcmp(setting, "imap-passwd"))
			snprintf(str, len, "%s", aucUserCache->passwd);
		else if (!strcmp(setting, "imap-server")) {
			snprintf(str, len, "None");
			if ((aucMailSourceType == IMAPSPOOL)
				&& !aucConfToken("child::Email/IMAPServer", &buf)) {
				snprintf(str, len, "%s", buf);
				free(buf);
			}
		} else if (!strcmp(setting, "inbox-path")) {
			snprintf(str, len, "/var/spool/mail/%s", aucUserCache->uname);
			if (!aucConfToken("child::Email/SpoolPath", &buf)) {
				while (1) {
					int len, tok, rpl;
					char *rplstr, *tokstr;
					len = strlen(buf);
					if ((a = strstr(buf, "$username"))) {
						tokstr = "$username";
						tok = strlen("$username");
						rpl = strlen(aucUserCache->uname);
						rplstr = aucUserCache->uname;
					} else if ((a = strstr(buf, "$homedir"))) {
						tokstr = "$homedir";
						tok = strlen("$homedir");
						rpl = strlen(aucUserCache->homedir);
						rplstr = aucUserCache->homedir;
					} else
						break;
					if (tok < rpl)
						buf = realloc(buf, len + rpl - tok + 1);
					a = strstr(buf, tokstr);
					memmove(a + rpl, a + tok, len - (a - buf) - tok + 1);
					memcpy(a, rplstr, rpl);
				}
				snprintf(str, len, "%s", buf);
				free(buf);
			}
		} else if (!strcmp(setting, "folder-collections")) {
			snprintf(str, len, "mail/");
			if (!aucConfToken("child::Email/FolderPath", &buf)) {
				snprintf(str, len, "%s", buf);
				free(buf);
			}
		} else if (!strcmp(setting, "user-domain")) {
			char hostname[128];
			getmailopt(hostname, sizeof(hostname), "imap-server");
			if (strcmp(hostname, _("None")) && (aucMailSourceType != IMAPSPOOL))
				snprintf(str, len, "%s", hostname);
			else if (!aucConfToken("child::Email/AddressDomain", &buf) && strlen(buf)) {
				snprintf(str, len, "%s", buf);
				free(buf);
			} else {
				gethostname(buffer, sizeof(buffer));
				host = gethostbyname(buffer);
				if (host)
					strncpy(str, host->h_name, len);
				else
					strncpy(str, buffer, len);
			}
		}
	}

}

void
getmailopt(char *str, size_t len, char *setting)
{
	int x;
	char buffer[len];

	getcustomopt(str, len, setting);
	if (!strlen(str))
		getdefaultopt(str, len, setting);
	for (x = 0; prefs[x].setting && strcmp(prefs[x].setting, setting); x++);
	if (prefs[x].path && *str != '/' && *str != '{') {
		strcpy(buffer, str);
		snprintf(str, len, "%s/%s", aucUserCache->homedir, buffer);
	}
}

int
getpinekey(char *str, size_t len, char *setting, char *filename)
{
	FILE *pinerc;
	char buf[1024], *a;
	int found = 0;

	if ((pinerc = fopen(filename, "r"))) {
		while (!found) {
			if (!fgets(buf, sizeof(buf), pinerc)) {
				fclose(pinerc);
				return (0);
			}
			if (!strncmp(buf, setting, strlen(setting)))
				found = 1;
		}
		fclose(pinerc);
		a = buf + strlen(setting);
		while (*a != '=')
			a++;
		a++;
		while (*a == ' ' || *a == '\t')
			a++;
		if (*a == '\n')
			return (0);
		snprintf(str, len, "%s", a);
		if (str[strlen(str) - 1] == '\n')
			str[strlen(str) - 1] = '\0';
		return (1);
	} else
		return (0);
}

void
getcustomopt(char *str, size_t len, char *setting)
{
	char buffer[512];

	snprintf(buffer, sizeof(buffer), "%s/%s", aucUserCache->homedir, aucMailPineUserConfig);
	if (!getpinekey(str, len, setting, buffer))
		strncpy(str, "", len);
}
