/*
 * Authenticated User Community
 * Copyright (c) 2000 David C. Moore
 *
 * You may distribute and/or use for any purpose modified or unmodified
 * copies of this software if you preserve the copyright notice above.
 */
 
/** 
 * @file files.c
 *
 * Routines comprising the web-based file manager of AUC
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>

#include <sys/stat.h>
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <dirent.h>

#ifdef HAVE_SYS_QUOTA_H
#include <sys/quota.h>
#endif

#include <libxml/tree.h>
#include <libxml/parser.h>
#include <libxml/xpath.h>

#include <auc/cgiupl.h>
#include <auc/cgic.h>
#include <auc/auc.h>
#include <auc/conf.h>
#include <auc/database.h>
#include <auc/error.h>
#include <auc/pages.h>
#include <auc/html.h>
#include <auc/mime.h>
#include <auc/forms.h>

#include "files.h"

struct virtualfs
{
	char *root;
	int cutoff;
	char *virtualmatch;
};

struct virtualfs *virtuals[32];
struct virtualfs **lastvirtual;
/* char *aucWordViewPath = "";
char *aucWordViewCharset = "iso-8859-1"; */
int aucWordViewInline = 1;
unsigned long aucNonLocalDiskLimit = 4000000;


#define GET_PROPFILE(str, path)	snprintf(str, sizeof(str), "%s%s", path, ".aucfolder")


char *filetypecaption(int);
char *filetypeimage(int);
int getfiletype(char *);
char *get_filename(char *);
int displayfilecontents(FILE *, char *, int);
int printtextfile(FILE *, char *);
int printtextfileforedit(FILE *, char *);
int getrealpath(char *, size_t, char *, char *);
int printfileops(FILE *, char *, unsigned long, int);
int writable(char *, uid_t, gid_t, int);
int fileerror(char *);
int fs_quota_exceeded(uid_t, char *);
char *getdevname(char *);
unsigned long spaceused(char *);
int matchvirtuals(char *, size_t, char *, char *);
int aucGetFileStats(char *path, char *fsize, char *fdate, int *writable, uid_t uid, gid_t gid);
int viewworddoc(FILE * output, char *path);
int defaultfileview(FILE * output, char *path);
int compare(void *, void *);

char * get_parent_path(char * path);
struct file_parameters_type * get_file_parameters(char * virtualfs,
		char * dir, char * topdirname, int expert);
void free_file_parameters(struct file_parameters_type * par);
int dropbox_index(FILE * output, char *path, int expert,
		int iswritable, aucFolderPtr folder_props);


char Pathbuf[2048];

char *typeimages[] = {
	"filetypes/type_default.gif",
	"filetypes/type_default.gif",
	"filetypes/type_dir.gif",
	"filetypes/type_doc.gif",
	"filetypes/type_text.gif",
	"filetypes/type_doc.gif"
};


struct file_parameters_type {
	char * name, * filename;
	char * path, * parent_path;
	char * parent_url, * current_url;
	char * parent_name;
	unsigned int ftype;
	aucFolderPtr props;
};

struct context_info_type {
	int requested_perm;
	int expert;
	int iswritable;
	shorthead_fn shorthead;
	shortfoot_fn shortfoot;
};

int action_save (FILE *, struct file_parameters_type *, struct context_info_type *, char *);
int action_newfile (FILE *, struct file_parameters_type *, struct context_info_type *, char *);
int action_delete (FILE *, struct file_parameters_type *, struct context_info_type *, char *);
int action_rename (FILE *, struct file_parameters_type *, struct context_info_type *, char *);
int action_props (FILE *, struct file_parameters_type *, struct context_info_type *, char *);
int action_upload (FILE *, struct file_parameters_type *, struct context_info_type *, char *);
int action_download (FILE *, struct file_parameters_type *, struct context_info_type *, char *);
int action_view (FILE *, struct file_parameters_type *, struct context_info_type *, char *);
int action_error (FILE *, struct file_parameters_type *, struct context_info_type *, char *);


typedef int (*file_action_func) (FILE *, struct file_parameters_type *, struct context_info_type *, char *);

struct file_action_type {
	char * name;
	file_action_func action;
	unsigned int typemask;
	unsigned int flags;
};

struct file_action_type normal_actions[] = {
	{ "newfile",	action_newfile,		FMASK_DIR,	IS_WRITABLE	},
	{ "save",		action_save,		FMASK_TEXTDOC | FMASK_HTMLDOC,	IS_WRITABLE | IS_POST },
	{ "delete",		action_delete,		FMASK_ANY,	IS_WRITABLE	},
	{ "rename",		action_rename,		FMASK_ANY,	IS_WRITABLE },
	{ "props",		action_props,		FMASK_DIR,	IS_WRITABLE },
	{ "upload",		action_upload,		FMASK_DIR,	IS_WRITABLE },
	{ "download",	action_download,	~FMASK_DIR,	0	},
	{ "edit",		action_view,		FMASK_TEXTDOC | FMASK_HTMLDOC,	IS_WRITABLE	},
	{ "view",		action_view,		FMASK_ANY,	0	},
	{ NULL }
};

struct file_action_type dropbox_actions[] = {
	{ "delete",		action_delete,		FMASK_ANY,	IS_WRITABLE	},
	{ "rename",		action_rename,		FMASK_ANY,	IS_WRITABLE },
	{ "props",		action_props,		FMASK_DIR,	IS_WRITABLE },
	{ "upload",		action_upload,		FMASK_DIR,	0	},
	{ "download",	action_download,	~FMASK_DIR,	IS_WRITABLE	},
	{ "view",		action_view,		FMASK_ANY,	IS_WRITABLE	},
	{ "view",		action_view,		FMASK_DIR,	0	},
	{ NULL }
};

struct file_action_type * folder_actions[] = {
	normal_actions, dropbox_actions
};



int
aucFilesInit(void)
{
	char *buf;

/*	aucConfToken("child::FileManager/WordViewPath", &aucWordViewPath);
	aucConfToken("child::FileManager/WordViewCharset", &aucWordViewCharset); */

	if (!aucConfToken("child::FileManager/WordViewInline", &buf))
		if (!strcasecmp(buf, "no"))
			aucWordViewInline = 0;

	if (!aucConfToken("child::FileManager/DefaultDiskQuota", &buf))
		aucNonLocalDiskLimit = atol(buf) * 1024;

	aucNodelettRegister("files",
						_("File Storage"),
						aucFilesNodeFrame, 1, 80, 50,
						_("Volumes where you store files."), "normal", 0);

	aucPagesRegister(NULL, "upload-file", PAGE_FILE, aucFileBrowserH, "normal", 0);
	aucPagesRegister(NULL, "homedir", PAGE_DIRFUNC, aucFileBrowserH, "normal", 0);

	aucDBRegister("files", "CREATE TABLE %s (path TEXT NOT NULL, \
         action VARCHAR(32) NOT NULL DEFAULT 'add', \
         contents TEXT NOT NULL)");

	*virtuals = NULL;
	lastvirtual = virtuals;

	return 0;
}

int
aucExceededQuota()
{
	if (aucUserCache->local)
		return fs_quota_exceeded(aucUserCache->uid, getdevname(aucUserCache->homedir));
	else if (strlen(aucUserCache->homedir))
		return (spaceused(aucUserCache->homedir) > aucNonLocalDiskLimit);
	else
		return 0;
}

char *
getdevname(char *path)
{
	struct stat statbuf;
	int pathmaj, pathmin, found = 0;
	char buf[1024], device[32], mntpt[64];
	FILE *mtab;

	if (stat(path, &statbuf))
		return NULL;
	pathmaj = (int) (statbuf.st_dev >> 8);
	pathmin = (int) (statbuf.st_dev & 0x00FF);
	if ((mtab = fopen("/etc/mtab", "r"))) {
		while (fgets(buf, sizeof(buf), mtab)) {
			sscanf(buf, "%s %s", device, mntpt);
			if (!stat(mntpt, &statbuf)) {
				if ((pathmaj == (int) (statbuf.st_dev >> 8)) &&
					(pathmin == (int) (statbuf.st_dev & 0x00FF))) {
					found = 1;
					break;
				}
			}
		}
		fclose(mtab);
	}
	if (found)
		return (char *) strdup(device);
	else
		return NULL;

}

unsigned long
spaceused(char *path)
{
	DIR *curdir;
	struct dirent *filebuf;
	struct stat statbuf;
	char buf[1024];
	unsigned long total = 0;

	if (!lstat(path, &statbuf))
		total += statbuf.st_size;

	if ((curdir = opendir(path))) {
		while ((filebuf = readdir(curdir)))
			if (strcmp(filebuf->d_name, ".") && strcmp(filebuf->d_name, "..")) {
				snprintf(buf, sizeof(buf), "%s/%s", path, filebuf->d_name);
				total += spaceused(buf);
			}
		closedir(curdir);
	}

	return total;
}

int
fs_quota_exceeded(uid_t uid, char *device)
{
#ifdef HAVE_SYS_QUOTA_H
	struct dqblk dqbuf;

	if (!quotactl(QCMD(Q_GETQUOTA, USRQUOTA), device, uid, (caddr_t) & dqbuf)) {
		if ((dqbuf.dqb_curblocks >= dqbuf.dqb_bsoftlimit) && (dqbuf.dqb_bsoftlimit != 0)
			)
			return 1;
		if ((dqbuf.dqb_curinodes >= dqbuf.dqb_isoftlimit) && (dqbuf.dqb_isoftlimit != 0)
			)
			return 1;
	}
#endif

	return 0;
}

void
aucFilesNodeFrame(FILE * output, aucNodelettInfoPtr nodelett, char *arg)
{
	fprintf(output, "<table width=100%% height=%d cellpadding=0 cellspacing=0 border=0 bgcolor=#FFFFFF>\n",
			nodelett->height);
	fprintf(output, "<tr valign=top><td>\n");
	
	fprintf(output, "<table width=100%% cellpadding=2 cellspacing=0 border=0 bgcolor=#FFFFFF>\n");
	
	fprintf(output, "<tr valign=top><td bgcolor=#CCCCCC><small><b>%s</b></small></td></tr>\n",
		_("Private Areas"));	
		
	fprintf(output, "<tr valign=top><td>\n");
	fprintf(output,
			"<p>&nbsp;<img src=\"%s/fileops/folder.gif\" alt=\"\" border=0 align=absmiddle>\n",
			aucGraphDir);
	fprintf(output, "<b><a href=\"homedir/\">%s</a>/</b></p>\n\n", _("Home Directory"));

	if (aucExceededQuota())
		fprintf(output, "<font color=\"red\"><b>%s</b></font>\n\n", _("Quota Exceeded"));
	
	fprintf(output, "</td></tr>\n");
	
/*
	fprintf(output, "<tr valign=top><td bgcolor=#CCCCCC><small><b>%s</b></small></td></tr>\n",
		_("Shared Folders"));	
	fprintf(output, "<tr valign=top><td><p><i>%s</i></p></td></tr>\n", _("none"));
	
	fprintf(output, "<tr valign=top><td bgcolor=#CCCCCC><small><b>%s</b></small></td></tr>\n",
		_("Dropboxes"));
	fprintf(output, "<tr valign=top><td><p><i>%s</i></p></td></tr>\n", _("none"));
*/

	fprintf(output, "</table></td></tr></table>\n");

}

int
copy_uploaded_files(char * path, int requested_perm, char * fileprefix)
{
	char file[256], dirstr[256];
	FILE *tmp, *new;
	char copybuf[1024];
	int count, mask=077;
	struct stat finfo;

	if (cgiFormSuccess == cgiFormString("filename", file, sizeof(file))) {
		if (fileprefix)
			snprintf(dirstr, sizeof(dirstr), "%s%s%s", path, fileprefix, file);
		else
			snprintf(dirstr, sizeof(dirstr), "%s%s", path, file);

		if (!stat(dirstr, &finfo)) {
			unlink(cgiUploadTemp);
			fileerror("File already exists");
			return -1;
		}

		if (requested_perm == TREE_FORCEWRITE) {
			aucSuRoot();
			mask = umask(022);
		}

		if ((tmp = fopen(cgiUploadTemp, "r"))) {
			if ((new = fopen(dirstr, "w"))) {
				while ((count = fread(copybuf, 1, sizeof(copybuf), tmp)))
					fwrite(copybuf, 1, count, new);
				fclose(new);
			}
			fclose(tmp);
		}
		unlink(cgiUploadTemp);

		if (requested_perm == TREE_FORCEWRITE) {
			umask(mask);
			aucSuUser();
		}

		return 0;
	}
	return -1;

}

int
aucChownUploadedFiles(uid_t uid, gid_t gid)
{
	char file[256], dirstr[256];

	if (cgiFormSuccess == cgiFormString("filename", file, sizeof(file))) {
		snprintf(dirstr, sizeof(dirstr), "%s", cgiUploadTemp);
		chown(dirstr, uid, gid);
		return 1;
	}
	return 0;
}

int
fileerror(char *str)
{
	char buf[1024];
	snprintf(buf, sizeof(buf),
"<table width=100%% border=0 cellpadding=5><tr bgcolor=#CCCCCC><td> \
<big><b><font color=\"#FF0000\">%s:</font> %s</b></big> \
</td></tr></table>\n",
		_("Error"), str);
	return aucErrorAppendStr(buf);
}

int
aucFileOp(char *path, char *path2, unsigned long fop)
{
	char *fields[4];
	int x;
	FILE *outfile;
	struct stat fileinfo;

	for (x = 0; x < 4; x++)
		fields[x] = NULL;

	switch (fop) {
	case FOP_WRITE:
		if (path2) {
			fields[0] = (char *) malloc(sizeof(char) * (strlen(path2) + 1));
			strcpy(fields[0], path2);
		} else
			aucHtmlReadForm("text", fields, 0);
		if ((outfile = fopen(path, "w"))) {
			fwrite(fields[0], sizeof(char), strlen(fields[0]), outfile);
			fclose(outfile);
		} else {
			fileerror((char *) strerror(errno));
			free(fields[0]);
			return -1;
		}
		free(fields[0]);
		break;
	case FOP_MKDIR:
		if (mkdir(path, 0777)) {
			fileerror((char *) strerror(errno));
			return -1;
		}
		break;
	case FOP_DELETE:
		if (unlink(path)) {
			fileerror((char *) strerror(errno));
			return -1;
		}
		break;
	case FOP_RMDIR:
		if (rmdir(path)) {
			fileerror((char *) strerror(errno));
			return -1;
		}
		break;
	case FOP_RENAME:
		if (stat(path2, &fileinfo)) {
			if (rename(path, path2)) {
				fileerror((char *) strerror(errno));
				return -1;
			}
		} else {
			fileerror("File of that name already exists");
			return -1;
		}
		break;
	default:
		return -1;
		break;
	}
	return 0;
}

int headfoot = 0;

void
printhtmlheader(FILE * output, shorthead_fn shorthead, char *path, char *name)
{
	char buf[512];
	int nohead = 0;

	if (cgiFormSuccess == cgiFormString("nohead", buf, sizeof(buf)))
		if (!strcmp(buf, _("yes")))
			nohead = 1;
	if (!strcmp(cgiRequestMethod, "POST"))
		nohead = 1;

	if (!nohead) {
		headfoot = 1;
		cgiHeaderContentType("text/html");
		shorthead(output, path, name);
	}
}

void
printhtmlfooter(FILE * output, shortfoot_fn shortfoot, char *path, char *name)
{
	if (headfoot)
		shortfoot(output, path, name);

}

int
action_newfile (FILE * output, struct file_parameters_type * par,
		struct context_info_type * context, char * action)
{
	char ext[16], fname[256];
	char newpath[2048];
	unsigned long fop = FOP_WRITE;
	char buf[1024];
	int mask=077;

	if (!strcmp(cgiRequestMethod, "POST")) {
		strcpy(ext, "");
		cgiFormString("extension", ext, sizeof(ext));
		cgiFormString("filename", fname, sizeof(fname));
		if (!strcmp(ext, "directory")) {
			fop = FOP_MKDIR;
			strcpy(ext, "");
			if (aucText2Filename(fname))
				strcpy(fname, _("New_Directory"));
		} else if (aucText2Filename(fname))
			strcpy(fname, _("New_File"));

		snprintf(newpath, sizeof(newpath), "%s%s%s", par->path, fname, ext);
		
		/* we need root permissions if the current uid cannot write the file */
		if (context->requested_perm == TREE_FORCEWRITE) {
			aucSuRoot();
			mask = umask(022);
		}
		
		aucFileOp(newpath, NULL, fop);
		
		if (context->requested_perm == TREE_FORCEWRITE) {
			umask(mask);
			aucSuUser();
		}

		snprintf(buf, sizeof(buf), "%s%s", fname, ext);
		cgiHeaderLocation(buf);
		return 0;
	} else {
		printhtmlheader(output, context->shorthead, par->path, _("Add New File"));
		fprintf(output, "<h2>%s</h2>\n\n", par->name);

		if (aucExceededQuota()) {
			aucGetFullPath("quota-mess.html", aucDataDir, buf, sizeof(buf));
			aucHtmlEchoFile(output, buf, 0);
		} else
			aucNewFilePrompt(output, par->current_url, context->expert);

		printhtmlfooter(output, context->shortfoot, par->path, par->name);
		return 0;
	}
}

int
action_save (FILE * output, struct file_parameters_type * par,
		struct context_info_type * context, char * action)
{
	int mask=077;
	
	/* we need root permissions if the current uid cannot write the file */
	if (context->requested_perm == TREE_FORCEWRITE) {
		aucSuRoot();
		mask = umask(022);
	}

	aucFileOp(par->path, NULL, FOP_WRITE);

	if (context->requested_perm == TREE_FORCEWRITE) {
		umask(mask);
		aucSuUser();
	}

	cgiHeaderLocation(par->parent_url);
	return 0;
}

int
action_delete (FILE * output, struct file_parameters_type * par,
		struct context_info_type * context, char * action)
{
	char confirm[64], buf[256];
	int mask=077;

	strcpy(confirm, "");
	cgiFormString("confirm", confirm, sizeof(confirm));
	if (!strcmp(confirm, "yes")) {

		if (context->requested_perm == TREE_FORCEWRITE) {
			aucSuRoot();
			mask = umask(022);
		}
		
		if (par->ftype == FILE_DIR) {
		
			/* If the directory contains a XML properties file, delete it first. */
			if (par->props->doc) {
				GET_PROPFILE(buf, par->path);
				unlink(buf);
			}
			
			/* Do the rmdir, and put back any XML properties file if the rmdir fails. */
			if ((aucFileOp(par->path, NULL, FOP_RMDIR) < 0) && par->props->doc)
				xmlSaveFile(buf, par->props->doc);
				
		} else
			aucFileOp(par->path, NULL, FOP_DELETE);
		
		if (context->requested_perm == TREE_FORCEWRITE) {
			umask(mask);
			aucSuUser();
		}

		cgiHeaderLocation(par->parent_url);
		return 0;
	} else {
		printhtmlheader(output, context->shorthead, par->path, _("Confirmation of Delete"));
		fprintf(output, "<h2>%s</h2>\n\n", par->name);

		fprintf(output, "<p>%s \"%s\"?</p>\n\n", _("Are you sure you want to delete the file"), par->name);
		fprintf(output, "<ul><p><b><a href=\"%s?action=delete&confirm=yes\">%s</a></b></p>\n",
				par->current_url, _("Yes"));
		fprintf(output, "<p><b><a href=\"%s\">%s</a></b></p></ul>\n", par->parent_url, _("No"));
		printhtmlfooter(output, context->shortfoot, par->path, par->name);
		return 0;
	}
}

int
action_rename (FILE * output, struct file_parameters_type * par,
		struct context_info_type * context, char * action)
{
	char newname[1024], ext[64];
	char buf[1024];
	int mask = 077;

	if (!strcmp(cgiRequestMethod, "POST")) {
		strcpy(newname, "");
		strcpy(ext, "");
		cgiFormString("newname", newname, sizeof(newname));
		cgiFormString("extension", ext, sizeof(ext));
		if (!aucText2Filename(newname)) {
			snprintf(buf, sizeof(buf), "%s%s%s", par->parent_path, newname, ext);

			if (context->requested_perm == TREE_FORCEWRITE) {
				aucSuRoot();
				mask = umask(022);
			}
		
			aucFileOp(par->path, buf, FOP_RENAME);
			
			if (context->requested_perm == TREE_FORCEWRITE) {
				umask(mask);
				aucSuUser();
			}
		}
		cgiHeaderLocation(par->parent_url);
		return 0;
	} else {
		printhtmlheader(output, context->shorthead, par->path, _("Rename File"));
		fprintf(output, "<h2>%s</h2>\n\n", par->name);

		aucRenameFilePrompt(output, par->current_url, par->filename, par->ftype, context->expert);
		printhtmlfooter(output, context->shortfoot, par->path, par->name);
		return 0;
	}
}


int
save_propfile(char * propfile, xmlDocPtr xmldoc, int requested_perm)
{
	int mask=0;

	if (requested_perm == TREE_FORCEWRITE) {
		aucSuRoot();
		mask = umask(022);
	}
	xmlSaveFile(propfile, xmldoc);
	if (requested_perm == TREE_FORCEWRITE) {
		umask(mask);
		aucSuUser();
	}
	
	return 0;
}

int
action_props (FILE * output, struct file_parameters_type * par,
		struct context_info_type * context, char * action)
{
	xmlDocPtr xmldoc;
	char template[256], propfile[256];
	aucVarField vars[64];
	
	vars[0].name = "form_action";
	vars[0].value = par->current_url;
	vars[1].name = NULL;
	
	aucGetFullPath("folderprops.xml", aucDataDir, template, sizeof(template));

	/* read the properties file for the folder if it exists or create it if it doesn't */
	GET_PROPFILE(propfile, par->path);
	if (!(xmldoc = xmlParseFile(propfile)))
		xmldoc = xmlNewDoc("1.0");


	if (!strcmp(cgiRequestMethod, "GET")) {
		printhtmlheader(output, context->shorthead, par->path, _("Folder Properties"));
		fprintf(output, "<h2>%s %s</h2>\n\n", _("Properties of"), par->name);

		aucFormsXMLDisplayFile(output, template, vars, xmldoc);

		printhtmlfooter(output, context->shortfoot, par->path, par->name);
		return 0;
	} else {
		aucFormsXMLProcessFile(output, template, vars, &xmldoc);
		
		/* save the properties of the folder */
		save_propfile(propfile, xmldoc, context->requested_perm);
		
		cgiHeaderLocation(par->parent_url);
	}
	
	xmlFreeDoc(xmldoc);
	return 0;
}

int
action_error (FILE * output, struct file_parameters_type * par,
		struct context_info_type * context, char * action)
{
		printhtmlheader(output, context->shorthead, par->path, _("Error"));
		
		fprintf(output, "<p><big><b>%s</b>: %s</big></p>\n\n", _("Error"),
			_("The requested file does not exist, or you do not have permission to \
perform the requested action."));
		
		printhtmlfooter(output, context->shortfoot, par->path, par->name);
}

int
add_dropbox_file(char * name, char * user, time_t create_time, xmlDocPtr propsdoc)
{
	xmlNodePtr entry;
	char buf[16];
	
	entry = xmlNewChild(propsdoc->children, NULL, "file", NULL);
	xmlSetProp(entry, "name", name);
	xmlSetProp(entry, "uname", user);
	snprintf(buf, sizeof(buf), "%ld", create_time);
	xmlSetProp(entry, "time", buf);
	
	return 0;
}


int
action_upload (FILE * output, struct file_parameters_type * par,
		struct context_info_type * context, char * action)
{
	char buf[1024], prefix[128];

	if (!strcasecmp(cgiRequestMethod, "POST")) {
		if (cgiFormSuccess == cgiFormString("filename", buf, sizeof(buf))) {
		
			if (par->props->type == FOLDER_DROPBOX) {
			
				snprintf(prefix, sizeof(prefix), "%s_", aucUserCache->uname);
				if (!copy_uploaded_files(par->path, TREE_FORCEWRITE, prefix)) {
				
					add_dropbox_file(buf, aucUserCache->uname, aucTime, par->props->doc);
					
					GET_PROPFILE(buf, par->path);
					/* save the properties of the folder */
					save_propfile(buf, par->props->doc, TREE_FORCEWRITE);
				}
				
			} else
				copy_uploaded_files(par->path, context->requested_perm, NULL);
				
			cgiHeaderLocation(par->current_url);
		} else {
			snprintf(buf, sizeof(buf), "%s?action=upload", par->current_url);
			cgiHeaderLocation(buf);
		}

	} else {
		printhtmlheader(output, context->shorthead, par->path, _("Upload File"));

		fprintf(output, "<h2>%s</h2>\n\n", par->name);

		if (aucExceededQuota()) {
			aucGetFullPath("quota-mess.html", aucDataDir, buf, sizeof(buf));
			aucHtmlEchoFile(output, buf, 0);
		} else {

			if (par->props->type == FOLDER_DROPBOX)
				aucGetFullPath("upload-dropbox-mess.html", aucDataDir, buf, sizeof(buf));
			else
				aucGetFullPath("upload-mess.html", aucDataDir, buf, sizeof(buf));

			aucHtmlEchoFile(output, buf, 0);

			fprintf(output,
					"<form method=POST enctype=\"multipart/form-data\" action=\"%s.uploadfile\">\n",
					par->current_url);
			fprintf(output, "<p><b>%s:</b>\n", _("File to upload"));
			fprintf(output, "<input type=hidden name=action value=\"upload\">\n");
			fprintf(output, "<input type=file name=filename></p>\n");
			fprintf(output, "<input type=submit value=\"%s\">\n", _("Upload"));
			fprintf(output, "</form>\n\n");
		}

		printhtmlfooter(output, context->shortfoot, par->path, par->name);
	}
	return 0;
}

int
mimeview(FILE * output, char * path, char * mimetype)
{
	struct stat finfo;
	
	stat(path, &finfo);
	
	fprintf(output, "Content-type: %s\n", mimetype);
	fprintf(output, "Content-length: %ld\n\n", finfo.st_size);
	
	aucHtmlEchoFile(output, path, 1);
	return 0;
}


int
action_download (FILE * output, struct file_parameters_type * par,
		struct context_info_type * context, char * action)
{
	mimeview(output, par->path, "application/octet-stream");
	return 0;
}


int
action_view (FILE * output, struct file_parameters_type * par,
		struct context_info_type * context, char * action)
{
	int edit=0, badquota, shown;
	char * mimetype;
	
	/* if the file type is not a type we handle specially, we just output it
	 * using the appropriate mime type and let the browser handle it. */
	if (par->ftype == FILE_DEFAULT) {
		mimetype = aucMimeGetType(aucGetFileExt(par->filename), aucEtcRoot);
		if (!mimetype)
			mimetype = "text/plain";
		mimeview(output, par->path, mimetype);
		return 0;
	}
	
	
	/* since this file is of a special type (word, text, html, etc.) we wrap
	 * it inside the file manager interface. */
	
	printhtmlheader(output, context->shorthead, par->path, par->name);

	if (!strcmp(action, "edit"))
		edit = 1;

	if ((badquota = aucExceededQuota())) {
		char buf[1024];
		aucGetFullPath("quota-mess.html", aucDataDir, buf, sizeof(buf));
		aucHtmlEchoFile(output, buf, 0);
	}

	fprintf(output, "<p>\n");
	fprintf(output, "<table cellpadding=0 cellspacing=0 border=0 width=100%%>\n");
	fprintf(output, "<tr><td>\n");
	if (par->parent_url)
		fprintf(output, "<font size=-1><b><a href=\"%s\">%s %s</a></b></font><br>\n",
				par->parent_url, _("Back to"), par->parent_name);

	fprintf(output, "<table cellpadding=2 cellspacing=0 border=0><tr valign=top><td>\n");
	fprintf(output, "<b>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<font size=+2>%s</font></b>\n",
			par->name);
	fprintf(output, "</td><td>\n");
	fprintf(output, "<b>\n");
	if (par->parent_url && !edit)
		printfileops(output, par->current_url, par->ftype, context->iswritable);
	fprintf(output, "</b>\n");
	fprintf(output, "</td></tr></table>\n\n");
	
	fprintf(output, "</td><td>\n");
	if (!context->expert) {
		fprintf(output, "<table width=250 cellpadding=1 cellspacing=0 border=0><tr valign=top>\n");
		fprintf(output,
				"<td width=250 colspan=2 bgcolor=#000000><font size=-1 color=#FFFFFF><b>%s</b></font></td>\n",
				_("Legend: File Operations"));
		fprintf(output, "<tr valign=top>\n");
		fprintf(output,
				"<td width=125 bgcolor=#EEEEEE><img src=\"%s/fileops/edit.gif\" width=20 height=20>\n",
				aucGraphDir);
		fprintf(output, "<font size=-1>= %s</font></td>\n", _("Edit File"));
		fprintf(output,
				"<td width=125 bgcolor=#EEEEEE><img src=\"%s/fileops/download.gif\" width=20 height=20>\n",
				aucGraphDir);
		fprintf(output, "<font size=-1>= %s</font></td>\n", _("Download File"));
		fprintf(output,
				"<tr valign=top><td width=175 bgcolor=#EEEEEE><img src=\"%s/fileops/trash.gif\" width=20 height=20>\n",
				aucGraphDir);
		fprintf(output, "<font size=-1>= %s</font></td>\n", _("Delete File"));
		fprintf(output,
				"<td width=125 bgcolor=#EEEEEE><img src=\"%s/fileops/rename.gif\" width=20 height=20>\n",
				aucGraphDir);
		fprintf(output, "<font size=-1>= %s</font></td></tr>\n", _("Rename File"));
		fprintf(output, "</table>\n\n");
	}

	if (!context->expert)
		fprintf(output, "</td></tr></table></p>\n\n");

	if ((par->ftype == FILE_DIR) && par->props->description && strlen(par->props->description))
		fprintf(output, "<dl><dt><small><u>%s</u>:</small></dt><dd>%s</dd></dl>\n\n",
		   _("Folder Description"), par->props->description);

	if (context->expert)
		fprintf(output, "</td></tr></table></p>\n\n");

	if (edit) {
		fprintf(output, "<form method=post action=\"%s\">\n", par->filename);
		fprintf(output, "<input type=hidden name=action value=save>\n");
		fprintf(output, "<textarea name=text rows=15 cols=70 wrap=soft>");
		printtextfileforedit(output, par->path);
		fprintf(output, "</textarea><br>\n");
		fprintf(output, "<input type=submit value=\"%s\"></form>\n\n", _("Save File"));
		printhtmlfooter(output, context->shortfoot, par->path, par->name);
		return 0;
	}


	switch (par->ftype) {
	case FILE_DIR:
		if (context->iswritable && !badquota && (par->props->type == FOLDER_NORMAL)) {
			fprintf(output, "<b>%s:</b>\n", _("Add New Content"));
			fprintf(output, "<font size=-1>\n");
			fprintf(output, "<b><a href=\"?action=upload\">");
			fprintf(output,
					"<img src=\"%s/fileops/upload.gif\" width=20 height=20 alt=\"\" border=0>\n",
					aucGraphDir);
			fprintf(output, "%s</a></b>,\n\n", _("Upload a File"));
			fprintf(output, "<b><a href=\"?action=newfile\">");
			fprintf(output,
					"<img src=\"%s/fileops/file.gif\" width=20 height=20 alt=\"\" border=0>\n",
					aucGraphDir);
			fprintf(output, "%s</a></b></font>\n", _("Create New Document or Folder"));
		}
		
		if (par->props->type == FOLDER_DROPBOX) {
			fprintf(output, "<b><a href=\"?action=upload\">");
			fprintf(output,
					"<img src=\"%s/fileops/upload.gif\" width=20 height=20 alt=\"\" border=0>\n",
					aucGraphDir);
			fprintf(output, "%s</a></b>\n\n", _("Upload a File into this Dropbox"));			
			shown = dropbox_index(output, par->path, context->expert, context->iswritable,
										par->props);
		} else
			shown = aucFileDetailsView(output, par->path, context->expert, 0, context->iswritable);
			
		break;
	case FILE_TEXTDOC:
		fprintf(output, "<p><small>Document Contents:</small><br>\n");
		fprintf(output,
				"<table width=100%% cellpadding=6 cellspacing=0 border=0 bgcolor=#EEEEEE><tr><td>\n\n");
		printtextfile(output, par->path);
		fprintf(output, "</td></tr></table></p>\n\n");
		break;
	case FILE_HTMLDOC:
		aucHtmlEchoFile(output, par->path, 1);
		break;
	case FILE_WORDDOC:
		viewworddoc(output, par->path);
		break;
	default:
		defaultfileview(output, par->path);
		break;
	}

	fprintf(output, "<p>\n\n");
	printhtmlfooter(output, context->shortfoot, par->path, par->name);

	return 0;
}



int
defaultfileview(FILE * output, char *path)
{
	fprintf(output, "FIXME: non-mime default file display\n\n");
	return 0;
}

/*
int
sed_escape(char * path, size_t len)
{
	int space, pathlen;
	char * a;
	
	pathlen = strlen(path) + 1;
	space = len - pathlen;
	for (a=path; *a; a++) {
		if (*a == '/') {
			if (space < 2)
				return -1;
			memmove(a+2, a, pathlen);
			*a = *(a+1) = '\\';
			space -= 2;
			a += 2;
		}
		pathlen -= 1;
	}
	return 0;
}
*/

int
viewworddoc(FILE * output, char *path)
{
	char buf[2048];
	char * args[11];
	int i;
	
	fprintf(output, "<hr noshade size=1>\n\n");

//	snprintf(buf, sizeof(buf), "%s/aucwv -x %s/wvHtml.xml -c UTF-8 -d \"%s/.auc/tmp\" -u \"%s/homedir/.auc/tmp/\" '%s'",
//			 aucDataDir, aucDataDir, aucUserCache->homedir, aucPagesRoot, path);
			 
	snprintf(buf, sizeof(buf), "%s/aucwv", aucDataDir);
	args[0] = strdup(buf);
	snprintf(buf, sizeof(buf), "-x");
	args[1] = strdup(buf);
	snprintf(buf, sizeof(buf), "%s/aucwv.xml", aucDataDir);
	args[2] = strdup(buf);
	snprintf(buf, sizeof(buf), "-c");
	args[3] = strdup(buf);
	snprintf(buf, sizeof(buf), "UTF-8");
	args[4] = strdup(buf);
	snprintf(buf, sizeof(buf), "-d");
	args[5] = strdup(buf);
	snprintf(buf, sizeof(buf), "%s/.auc/tmp", aucUserCache->homedir);
	args[6] = strdup(buf);
	snprintf(buf, sizeof(buf), "-u");
	args[7] = strdup(buf);
	snprintf(buf, sizeof(buf), "%s/homedir/.auc/tmp/", aucPagesRoot);
	args[8] = strdup(buf);
	snprintf(buf, sizeof(buf), "%s", path);
	args[9] = strdup(buf);
	args[10] = NULL;
	
	if (aucHtmlEchoComm(output, args[0], args)) {
		aucGetFullPath("word-error-mess.html", aucDataDir, buf, sizeof(buf));
		aucHtmlEchoFile(output, buf, 1);
	}
	
	for (i=0; i<10; i++)
		free(args[i]);
	
	return 0;
}

int
match_action(struct file_action_type * action, char * str, struct file_parameters_type * par,
		struct context_info_type * context)
{
	if (strcmp(action->name, str))
		return 0;
		
	if (!(action->typemask & (1 << par->ftype)))
		return 0;
		
	if ((action->flags & IS_WRITABLE) && !(context->iswritable))
		return 0;
		
	if ((action->flags & IS_POST) && strcmp(cgiRequestMethod, "POST"))
		return 0;
		
	return 1;
}



/**
 * Displays an embedded or standalone file manager page.
 *
 * The file manager capability of AUC can be used anywhere by a programmer.  In order to
 * use it, the programmer should define a "virtual volume" using aucRegisterFilePath().
 * This is merely an association of a name with a given real path on the server.  Then
 * a header and footer function should be created.  Finally, all this should be inside
 * a page function that parses its extrapath argument and supplies it to this function
 * as the dir argument.
 *
 * @param output Output stream.
 * @param shorthead Function called to print a page header.  This argument as well as the
 *    next argument enable a programmer to "embed" the file manager within a custom-designed
 *    page by supplying their own header or footer which may or may not be used (depending
 *    on whether a POST operation is being satisfied, etc).
 * @param shortfoot Function called to print a page footer.
 * @param topdirname A name to be given to the toplevel directory (may include spaces or
 *    other formatting to make it "pretty").  An example of this is a user's home directory
 *    where the topdirname is "Home Directory."
 * @param virtualfs The "virtual volume" that this file manager should display.  Virtual
 *    volumes are assigned to a real path on the server at runtime using the
 *    aucRegisterFilePath() function.
 * @param dir The subdirectory within the virtual volume that the file manager should display.
 * @param requested_perm A flag to indicate the permissions that the web user should have
 *    in the current volume.  TREE_FORCEWRITE will give the user write permissions regardless
 *    of the directory's UNIX permissions on disk.  TREE_USEUNIX will cause the web user to
 *    have the same permissions as the UNIX file permissions allow.  TREE_READONLY will
 *    provide read-only capability to the web user.
 * @param expert 0 to disable "expert" view.  1 to enable it.  -1 to use whatever the current
 *    web user has in his/her preferences.
 * @param hidden 1 to show hidden files.  0 to hide them.
 *
 * @return 0 on success.  Non-zero on error.
 */



int
aucViewPath(FILE * output, shorthead_fn shorthead,
			shortfoot_fn shortfoot, char *topdirname,
			char *virtualfs, char *dir, int requested_perm, int expert, int hidden)
{
	char buf[1024];
	char action[64];
	char * buf2;
	struct file_parameters_type * par;
	struct context_info_type context;
	struct file_action_type * curr_action;
	file_action_func action_func = action_error;

	/* if uploading, remove the dummy filename */
	buf2 = get_filename(dir);
	if (buf2 && !strcmp(buf2, ".uploadfile"))
		dir[strlen(dir) - strlen(buf2)] = '\0';

	context.requested_perm = requested_perm;
	context.shorthead = shorthead;
	context.shortfoot = shortfoot;
	if (expert == -1)
		context.expert = aucFileExpert;
	else
		context.expert = expert;

	par = get_file_parameters(virtualfs, dir, topdirname, context.expert);

	if ((par->ftype == FILE_DIR) && (par->path[strlen(par->path) - 1] != '/')) {
		snprintf(buf, sizeof(buf), "%s/", par->filename);
		free_file_parameters(par);
		cgiHeaderLocation(buf);
		return 0;
	}

	context.iswritable = writable(par->path, aucUserCache->uid, aucUserCache->gid, context.requested_perm);

	if (cgiFormString("action", action, sizeof(action)) != cgiFormSuccess)
		strcpy(action, "view");

	if (par->ftype != FILE_ERROR) {

		curr_action = folder_actions[par->props->type];
		while (curr_action->name) {
			if (match_action(curr_action, action, par, &context)) {
				action_func = curr_action->action;
				break;
			}
			curr_action++;
		}
		if (!curr_action->name)
			action_func = action_error;
	} else
		action_func = action_error;

	action_func(output, par, &context, action);
	
	free_file_parameters(par);
	
	return 0;
}
	


/**
 * Frees the memory of a file_parameters_type structure.
 *
 * @param	par		Pointer to the structure
 *
 * @return	Nothing.
 */

void
free_file_parameters(struct file_parameters_type * par)
{
	free(par->path);
	free(par->filename);
	free(par->parent_path);
	aucFilesFreeFolderProps(par->props);
	free(par->current_url);
	free(par->name);
	free(par->parent_url);
	free(par->parent_name);
	free(par);
	par = NULL;
}


/**
 * Stores useful paths and information about a file in a structure.
 *
 * @param	virtualfs	The "virtual volume" of the desired file
 * @param	dir			The path of the desired file within the "virtual volume"
 * @param	topdirname	The name of the topmost directory (the root of the "virtual volume")
 * @param	expert		True when the user should see the "expert mode" file view
 *
 * @return	A pointer to the malloc'ed structure
 */

struct file_parameters_type *
get_file_parameters(char * virtualfs, char * dir, char * topdirname, int expert)
{
	char buf[256], url[256], * file_url, * parent_url;
	struct file_parameters_type * par;
	
	par = malloc(sizeof(struct file_parameters_type));

						/* store basic info about the file/directory */
	getrealpath(buf, sizeof(buf), virtualfs, dir);	
	par->path = strdup(buf);
	par->filename = strdup(get_filename(buf));
	par->ftype = aucGetFileType(par->path, NULL);
	par->parent_path = get_parent_path(par->path);
	
						/* get the url to this file, and cache it's name and path */
	snprintf(url, sizeof(url), "%s/%s", virtualfs, dir);
	file_url = strdup(get_filename(url));
	parent_url = get_parent_path(url);
	
	if (par->ftype == FILE_DIR) {
						/* get the XML properties associated with this folder */
		par->props = aucFilesGetFolderProps(par->path);
		
						/* store a url to get back to this same folder */
		snprintf(buf, sizeof(buf), "../%s/", file_url);
		par->current_url = aucLinkText(buf);
		
		if (strlen(dir)) { /* we are not in the topmost folder */
						/* store a formatted name of the current folder for later display */
			aucFormatPathText(buf, sizeof(buf), file_url, par->ftype, expert);
			par->name = strdup(buf);
			
						/* the URL to the parent path */
			par->parent_url = strdup("../");
			
		} else { /* we are in the topmost folder */
						/* store a formatted name of the current folder for later display */
			aucFormatPathText(buf, sizeof(buf), topdirname, par->ftype, expert);
			par->name = strdup(buf);
			
						/* since we are in the topmost folder, there is no parent URL */
			par->parent_url = NULL;
		}
	} else {
						/* if a file, parse the contents of the parent dir's properties */
		par->props = aucFilesGetFolderProps(par->parent_path);

						/* store the url to the current file and it's parent directory */
		par->current_url = aucLinkText(file_url);
		snprintf(buf, sizeof(buf), "../%s/", get_filename(parent_url));
		par->parent_url = aucLinkText(buf);

						/* store a formatted name of the current file for later display */
		aucFormatPathText(buf, sizeof(buf), file_url, par->ftype, expert);
		par->name = strdup(buf);
	}
	
	if (strlen(parent_url) && strstr(dir, get_filename(parent_url))) /* the parent directory is not the topmost */
		aucFormatPathText(buf, sizeof(buf), get_filename(parent_url), FILE_DIR, expert);
	else /* the parent directory is the topmost */
		aucFormatPathText(buf, sizeof(buf), topdirname, FILE_DIR, expert);
					/* store a formatted name of the parent directory */
	par->parent_name = strdup(buf);
	
			/* free our cached strings */
	free(file_url);
	free(parent_url);
	
	return par;
}



int
aucNewFilePrompt(FILE * output, char *action, int expert)
{
	if (expert == -1)
		expert = aucFileExpert;
	fprintf(output, "<form method=post action=\"%s\">\n", action);
	fprintf(output, "<input type=hidden name=action value=newfile>\n");
	fprintf(output, "<input type=hidden name=text value=\"%s\">", _("Click on the pencil+paper icon above to edit the contents of this file."));
	fprintf(output, "<p><b>%s:</b>\n", _("Name of file to Create"));
	fprintf(output, "<input type=text name=filename size=40></p>\n");
	fprintf(output, "<dl><dt><b>File Type:</b></dt>\n\n");

	fprintf(output, "<dd><input type=radio name=extension value=\".txt\" checked>\n");
	fprintf(output, "<b>%s:</b>  %s</dd>\n\n", _("Text File"),
			_("The easiest format to use for putting text online."));

	fprintf(output, "<dd><input type=radio name=extension value=directory>\n");
	fprintf(output, "<b>%s:</b>  %s</dd>\n\n", _("Folder"), _("Creates a folder in which additional documents can be stored."));

	if (expert) {
		fprintf(output, "<dd><input type=radio name=extension value=\".html\">\n");
		fprintf(output, "<b>%s:</b>  %s</dd>\n\n", _("HTML formatted text"), _("Allows advanced formatting, but HTML codes must be employed."));

		fprintf(output, "<dd><input type=radio name=extension value=\"\">\n");
		fprintf(output, "<b>%s:</b>  %s</dd>\n\n", _("Fixed Text"), _("Text is preserved exactly as entered into the file, displayed in a fixed-width font."));
	}

	fprintf(output, "</dl>\n\n");
	fprintf(output, "<input type=submit value=\"%s\"></form>\n\n", _("Create File"));
	return 0;
}

int
aucRenameFilePrompt(FILE * output, char *action, char *fname, unsigned long ftype, int expert)
{
	char buf[2048], buf2[64], *a;

	if (expert == -1)
		expert = aucFileExpert;
	if (ftype == FILE_DIR)
		ftype = FILE_DEFAULT;
	a = aucFormatPathText(buf, sizeof(buf), fname, ftype, expert);
	if (strlen(a))
		snprintf(buf2, sizeof(buf2), ".%s", a);
	else
		strcpy(buf2, "");
	fprintf(output, "<form method=post action=\"%s\">\n", action);
	fprintf(output, "<input type=hidden name=action value=rename>\n");
	fprintf(output, "<input type=hidden name=extension value=\"%s\">\n", buf2);
	fprintf(output, "<p><b>%s:</b>\n", _("New Name"));
	fprintf(output, "<input type=text name=newname size=40 value=\"%s\"></p>\n", buf);
	fprintf(output, "<input type=submit value=\"%s\"></form>\n\n", _("Rename File"));
	return 0;
}

int
aucText2Filename(char *filename)
{
	char *a;
	int valid = 0;

	for (a = filename; *a; a++)
		switch (*a) {
		case ' ':
			*a = '_';
			break;
		case '/':
			*a = '\\';
			break;
		case '.':
		case '~':
			break;
		default:
			valid = 1;
			break;
		}

	if (valid)
		return 0;
	else
		return -1;
}

char *
get_filename(char *path)
{
	char *a;

	snprintf(Pathbuf, sizeof(Pathbuf), "%s", path);
	if (Pathbuf[strlen(Pathbuf) - 1] == '/')
		Pathbuf[strlen(Pathbuf) - 1] = '\0';

	for (a = &Pathbuf[strlen(Pathbuf) - 2]; a >= Pathbuf; a--) {
		if (*a == '/')
			return a + 1;
	}
	if (strlen(Pathbuf))
		return Pathbuf;
	else
		return NULL;
}


/**
 * Returns the path to a given file or directory's parent directory with a
 * trailing '/'.
 *
 * This function expects to be passed a full path.  Since the resulting path
 * is returned, it must be free'ed when done.
 *
 * @param	path	Full path of current file or directory
 *
 * @return	A malloc'ed string containing the full path of the parent
 *			directory, including a trailing '/'.  Returns an empty
 *			string if the parent directory cannot be determined.
 */

char *
get_parent_path(char * path)
{
	char *a, *str;

	for (a = &path[strlen(path) - 2]; a >= path; a--) {
		if (*a == '/')
			break;
	}
	str = (char *) malloc(sizeof(char) * (a - path + 2));
	strncpy(str, path, a - path + 1);
	str[a - path + 1] = '\0';
	return str;
}

int
compare(void *a, void *b)
{
	const char **s1, **s2;
	s1 = a;
	s2 = b;
	return strcmp(*s1, *s2);
}

int
writable(char *path, uid_t uid, gid_t gid, int requested_perm)
{
	int retval;
	char buf1[32], buf2[32];

	if (requested_perm == TREE_USEUNIX) {
		aucGetFileStats(path, buf1, buf2, &retval, uid, gid);
		return retval;
	} else if (requested_perm == TREE_FORCEWRITE)
		return 1;

	return 0;
}

int
aucGetFileStats(char *path, char *fsize, char *fdate, int *writable, uid_t uid, gid_t gid)
{
	struct stat fileinfo;
	time_t now;
	char format[64];
	struct tm *tmptr;

	*writable = 0;

	if (stat(path, &fileinfo))
		if (lstat(path, &fileinfo))
			return -1;

	if (fileinfo.st_uid == uid) {
		if (fileinfo.st_mode & S_IWUSR)
			*writable = 1;
	} else if (fileinfo.st_gid == gid) {
		if (fileinfo.st_mode & S_IWGRP)
			*writable = 1;
	} else {
		if (fileinfo.st_mode & S_IWOTH)
			*writable = 1;
	}

	now = time(NULL);

	if ((now - fileinfo.st_mtime) > (335 * 24 * 60 * 60))
		strcpy(format, "&nbsp;%b&nbsp;%d&nbsp;&nbsp;%Y");
	else
		strcpy(format, "&nbsp;%b&nbsp;%d&nbsp;%H:%M");
	tmptr = localtime(&fileinfo.st_mtime);
	strftime(fdate, 64, format, tmptr);

	if (fileinfo.st_size < 10000)
		sprintf(fsize, "%7ld", (long int) fileinfo.st_size);
	else
		sprintf(fsize, "%6ldK", fileinfo.st_size / 1024);
	return 0;
}

int
dropbox_index(FILE * output, char *path, int expert,
		int iswritable, aucFolderPtr folder_props)
{
	xmlNodePtr node;
	char * name, * user, * time_str;
	time_t create_time;
	char buf[256];
	unsigned long ftype;
	
	node = xmlDocGetRootElement(folder_props->doc);
	node = node->children;
	
	fprintf(output, "<table cellpadding=1 cellspacing=0 border=0 width=600>\n");
	while (node) {
		int temp;
		char fsize[64], fdate[64], timestr[64];
		struct tm * ltime;
		
		if (!node->name || strcmp(node->name, "file")) {
			node = node->next;
			continue;
		}
		name = xmlGetProp(node, "name");
		user = xmlGetProp(node, "uname");
		
		if (!iswritable && strcmp(user, aucUserCache->uname)) {
			free(name);
			free(user);
			node = node->next;
			continue;
		}
		

		snprintf(buf, sizeof(buf), "%s%s_%s", path, user, name);
		if (aucGetFileStats(buf, fsize, fdate, &temp, aucUserCache->uid, aucUserCache->gid)) {
			xmlNodePtr nextnode = node->next;
			
			if (iswritable) {
				xmlUnlinkNode(node);
				xmlFreeNode(node);
				GET_PROPFILE(buf, path);
				save_propfile(buf, folder_props->doc, TREE_FORCEWRITE);
			}
			free(name);
			free(user);
			node = nextnode;
			continue;
		}

		time_str = xmlGetProp(node, "time");
		create_time = strtoul(time_str, NULL, 10);
		free(time_str);
		
		fprintf(output, "<tr bgcolor=#EEEEEE><td width=350>\n");
		snprintf(buf, sizeof(buf), "%s_%s", user, name);
		ftype = aucGetFileType(path, buf);
		
		if (!iswritable)
			fprintf(output, "<b>%s</b>\n", name);
		else
			aucLinkFileShort(output, aucLinkText(buf), name, ftype, expert);
			
		fprintf(output, "</td><td width=50><font size=-1><tt>%s</td></font></td>\n", fsize);
		fprintf(output, "<td width=200>\n");
		
		snprintf(buf, sizeof(buf), "%s_%s", user, name);
		
		if (iswritable) {
			fprintf(output, "<a href=\"%s?action=download\">Download</a>,\n", aucLinkText(buf));
			fprintf(output, "<a href=\"%s?action=delete\">Delete</a>\n", aucLinkText(buf));
		}
		
		fprintf(output, "</td></tr>\n\n");
		
		aucAuthGetFullName(buf, sizeof(buf), user, NAME_LINK);
		ltime = localtime(&create_time);
		strftime(timestr, sizeof(timestr), "%a, %d %B %Y, %l:%M %P", ltime);
		fprintf(output, "<tr><td colspan=3><blockquote>\n");
		
		if (iswritable)
			fprintf(output, "Submitted by <b>%s</b> on <b>%s</b>.\n", buf, timestr);
		else
			fprintf(output, "Submitted by you on <b>%s</b>.\n", timestr);
			
		fprintf(output, "</td></tr>\n\n");

		free(name);
		free(user);
		node = node->next;
	}
	fprintf(output, "</table>\n\n");
	
	if (!iswritable)
		fprintf(output, "<p>(%s)</p>\n", _("Only files submitted by you are listed here"));
	
	return 0;
}

int
aucFileDetailsView(FILE * output, char *path, int expert, int hidden, int iswritable)
{
	DIR *dirptr;
	struct dirent *entry;
	char *files[2048];
	int x = 0, num = 0, shown = 0;
	unsigned long ftype;
	char buf[1024];

	if (expert == -1)
		expert = aucFileExpert;

	dirptr = opendir(path);
	if (!dirptr) {
		fprintf(output, "<i>%s</i>\n\n", strerror(errno));
		return 0;
	}

	x = 0;
	while ((entry = readdir(dirptr)) && (x < sizeof(files)))
		files[x++] = (char *) strdup(entry->d_name);
	num = x;
	qsort(files, num, sizeof(char *), (compareFunc) compare);
	fprintf(output, "<table cellpadding=1 cellspacing=0 border=0 width=600>\n");
	fprintf(output, "<tr><td width=280><b><u>%s</u></b></td>\n", _("Name"));
	fprintf(output, "<td width=50><p align=right><b><u>%s</u></b></p></td>\n", _("Size"));
	fprintf(output, "<td width=100><p align=right><b><u>%s</u></b></p></td>\n", _("Date"));
	fprintf(output, "<td width=170>&nbsp;&nbsp;<b><u>%s</u></b></td></tr>\n\n",
			_("File Operations"));
	for (x = 0; x < num; x++) {
		int temp;
		char fsize[64], fdate[64];
		aucFolderPtr props;
		
		if (!hidden && (*files[x] == '.')) {
			free(files[x]);
			continue;
		}
		shown++;
		ftype = aucGetFileType(path, files[x]);
		/*   snprintf(buf, sizeof(buf), "%s%s/%s/%s%s", aucRootDir, aucBaseName,
		   virtualfs, newdir, aucLinkText(files[x])); */
		fprintf(output, "<tr bgcolor=%s><td width=280>\n", (shown % 2) ? "#EEEEEE" : "#FFFFFF");
		aucLinkFileShort(output, aucLinkText(files[x]), files[x], ftype, expert);
		snprintf(buf, sizeof(buf), "%s/%s", path, files[x]);
		aucGetFileStats(buf, fsize, fdate, &temp, aucUserCache->uid, aucUserCache->gid);
		fprintf(output,
				"</td><td width=50><p align=right><font size=-1><tt>%s</tt></font></p></td>\n",
				fsize);
		fprintf(output, "<td width=100><p align=right><font size=-1><tt>%s</tt></font></p></td>\n",
				fdate);
		fprintf(output, "<td width=170>\n");
		/*   snprintf(buf, sizeof(buf), "%s%s/%s/%s%s", aucRootDir, aucBaseName,
		   virtualfs, newdir, files[x]); */
		//    fprintf(output, "&nbsp;&nbsp;&nbsp;<font size=-1>\n");
		printfileops(output, aucLinkText(files[x]), ftype, iswritable);
		//     fprintf(output, "</font>\n\n");
		fprintf(output, "</td></tr>\n\n");
		
		if ((ftype == FILE_DIR) && (props = aucFilesGetFolderProps(buf))) {
			if (props->description && strlen(props->description)) {
				fprintf(output, "<tr bgcolor=%s><td colspan=4 width=600>\n", (shown % 2) ? "#EEEEEE" : "#FFFFFF");
				fprintf(output, "<blockquote>%s: %s</blockquote>\n",
					_("Description"), props->description);
				fprintf(output, "</td></tr>\n\n");
			}
			aucFilesFreeFolderProps(props);
		}
		
		free(files[x]);
	}
	if (shown == 0) {
		fprintf(output, "<tr><td><i>%s</i></td></tr>\n\n", _("Folder is empty"));
	}
	fprintf(output, "</table>\n\n");
	closedir(dirptr);
	return shown;
}

int
aucFileListView(FILE * output, char *virtualfs, char *dir, int expert, int hidden, int cols)
{
	DIR *dirptr;
	struct dirent *entry;
	char path[1024], newdir[1024];
	char *files[2048];
	int x = 0, num = 0, shown = 0;
	unsigned long ftype;
	char buf[1024];

	if (expert == -1)
		expert = aucFileExpert;

	if (strlen(dir) && (dir[strlen(dir) - 1] != '/'))
		snprintf(newdir, sizeof(newdir), "%s/", dir);
	else
		snprintf(newdir, sizeof(newdir), "%s", dir);

	if (getrealpath(path, sizeof(path), virtualfs, newdir)) {
		snprintf(buf, sizeof(buf), "%s/homedir", aucPagesRoot);
		aucLinkFileShort(output, buf, _("Home Directory"), FILE_DIR, 0);
		return 1;
	}

	dirptr = opendir(path);
	if (!dirptr) {
		fprintf(output, "<i>%s</i>\n\n", strerror(errno));
		return 0;
	}

	x = 0;
	while ((entry = readdir(dirptr)) && (x < sizeof(files)))
		files[x++] = (char *) strdup(entry->d_name);
	num = x;
	qsort(files, num, sizeof(char *), (compareFunc) compare);
	for (x = 0; x < num; x++) {
		if (!hidden && (*files[x] == '.')) {
			free(files[x]);
			continue;
		}
		shown++;
		ftype = aucGetFileType(path, files[x]);
		snprintf(buf, sizeof(buf), "%s/%s/%s%s", aucPagesRoot,
				 aucLinkText(virtualfs), newdir, aucLinkText(files[x]));
		aucLinkFileShort(output, buf, files[x], ftype, expert);
		free(files[x]);
	}
	if (shown == 0) {
		fprintf(output, "<i>%s</i>\n\n", _("Folder is empty"));
	}
	closedir(dirptr);
	return shown;
}

aucFolderPtr
aucFilesGetFolderProps(char * path)
{
	char xmlpath[256];
	xmlDocPtr props;
	xmlXPathObjectPtr xpathobj;
	xmlXPathContextPtr xpathctx;
	aucFolderPtr result;
	
	snprintf(xmlpath, sizeof(xmlpath), "%s/%s", path, ".aucfolder");
	props = xmlParseFile(xmlpath);
	
	result = (aucFolderPtr) malloc(sizeof(aucFolder));
	
	if (!props) {
		result->doc = NULL;
		result->description = NULL;
		result->type = FOLDER_NORMAL;
		return result;
	}
	
	xpathctx = xmlXPathNewContext(props);

	result->doc = props;
	result->description = NULL;
	xpathobj = xmlXPathEval("/folder/description", xpathctx);

	if (xpathobj->nodesetval->nodeNr > 0) {
		char * str;
		if ((str = xmlNodeGetContent(xpathobj->nodesetval->nodeTab[0])))
			result->description = strdup(str);
	}

	xmlXPathFreeObject(xpathobj);

	result->type = FOLDER_NORMAL;
	
	xpathobj = xmlXPathEval("/folder/attribute::type", xpathctx);
	if (xpathobj->nodesetval->nodeNr > 0) {
		xmlNodePtr node;
		node = xpathobj->nodesetval->nodeTab[0];
		if (((xmlAttrPtr) node)->children && ((xmlAttrPtr) node)->children->content) {
			if (!strcmp(((xmlAttrPtr) node)->children->content, "dropbox"))
				result->type = FOLDER_DROPBOX;
		}
	}

	xmlXPathFreeObject(xpathobj);

	xmlXPathFreeContext(xpathctx);
	
	return result;
}

void
aucFilesFreeFolderProps(aucFolderPtr props)
{
	xmlFreeDoc(props->doc);
	free(props->description);
	free(props);
}

int
printfileops(FILE * output, char *url, unsigned long ftype, int iswritable)
{
	char buffer[1024];

	if ((ftype == FILE_DIR) && (url[strlen(url) - 1] != '/'))
		sprintf(buffer, "%s/", url);
	else
		strcpy(buffer, url);

	fprintf(output, "<table cellpadding=0 cellspacing=2 border=0 width=134>\n");
	fprintf(output, "<tr valign=top>\n");

	fprintf(output, "<td width=20>&nbsp;</td>\n");

	if (iswritable && ((ftype == FILE_TEXTDOC) || (ftype == FILE_HTMLDOC))) {
		fprintf(output, "<td width=20><a href=\"%s?action=edit\" onMouseOver=\"window.status='%s'; return true\">",
			buffer, _("Edit File"));
		fprintf(output,
				"<img src=\"%s/fileops/edit.gif\" width=20 height=20 alt=\"Edit\" border=0>",
				aucGraphDir);
		fprintf(output, "</a></td>\n");
	} else
		fprintf(output, "<td width=20>&nbsp;</td>\n");

	if ((ftype != FILE_DIR) && (ftype != FILE_ERROR)) {
		fprintf(output, "<td width=20><a href=\"%s?action=download\" onMouseOver=\"window.status='%s'; return true\">",
			buffer, _("Download File"));
		fprintf(output,
				"<img src=\"%s/fileops/download.gif\" width=20 height=20 alt=\"Dnload\" border=0>",
				aucGraphDir);
		fprintf(output, "</a></td>\n");
	} else
		fprintf(output, "<td width=20>&nbsp;</td>\n");

	if (iswritable) {
		fprintf(output, "<td width=20><a href=\"%s?action=delete\" onMouseOver=\"window.status='%s'; return true\">",
			buffer, _("Delete"));
		fprintf(output,
				"<img src=\"%s/fileops/trash.gif\" width=20 height=20 alt=\"Del\" border=0>",
				aucGraphDir);
		fprintf(output, "</a></td>\n");
	} else
		fprintf(output, "<td width=20>&nbsp;</td>\n");

	if (iswritable) {
		fprintf(output, "<td width=20><a href=\"%s?action=rename\" onMouseOver=\"window.status='%s'; return true\">",
			buffer, _("Rename"));
		fprintf(output,
				"<img src=\"%s/fileops/rename.gif\" width=20 height=20 alt=\"Rename\" border=0>",
				aucGraphDir);
		fprintf(output, "</a></td>\n");
	} else
		fprintf(output, "<td width=20>&nbsp;</td>\n");
		
	if (iswritable && (ftype == FILE_DIR)) {
		fprintf(output, "<td width=20><a href=\"%s?action=props\" onMouseOver=\"window.status='%s'; return true\">",
			buffer, _("Edit Folder Properties"));
		fprintf(output,
				"<img src=\"%s/fileops/props.gif\" width=20 height=20 alt=\"Properties\" border=0>",
				aucGraphDir);
		fprintf(output, "</a></td>\n");
	} else
		fprintf(output, "<td width=20>&nbsp;</td>\n");

	fprintf(output, "</tr></table>\n\n");

	return 0;
}

int
aucRegisterFilePath(char *root, char *virtualmatch, int cutoff)
{
	(*lastvirtual) = (struct virtualfs *) malloc(sizeof(struct virtualfs));

	(*lastvirtual)->root = root;
	(*lastvirtual)->virtualmatch = virtualmatch;
	(*lastvirtual)->cutoff = cutoff;

	lastvirtual++;
	*lastvirtual = NULL;
	return 0;
}

int
matchvirtuals(char *str, size_t len, char *virtualfs, char *dir)
{
	struct virtualfs **currvirt;

	for (currvirt = virtuals; *currvirt; currvirt++) {
		if ((*currvirt)->cutoff) {
			if (!strncmp(virtualfs, (*currvirt)->virtualmatch, (*currvirt)->cutoff)) {
				snprintf(str, len, "%s/%s/%s", (*currvirt)->root,
						 virtualfs + (*currvirt)->cutoff, dir);
				return 0;
			}
		} else {
			if (!strcmp(virtualfs, (*currvirt)->virtualmatch)) {
				snprintf(str, len, "%s/%s", (*currvirt)->root, dir);
				return 0;
			}
		}
	}
	return -1;
}

int
getrealpath(char *str, size_t len, char *virtualfs, char *dir)
{
	if (!strcmp(virtualfs, "usertree")) {
		return 1;
	} else if (!strncmp(virtualfs, "root", 4)) {
		snprintf(str, len, "%s/%s", virtualfs + 4, dir);
	} else if (!strncmp(virtualfs, "homedir", 7)) {
		snprintf(str, len, "%s%s/%s", aucUserCache->homedir, virtualfs + 7, dir);
	} else if (!matchvirtuals(str, len, virtualfs, dir)) {
	} else {
		snprintf(str, len, "%s/%s/%s", aucVarRoot, virtualfs, dir);
	}
	return 0;
}

unsigned long
aucGetFileType(char *path, char *name)
{
	struct stat fileinfo;
	char buf[1024];

	if (path && name)
		snprintf(buf, sizeof(buf), "%s/%s", path, name);
	else
		snprintf(buf, sizeof(buf), "%s", path);

	if (stat(buf, &fileinfo)) {
//      if (lstat(buf, &fileinfo))
		return (FILE_ERROR);
//      return(FILE_DEFAULT);
	}

	if (S_ISDIR(fileinfo.st_mode))
		return (FILE_DIR);
	else
		return (getfiletype(buf));
}

char *
aucFormatPathText(char *str, size_t len, char *path, unsigned long ftype, int expert)
{
	char buf[2048];
	int foundext = 1;
	char *a, *res;

	res = &path[strlen(path)];
	if (expert == -1)
		expert = aucFileExpert;

	if (ftype > FILE_DIR)
		foundext = 0;

	if ((ftype == FILE_DIR) && (path[strlen(path) - 1] != '/'))
		snprintf(str, len, "%s/", path);
	else
		snprintf(str, len, "%s", path);
	for (a = &str[strlen(str)]; a >= str; a--) {
		if (*a == '/') {
			snprintf(buf, sizeof(buf), "<b>/</b>%s", a + 1);
			snprintf(a, len - (a - str), "%s", buf);
			continue;
		}
		if (!expert && (*a == '_'))
			*a = ' ';
		if (!expert && (*a == '\\'))
			*a = '/';
		if (!expert && !foundext && (*a == '.')) {
			foundext = 1;
			res = a + 1;
			*a = '\0';
		}
	}
	return res;
}

int
aucLinkFileShort(FILE * output, char *link, char *name, unsigned long ftype, int expert)
{
	char realname[1024];

	if (expert == -1)
		expert = aucFileExpert;

	aucFormatPathText(realname, sizeof(realname), name, ftype, expert);

	fprintf(output, "<table cellpadding=0 cellspacing=1 border=0>\n");
	fprintf(output, "<tr><td width=20><img src=\"%s/%s\" width=20 height=22 border=0 alt=\"\"></td>\n",
			aucGraphDir, typeimages[ftype]);
	fprintf(output, "<td><a href=\"%s%s\">", link, (ftype == FILE_DIR) ? "/" : "");
	fprintf(output, "%s</a></td></tr></table>\n\n", realname);
	return 0;
}

int
aucFileBrowserH(FILE * output, char *extrapath)
{
	int homedirhead(FILE * output, char *path, char *name)
	{
		aucHtmlSlimHead("%s", _("Home Directory"));
		return 0;
	}
	int homedirfoot(FILE * output, char *path, char *name)
	{
		aucHtmlSlimFoot();
		return 0;
	}

	aucViewPath(output, homedirhead, homedirfoot, _("Home Directory"), "homedir", extrapath,
			 TREE_USEUNIX, -1, 0);
	return 0;
}

int
aucFileRemoveDir(char * path, int contents_only)
{
	char buf[256];
	struct stat statbuf;
	struct dirent * dir_ent;
	DIR * dir;
	
	if (!(dir = opendir(path)))
		return -1;

	while ((dir_ent = readdir(dir))) {
		snprintf(buf, sizeof(buf), "%s/%s", path, dir_ent->d_name);
		if (stat(buf, &statbuf) || !strcmp(dir_ent->d_name, ".") || !strcmp(dir_ent->d_name, ".."))
			continue;
		if (S_ISDIR(statbuf.st_mode))
			aucFileRemoveDir(buf, 0);
		else
			unlink(buf);
	}
	
	closedir(dir);
	if (!contents_only)
		rmdir(path);
	return 0;
}

int
getfiletype(char *name)
{
	char *end, ext[256];

	end = name + strlen(name);
	while ((end != name) && (*end != '.'))
		end--;
	if (*end == '.')
		snprintf(ext, sizeof(ext), "%s", end + 1);
	else
		strcpy(ext, "");

	if (aucWordViewInline) {
		if (!strcmp(ext, "doc") || !strcmp(ext, "DOC"))
			return FILE_WORDDOC;
	}
	if (!strcmp(ext, "txt") || !strcmp(ext, "TXT"))
		return FILE_TEXTDOC;
	if (!strcmp(ext, "htm") || !strcmp(ext, "HTM"))
		return FILE_HTMLDOC;
	if (!strcmp(ext, "html") || !strcmp(ext, "HTML"))
		return FILE_HTMLDOC;
	return FILE_DEFAULT;
}

char *
filetypeimage(int type)
{
	char buf[1024];

	switch (type) {
	case FILE_WORDDOC:
		sprintf(buf, "%s/%s", aucGraphDir, "filetypes/type_doc.gif");
		return (char *) strdup(buf);
	case FILE_TEXTDOC:
		sprintf(buf, "%s/%s", aucGraphDir, "filetypes/type_text.gif");
		return (char *) strdup(buf);
	case FILE_HTMLDOC:
		sprintf(buf, "%s/%s", aucGraphDir, "filetypes/type_image.gif");
		return (char *) strdup(buf);
	default:
		sprintf(buf, "%s/%s", aucGraphDir, "filetypes/type_default.gif");
		return (char *) strdup(buf);
	}
}

int
printtextfile(FILE * output, char *path)
{
	FILE *infile;
	char bigbuf[1024 * 16];
	int count;

	if ((infile = fopen(path, "r"))) {
		while ((count = fread(bigbuf, 1, sizeof(bigbuf), infile)))
			aucPrintText(bigbuf, output, 0);
		fclose(infile);
	}
	return 0;
}

int
printtextfileforedit(FILE * output, char *path)
{
	FILE *infile;
	char bigbuf[1024 * 16];
	int count;

	if ((infile = fopen(path, "r"))) {
		while ((count = fread(bigbuf, 1, sizeof(bigbuf), infile)))
			aucPrintText(bigbuf, output, TEXT_NOLINKS | TEXT_BSPACES | TEXT_NOBREAKS);
		fclose(infile);
	}
	return 0;
}
