/*
 * Authenticated User Community
 * Copyright (c) 2000 David C. Moore
 *
 * You may distribute and/or use for any purpose modified or unmodified
 * copies of this software if you preserve the copyright notice above.
 *
 * forms.c
 *
 * Functions to display and process forms from an XML template file
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <libxml/parser.h>
#include <libxml/parserInternals.h>
#include <libxml/xpath.h>
#include <libxml/HTMLparser.h>

#include <auc/auc.h>
#include <auc/cgic.h>
#include <auc/util.h>
#include <auc/database.h>
#include <auc/plugin.h>
#include <auc/nodelett.h>
#include <auc/html.h>

#include "forms.h"

typedef enum
{
	START_FORM,
	IN_TITLE,
	IN_DESC,
	IN_TOKEN,
	IN_OPTION,
	IN_OPTION_TITLE,
	IGNORE,
	FINISH_FORM
}
statesEnum;

typedef enum
{
	FIXED,
	INLINE,
	GETVERSION
}
handlerEnum;

typedef enum
{
	TEXT,
	TEXTAREA,
	RADIO,
	CHECKBOX,
	DROPDOWN,
	LIST,
	PASSWORD,
	SUBMIT,
	LABEL,
	HIDDEN,
	ROLEMATRIX
}
typeEnum;

typedef struct field_info
{
	char name[32];
	handlerEnum handler;
	typeEnum type;
	int length, width, height, maxlen;
	int ptitle, pfield, pdesc, printable, storexml, hascontent;
	char xmlpath[256];
	xmlNodePtr xmlnode;
	char defaultvalue[256];
	char link[256];
}
fieldInfo;

typedef struct state_info
{
	statesEnum state;
	int ignore_depth;
	statesEnum saved_state;
	int heading_depth;
	int is_attribute;
	int cellcount;
	FILE *output;
	aucVarField *vars;
	xmlDocPtr xmldoc;
	fieldInfo field;
	fieldInfo field_attr;
	fieldInfo *currfield;
}
sax_state;

void start_el(sax_state *, xmlChar *, xmlChar **);
void end_el(sax_state *, xmlChar *);
void get_chars(sax_state *, xmlChar *, int len);
void start_doc(sax_state *);
void end_doc(sax_state *);
void start_el_w(sax_state *, xmlChar *, xmlChar **);
void end_el_w(sax_state *, xmlChar *);
void start_doc_w(sax_state *);
void end_doc_w(sax_state *);
int test_form_conditional(sax_state *, xmlChar **);
xmlEntityPtr get_entity(sax_state *, xmlChar *);
void p_warning(sax_state *, char *, ...);
void p_error(sax_state *, char *, ...);
void p_fatal_error(sax_state *, char *, ...);
char *get_token(aucVarField *, char *);
void initialize_field(fieldInfo *, xmlChar **, aucVarField *);
handlerEnum str2handler(char *);
typeEnum str2type(char *);
void end_field_row(sax_state *);

char *get_form_input(fieldInfo * field);
void xml_process_formattr(xmlDocPtr doc, char *ppath, char *pname, fieldInfo * attrfield);
void xml_process_formtoken(xmlDocPtr doc, fieldInfo * field);
void xml_cond_process_formtoken(xmlDocPtr doc, fieldInfo * field, char *name);
void store_text_var(aucVarField * vars, fieldInfo * field);
void store_boolean_var(aucVarField * vars, fieldInfo * field, char *name);
xmlNodePtr xml_add_node(xmlDocPtr doc, char *name, char *path, int allowdups);
xmlNodePtr xml_find_node(xmlDocPtr doc, char *name, char *path);
void xml_del_node(xmlDocPtr doc, char *name, char *path);
void xml_print_node(xmlDocPtr doc, char *name, char *path, FILE * output);
void xml_tokenattr_default(fieldInfo * field, char *ppath, char *pname, xmlDocPtr doc);
void xml_token_default(fieldInfo * field, xmlDocPtr doc);
int xml_token_exist(fieldInfo * field, char *content, xmlDocPtr doc);
void insert_var(aucVarField * vars, char *name, char *value);
// void xml_remove_blank_nodes(xmlDocPtr doc);
char *xml_get_content(xmlNodePtr node);
char *xml_xpath_value(char *xpath, xmlDocPtr doc);
void process_option_method(sax_state * docinfo, char *method);

void table_start(FILE *, aucVarField *);
void table_end(FILE *);
void row_start(FILE *, int);
void row_end(FILE *);
void cell_start(FILE *);
void cell_start_long(FILE *, int);
void cell_end(FILE *);
void cell_blank(FILE *);
void cell_field_start(FILE *, fieldInfo *, xmlDocPtr);
void cell_field_end(FILE *, fieldInfo *);
void cell_field_start_option(FILE *, fieldInfo *, xmlChar **, xmlDocPtr);
void cell_field_end_option(FILE *, fieldInfo *);

/* The xml parsing callbacks used when parsing a form template and displaying
   the form on a webpage. */
static xmlSAXHandler saxhandler = {
	NULL,						/* internalSubsetSAXFunc */
	NULL,						/* isStandaloneSAXFunc */
	NULL,						/* hasInternalSubsetSAXFunc */
	NULL,						/* hasExternalSubsetSAXFunc */
	NULL,						/* resolveEntitySAXFunc */
	(getEntitySAXFunc) get_entity,	/* getEntitySAXFunc */
	NULL,						/* entityDeclSAXFunc */
	NULL,						/* notationDeclSAXFunc */
	NULL,						/* attributeDeclSAXFunc */
	NULL,						/* elementDeclSAXFunc */
	NULL,						/* unparsedEntityDeclSAXFunc */
	NULL,						/* setDocumentLocatorSAXFunc */
	(startDocumentSAXFunc) start_doc,	/* startDocumentSAXFunc */
	(endDocumentSAXFunc) end_doc,	/* endDocumentSAXFunc */
	(startElementSAXFunc) start_el,	/* startElementSAXFunc */
	(endElementSAXFunc) end_el,	/* endElementSAXFunc */
	NULL,						/* referenceSAXFunc */
	(charactersSAXFunc) get_chars,	/* charactersSAXFunc */
	NULL,						/* ignorableWhitespaceSAXFunc */
	NULL,						/* processingInstructionSAXFunc */
	NULL,						/* commentSAXFunc */
	(warningSAXFunc) p_warning,	/* warningSAXFunc */
	(errorSAXFunc) p_error,		/* errorSAXFunc */
	(fatalErrorSAXFunc) p_fatal_error,	/* fatalErrorSAXFunc */
	NULL,						/* getParameterEntitySAXFunc */
	NULL,						/* cdataBlockSAXFunc */
	NULL						/* externalSubsetSAXFunc */
};

/* The xml parsing callbacks used when parsing a form template and writing
   the form POST results to variables or an xml document. */
static xmlSAXHandler saxhandler_w = {
	NULL,						/* internalSubsetSAXFunc */
	NULL,						/* isStandaloneSAXFunc */
	NULL,						/* hasInternalSubsetSAXFunc */
	NULL,						/* hasExternalSubsetSAXFunc */
	NULL,						/* resolveEntitySAXFunc */
	(getEntitySAXFunc) get_entity,	/* getEntitySAXFunc */
	NULL,						/* entityDeclSAXFunc */
	NULL,						/* notationDeclSAXFunc */
	NULL,						/* attributeDeclSAXFunc */
	NULL,						/* elementDeclSAXFunc */
	NULL,						/* unparsedEntityDeclSAXFunc */
	NULL,						/* setDocumentLocatorSAXFunc */
	(startDocumentSAXFunc) start_doc_w,	/* startDocumentSAXFunc */
	(endDocumentSAXFunc) end_doc_w,	/* endDocumentSAXFunc */
	(startElementSAXFunc) start_el_w,	/* startElementSAXFunc */
	(endElementSAXFunc) end_el_w,	/* endElementSAXFunc */
	NULL,						/* referenceSAXFunc */
	NULL,						/* charactersSAXFunc */
	NULL,						/* ignorableWhitespaceSAXFunc */
	NULL,						/* processingInstructionSAXFunc */
	NULL,						/* commentSAXFunc */
	(warningSAXFunc) p_warning,	/* warningSAXFunc */
	(errorSAXFunc) p_error,		/* errorSAXFunc */
	(fatalErrorSAXFunc) p_fatal_error,	/* fatalErrorSAXFunc */
	NULL,						/* getParameterEntitySAXFunc */
	NULL,						/* cdataBlockSAXFunc */
	NULL						/* externalSubsetSAXFunc */
};

int
aucFormsXMLDisplayFile(FILE * output, char *file, aucVarField * vars, xmlDocPtr reference_doc)
{
	xmlParserCtxtPtr ctxt;
	sax_state *docinfo;
	char buf[128];

	docinfo = (sax_state *) malloc(sizeof(sax_state));

	docinfo->output = output;
	docinfo->vars = vars;
	docinfo->xmldoc = reference_doc;

	snprintf(buf, sizeof(buf), "%s", file);
	if (*file != '/')
		aucGetFullPath(file, aucDataDir, buf, sizeof(buf));
	ctxt = xmlCreateFileParserCtxt(buf);
	if (ctxt == NULL) {
		free(docinfo);
		if (file && strlen(file))
			fprintf(output, "<p><b>%s</b>: %s %s. %s</p>\n\n",
					_("Error"), file, _("not found"),
					_("AUC requires this file to display the requested form. \
Please reinstall it from the AUC distribution."));
		return -1;
	}

	ctxt->sax = &saxhandler;
	ctxt->userData = (void *) docinfo;

	xmlParseDocument(ctxt);

	ctxt->sax = NULL;
	free(docinfo);
	xmlFreeParserCtxt(ctxt);
	return 0;
}

int
aucFormsXMLDisplay(FILE * output, char *xmldata, aucVarField * vars, xmlDocPtr reference_doc)
{
	xmlParserCtxtPtr ctxt;
	sax_state *docinfo;

	docinfo = (sax_state *) malloc(sizeof(sax_state));

	docinfo->output = output;
	docinfo->vars = vars;
	docinfo->xmldoc = reference_doc;

	ctxt = xmlCreateMemoryParserCtxt(xmldata, strlen(xmldata));
	if (ctxt == NULL)
		return -1;

	ctxt->sax = &saxhandler;
	ctxt->userData = (void *) docinfo;

	xmlParseDocument(ctxt);

	ctxt->sax = NULL;
	free(docinfo);
	xmlFreeParserCtxt(ctxt);
	return 0;
}

int
aucFormsXMLProcessFile(FILE * output, char *file, aucVarField * vars, xmlDocPtr * output_doc)
{
	xmlParserCtxtPtr ctxt;
	sax_state *docinfo;
	char buf[128];

	docinfo = (sax_state *) malloc(sizeof(sax_state));

	docinfo->output = output;
	docinfo->vars = vars;

	if (!output_doc)
		docinfo->xmldoc = NULL;
	else if (!*output_doc) {
		docinfo->xmldoc = xmlNewDoc("1.0");
		*output_doc = docinfo->xmldoc;
	} else {
		docinfo->xmldoc = *output_doc;
	}

	snprintf(buf, sizeof(buf), "%s", file);
	if (*file != '/')
		aucGetFullPath(file, aucDataDir, buf, sizeof(buf));
	ctxt = xmlCreateFileParserCtxt(buf);
	if (ctxt == NULL)
		return -1;

	ctxt->sax = &saxhandler_w;
	ctxt->userData = (void *) docinfo;

	xmlParseDocument(ctxt);

	ctxt->sax = NULL;
	free(docinfo);
	xmlFreeParserCtxt(ctxt);
	return 0;
}

int
aucFormsXMLProcess(FILE * output, char *xmldata, aucVarField * vars, xmlDocPtr * output_doc)
{
	xmlParserCtxtPtr ctxt;
	sax_state *docinfo;

	docinfo = (sax_state *) malloc(sizeof(sax_state));

	docinfo->output = output;
	docinfo->vars = vars;

	if (!output_doc)
		docinfo->xmldoc = NULL;
	else if (!*output_doc) {
		docinfo->xmldoc = xmlNewDoc("1.0");
		*output_doc = docinfo->xmldoc;
	} else {
		docinfo->xmldoc = *output_doc;
	}

	ctxt = xmlCreateMemoryParserCtxt(xmldata, strlen(xmldata));
	if (ctxt == NULL)
		return -1;

	ctxt->sax = &saxhandler_w;
	ctxt->userData = (void *) docinfo;

	xmlParseDocument(ctxt);

	ctxt->sax = NULL;
	free(docinfo);
	xmlFreeParserCtxt(ctxt);
	return 0;
}

xmlXPathContextPtr
xmlXPathNewContextNode(xmlDocPtr doc, xmlNodePtr node)
{
	xmlXPathContextPtr retval;

	retval = xmlXPathNewContext(doc);
//	retval->nodelist = xmlXPathNodeSetCreate(node);
	retval->node = node;
	return retval;
}

int
aucFormsProcessListEdit(FILE * output, char *file, xmlDocPtr doc, char *action)
{
	xmlXPathContextPtr tokenctxt, ctxt;
	xmlXPathObjectPtr tokenobj, optionobj, titleobj, headersobj, listobj;
	xmlDocPtr refdoc;
	xmlNodePtr token, olditem = NULL;
	char *values[8], buf[256], *listname;
	int x, retval = 1, updown = 0, num = 0;

	aucHtmlReadForm("editlist edititem deleteitem upitem downitem", values, 0);
	if (!strlen(values[0]))
		return 0;
	listname = values[0];

	refdoc = xmlParseFile(file);

	tokenctxt = xmlXPathNewContext(refdoc);
	snprintf(buf, sizeof(buf),
			 "/descendant::ConfigToken[@name='%s'][@type='list']", listname);
	tokenobj = xmlXPathEval(buf, tokenctxt);
	if (tokenobj && (tokenobj->type == 1) && (tokenobj->nodesetval->nodeNr))
		token = tokenobj->nodesetval->nodeTab[0];
	else
		return 0;

	if (strlen(values[2]) || strlen(values[1])) {
		if (strlen(values[1])) {
			snprintf(buf, sizeof(buf), "%s[%s]", listname, values[1]);
			olditem = xmlCopyNode(xml_find_node(doc, buf, xmlGetProp(token, "parentpath")), 1);
		} else
			snprintf(buf, sizeof(buf), "%s[%s]", listname, values[2]);

		xml_del_node(doc, buf, xmlGetProp(token, "parentpath"));
		retval = 2;
	}

	if (strlen(values[3])) {
		updown = -1;
		num = strtoul(values[3], NULL, 10);
	} else if (strlen(values[4])) {
		updown = 1;
		num = strtoul(values[4], NULL, 10);
	}

	if (updown) {
		xmlNodePtr movednode, curnode;

		snprintf(buf, sizeof(buf), "%s[%d]", listname, num + updown);
		movednode = xml_find_node(doc, buf, xmlGetProp(token, "parentpath"));
		snprintf(buf, sizeof(buf), "%s[%d]", listname, num);
		curnode = xml_find_node(doc, buf, xmlGetProp(token, "parentpath"));
		if (movednode) {
			xmlUnlinkNode(movednode);
			if (updown == -1)
				xmlAddNextSibling(curnode, movednode);
			else if (updown == 1)
				xmlAddPrevSibling(curnode, movednode);
			retval = 2;
		}
	}

	ctxt = xmlXPathNewContextNode(refdoc, token);
	optionobj = xmlXPathEval("child::option", ctxt);
	xmlXPathFreeContext(ctxt);

	if (!strcmp(cgiRequestMethod, "GET")) {
		char title[256], *length;

		snprintf(title, sizeof(title), "%s", _("Edit List"));

		cgiStaticContentType("text/html");

		ctxt = xmlXPathNewContextNode(refdoc, token);
		titleobj = xmlXPathEval("child::title", ctxt);
		xmlXPathFreeContext(ctxt);
		if (titleobj && (titleobj->type == 1) && titleobj->nodesetval->nodeTab[0])
			snprintf(title, sizeof(title), "%s %s", _("Edit"),
					 xmlNodeGetContent(titleobj->nodesetval->nodeTab[0]));

		aucHtmlPopupHead(title);

		fprintf(output, "<table cellpadding=0 cellspacing=3 border=0>\n");

		ctxt = xmlXPathNewContextNode(refdoc, token);
		headersobj = xmlXPathEval("child::option/title", ctxt);
		xmlXPathFreeContext(ctxt);

		if (headersobj && (headersobj->type == 1)) {
			fprintf(output, "<tr valign=top>\n");
			for (x = 0; x < headersobj->nodesetval->nodeNr; x++)
				fprintf(output, "<td><p><b>%s:</b></p></td>\n",
						xmlNodeGetContent(headersobj->nodesetval->nodeTab[x]));
			fprintf(output, "</tr>\n\n");
		}
		xmlXPathFreeObject(headersobj);

		if (optionobj && (optionobj->type == 1)) {
			fprintf(output, "<tr valign=top>\n");
			fprintf(output, "<form method=\"POST\" action=\"%s-POST\">\n", action);
			fprintf(output, "<input type=\"hidden\" name=\"editlist\" value=\"%s\">\n\n", listname);
			for (x = 0; x < optionobj->nodesetval->nodeNr; x++) {
				xmlNodePtr node = optionobj->nodesetval->nodeTab[x];
				char defaultval[256];
				strcpy(defaultval, "");

				if (olditem) {
					char *val, *type = xmlGetProp(node, "type");

					if (!strcasecmp(type, "attribute"))
						val = xmlGetProp(olditem, xmlGetProp(node, "name"));
					else
						val = xmlNodeGetContent(olditem);

					snprintf(defaultval, sizeof(defaultval), " value=\"%s\"", val);
				}

				if (!(length = xmlGetProp(node, "length")))
					length = "30";

				fprintf(output, "<td><input name=\"%s\" type=\"text\" size=\"%s\"%s></td>\n",
						xmlGetProp(node, "name"), length, defaultval);
			}
			fprintf(output, "<td><input type=\"submit\" value=\"%s\"></td>\n", _("Insert"));
			fprintf(output, "</form></tr>\n\n");
		}

		ctxt = xmlXPathNewContext(doc);
		snprintf(buf, sizeof(buf), "/child::%s/%s", xmlGetProp(token, "parentpath"), listname);
		listobj = xmlXPathEval(buf, ctxt);
		xmlXPathFreeContext(ctxt);

		if (listobj && (listobj->type == 1)) {
			for (x = 0; x < listobj->nodesetval->nodeNr; x++) {
				int y;
				xmlNodePtr listitem = listobj->nodesetval->nodeTab[x];
				fprintf(output, "<tr valign=top>\n");
				for (y = 0; y < optionobj->nodesetval->nodeNr; y++) {
					xmlNodePtr node = optionobj->nodesetval->nodeTab[y];
					char *type = xmlGetProp(node, "type");

					if (!strcasecmp(type, "attribute"))
						fprintf(output, "<td>%s</td>\n",
								xmlGetProp(listitem, xmlGetProp(node, "name")));
					else
						fprintf(output, "<td>%s</td>\n", xmlNodeGetContent(listitem));
				}
				fprintf(output, "<td><a href=\"%s?editlist=%s&deleteitem=%d\">%s</a>,\n",
						action, listname, x + 1, _("Delete"));
				fprintf(output, "<a href=\"%s?editlist=%s&edititem=%d\">%s</a>,\n",
						action, listname, x + 1, _("Modify"));
				fprintf(output, "<a href=\"%s?editlist=%s&upitem=%d\">%s</a>,\n",
						action, listname, x + 1, _("Up"));
				fprintf(output, "<a href=\"%s?editlist=%s&downitem=%d\">%s</a></td>\n",
						action, listname, x + 1, _("Down"));
				fprintf(output, "</tr>\n\n");
			}
		}
		xmlXPathFreeObject(listobj);

		fprintf(output, "</table>\n\n");
		aucHtmlPopupFoot();
	} else {
		xmlNodePtr listitem;
		listitem = xml_add_node(doc, listname, xmlGetProp(token, "parentpath"), 1);
		if (optionobj && (optionobj->type == 1)) {
			for (x = 0; x < optionobj->nodesetval->nodeNr; x++) {
				xmlNodePtr node = optionobj->nodesetval->nodeTab[x];
				char *type = xmlGetProp(node, "type");
				char *fields[2];

				aucHtmlReadForm(xmlGetProp(node, "name"), fields, 0);
				if (!strcasecmp(type, "attribute"))
					xmlSetProp(listitem, xmlGetProp(node, "name"), fields[0]);
				else
					xmlNodeSetContent(listitem, fields[0]);
			}
		}
		snprintf(buf, sizeof(buf), "%s?editlist=%s", action, listname);
		cgiHeaderLocation(buf);
		retval = 2;
	}

	free(listname);
	free(values[1]);
	free(values[2]);
	free(values[3]);
	free(values[4]);
	xmlFreeDoc(refdoc);
	xmlXPathFreeContext(tokenctxt);
	xmlXPathFreeObject(optionobj);
	xmlXPathFreeObject(tokenobj);

	return retval;
}

void
print_role_matrix(sax_state * docinfo)
{
	xmlXPathContextPtr ctxt;
	xmlXPathObjectPtr obj;
	int x;
	char buf[256];

	if (docinfo->cellcount == 0)
		table_start(docinfo->output, docinfo->vars);

	docinfo->cellcount++;
	row_start(docinfo->output, docinfo->cellcount & 0x1);
	cell_blank(docinfo->output);
	cell_start_long(docinfo->output, 40);
	fprintf(docinfo->output, "<p>Add New Role:\n");
	fprintf(docinfo->output, "<input name=\"roleadd\" type=\"text\" size=\"12\">\n");
	fprintf(docinfo->output, "<input type=\"submit\" value=\"Add Role\"></p>\n");
	cell_end(docinfo->output);
	row_end(docinfo->output);

	snprintf(buf, sizeof(buf), "/child::%s/%s", docinfo->currfield->xmlpath,
			 docinfo->currfield->name);
	ctxt = xmlXPathNewContext(docinfo->xmldoc);
	obj = xmlXPathEval(buf, ctxt);
	if (obj && (obj->type == 1)) {
		aucNodelettListPtr pane;
		int y;

		docinfo->cellcount++;
		row_start(docinfo->output, docinfo->cellcount & 0x1);
		cell_start(docinfo->output);
		fprintf(docinfo->output, "<p align=center><br><br><b>%s</b></p>\n", _("Roles"));
		cell_end(docinfo->output);

		pane = aucNodelettGetList();
		for (y = 0; y < pane->len; y++) {
			cell_start(docinfo->output);
			fprintf(docinfo->output, "<p align=center><font size=-1>%s<br>[%s]</font></p>\n",
					pane->info[y]->title, pane->info[y]->reqpriv);
			cell_end(docinfo->output);
		}

		cell_start_long(docinfo->output, 38 - y);
		fprintf(docinfo->output, "&nbsp;\n");
		cell_end(docinfo->output);

		row_end(docinfo->output);

		for (x = 0; x < obj->nodesetval->nodeNr; x++) {
			xmlNodePtr role = obj->nodesetval->nodeTab[x];
			char *rolename = xmlGetProp(role, "name");

			docinfo->cellcount++;
			row_start(docinfo->output, docinfo->cellcount & 0x1);
			cell_start(docinfo->output);
			fprintf(docinfo->output, "<p align=right>%s:</p>\n", rolename);
			cell_end(docinfo->output);

			for (y = 0; y < pane->len; y++) {
				xmlXPathObjectPtr optionobj;
				xmlXPathContextPtr optionctxt = xmlXPathNewContextNode(docinfo->xmldoc, role);
				int checked = 0;

				cell_start(docinfo->output);

				snprintf(buf, sizeof(buf), "child::pane[@name='%s']",
						 pane->info[y]->handle);
				optionobj = xmlXPathEval(buf, optionctxt);
				if (optionobj && (optionobj->type == 1) && optionobj->nodesetval->nodeNr)
					checked = 1;
				fprintf(docinfo->output,
						"<center><input type=\"checkbox\" name=\"%s_%s_%s\"%s></center>\n",
						docinfo->currfield->name, rolename, pane->info[y]->handle,
						(checked) ? " checked" : "");
				cell_end(docinfo->output);

				xmlXPathFreeObject(optionobj);
				xmlXPathFreeContext(optionctxt);
			}

			cell_start_long(docinfo->output, 38 - y);
			fprintf(docinfo->output, "<input type=\"submit\" name=\"delrole_%s\" value=\"%s\">\n",
					rolename, _("Delete"));
			cell_end(docinfo->output);

			row_end(docinfo->output);
		}
	}

	docinfo->cellcount++;
	row_start(docinfo->output, docinfo->cellcount & 0x1);
	cell_blank(docinfo->output);
	cell_start_long(docinfo->output, 40);
	fprintf(docinfo->output, "<input type=\"submit\" value=\"Save Settings\">\n");
	cell_end(docinfo->output);
	row_end(docinfo->output);

	xmlXPathFreeObject(obj);
	xmlXPathFreeContext(ctxt);
}

void
process_role_matrix(sax_state * docinfo)
{
	xmlXPathContextPtr ctxt;
	xmlXPathObjectPtr obj;
	char *values[2];
	char buf[256];
	int x, y;

	aucHtmlReadForm("roleadd", values, 0);
	if (strlen(values[0])) {
		xmlNodePtr role;
		role = xml_add_node(docinfo->xmldoc, docinfo->currfield->name,
							docinfo->currfield->xmlpath, 1);
		xmlSetProp(role, "name", values[0]);
	}

	snprintf(buf, sizeof(buf), "/child::%s/%s", docinfo->currfield->xmlpath,
			 docinfo->currfield->name);
	ctxt = xmlXPathNewContext(docinfo->xmldoc);
	obj = xmlXPathEval(buf, ctxt);
	if (obj && (obj->type == 1)) {
		aucNodelettListPtr pane;

		for (x = 0; x < obj->nodesetval->nodeNr; x++) {
			xmlNodePtr role = obj->nodesetval->nodeTab[x];
			char *rolename = xmlGetProp(role, "name");

			snprintf(buf, sizeof(buf), "delrole_%s", rolename);
			if (cgiFormSuccess == cgiFormCheckboxSingle(buf)) {
				snprintf(buf, sizeof(buf), "%s[@name='%s']",
						 docinfo->currfield->name, rolename);
				xml_del_node(docinfo->xmldoc, buf, docinfo->currfield->xmlpath);
				continue;
			}

			snprintf(buf, sizeof(buf), "%s[@name='%s']/pane",
					 docinfo->currfield->name, rolename);
			xml_del_node(docinfo->xmldoc, buf, docinfo->currfield->xmlpath);
			pane = aucNodelettGetList();
			for (y = 0; y < pane->len; y++) {
				snprintf(buf, sizeof(buf), "%s_%s_%s", docinfo->currfield->name,
						 rolename, pane->info[y]->handle);
				if (cgiFormSuccess == cgiFormCheckboxSingle(buf)) {
					xmlNodePtr panenode;

					snprintf(buf, sizeof(buf), "%s/%s[@name='%s']",
							 docinfo->currfield->xmlpath, docinfo->currfield->name, rolename);
					panenode = xml_add_node(docinfo->xmldoc, "pane", buf, 1);
					xmlSetProp(panenode, "name", pane->info[y]->handle);
				}
			}

		}
	}
	xmlXPathFreeObject(obj);
	xmlXPathFreeContext(ctxt);
}

void
start_el(sax_state * docinfo, xmlChar * name, xmlChar ** attrs)
{
	if (docinfo->state == IGNORE) {
		docinfo->ignore_depth++;
		return;
	}

	if (!xmlStrcmp(name, "if")) {
		if (!test_form_conditional(docinfo, attrs)) {
			docinfo->saved_state = docinfo->state;
			docinfo->state = IGNORE;
			docinfo->ignore_depth = 0;
		}
	} else if (!xmlStrcmp(name, "ConfigToken")) {
		initialize_field(&(docinfo->field), attrs, docinfo->vars);
		docinfo->currfield = &(docinfo->field);
		docinfo->state = IN_TOKEN;
		docinfo->heading_depth++;
		if (docinfo->xmldoc && docinfo->currfield->storexml)
			xml_token_default(docinfo->currfield, docinfo->xmldoc);
		if (docinfo->currfield->type == ROLEMATRIX)
			print_role_matrix(docinfo);
	} else if (!xmlStrcmp(name, "title")) {
		if (docinfo->state == IN_OPTION) {
			docinfo->state = IN_OPTION_TITLE;
		} else {
			char *tmp;
			docinfo->state = IN_TITLE;
			if (docinfo->heading_depth == 0)
				fprintf(docinfo->output, "<h2>");
			else if (docinfo->currfield->printable) {
				if (docinfo->cellcount == 0)
					table_start(docinfo->output, docinfo->vars);
				docinfo->cellcount++;
				docinfo->currfield->ptitle = 1;
				row_start(docinfo->output, docinfo->cellcount & 0x1);
				cell_start(docinfo->output);
				fprintf(docinfo->output, "<p align=right>");
			}
			if ((tmp = get_token((aucVarField *) attrs, "xpath")))
				fprintf(docinfo->output, "%s", xml_xpath_value(tmp, docinfo->xmldoc));
		}
	} else if (!xmlStrcmp(name, "description")) {
		if (docinfo->heading_depth == 0)
			fprintf(docinfo->output, "<p>\n");
		else if ((docinfo->state == IN_TOKEN) && docinfo->currfield->printable) {
			if (docinfo->currfield->ptitle == 0) {
				docinfo->cellcount++;
				row_start(docinfo->output, docinfo->cellcount & 0x1);
				cell_blank(docinfo->output);
				docinfo->currfield->ptitle = 1;
			}
			if (docinfo->currfield->pfield == 0) {
				cell_field_start(docinfo->output, docinfo->currfield, docinfo->xmldoc);
				docinfo->currfield->pfield = 1;
			}
			if (docinfo->currfield->pfield == 1) {
				cell_field_end(docinfo->output, docinfo->currfield);
				docinfo->currfield->pfield = 2;
			}
			docinfo->currfield->pdesc = 1;
			cell_start_long(docinfo->output, 38);
			fprintf(docinfo->output, "<p><small>\n");
		}
		docinfo->state = IN_DESC;
	} else if (!xmlStrcmp(name, "option")) {
		if (docinfo->currfield->printable) {
			if (docinfo->currfield->ptitle == 0) {
				docinfo->cellcount++;
				row_start(docinfo->output, docinfo->cellcount & 0x1);
				cell_blank(docinfo->output);
				docinfo->currfield->ptitle = 1;
			}
			if (docinfo->currfield->pfield == 0) {
				cell_field_start(docinfo->output, docinfo->currfield, docinfo->xmldoc);
				docinfo->currfield->pfield = 1;
			}
			cell_field_start_option(docinfo->output, docinfo->currfield, attrs, docinfo->xmldoc);
		}
		docinfo->state = IN_OPTION;
	} else if (!xmlStrcmp(name, "attribute")) {
		end_field_row(docinfo);
		initialize_field(&(docinfo->field_attr), attrs, docinfo->vars);
		docinfo->currfield = &(docinfo->field_attr);
		docinfo->is_attribute = 1;
		if (docinfo->xmldoc)
			xml_tokenattr_default(docinfo->currfield, docinfo->field.xmlpath,
								  docinfo->field.name, docinfo->xmldoc);
	} else {
	}
}

void
end_el(sax_state * docinfo, xmlChar * name)
{
	if (docinfo->state == IGNORE) {
		if (docinfo->ignore_depth == 0)
			docinfo->state = docinfo->saved_state;
		else
			docinfo->ignore_depth--;
		return;
	}

	if (!xmlStrcmp(name, "ConfigToken")) {
		end_field_row(docinfo);
		docinfo->heading_depth--;
		if (docinfo->heading_depth == 0)
			docinfo->state = START_FORM;
	} else if (!xmlStrcmp(name, "title")) {
		if (docinfo->state == IN_OPTION_TITLE) {
			docinfo->state = IN_OPTION;
		} else {
			if (docinfo->heading_depth == 0) {
				docinfo->state = START_FORM;
				fprintf(docinfo->output, "</h2>\n\n");
			} else {
				docinfo->state = IN_TOKEN;
				if (docinfo->currfield->printable) {
					fprintf(docinfo->output, ":<br>\n");
					if (strlen(docinfo->currfield->link) && (docinfo->currfield->type == LIST))
						fprintf(docinfo->output,
								"<i><a href=\"javascript: newWindow('%s?editlist=%s','editlist','500')\">%s</a></i>\n",
								docinfo->currfield->link, docinfo->currfield->name, _("Customize"));
					fprintf(docinfo->output, "</p>\n\n");
					cell_end(docinfo->output);
				}
			}
		}
	} else if (!xmlStrcmp(name, "description")) {
		if (docinfo->heading_depth == 0) {
			docinfo->state = START_FORM;
			fprintf(docinfo->output, "</p>\n");
		} else {
			docinfo->state = IN_TOKEN;
			if (docinfo->currfield->printable) {
				fprintf(docinfo->output, "</small></p>\n");
				cell_end(docinfo->output);
			}
		}
	} else if (!xmlStrcmp(name, "option")) {
		if (docinfo->currfield->printable)
			cell_field_end_option(docinfo->output, docinfo->currfield);
		docinfo->state = IN_TOKEN;
	} else if (!xmlStrcmp(name, "attribute")) {
		end_field_row(docinfo);
		docinfo->currfield = &(docinfo->field);
		docinfo->is_attribute = 0;
	} else {
	}
}

void
get_chars(sax_state * docinfo, xmlChar * str, int len)
{
	switch (docinfo->state) {
	case IN_TITLE:
	case IN_DESC:
		fprintf(docinfo->output, "%.*s", len, str);
		break;
	case IN_OPTION_TITLE:
		if (docinfo->currfield->printable && (docinfo->currfield->type != LIST))
			fprintf(docinfo->output, "%.*s", len, str);
		break;
	default:
		break;
	}
}

void
start_el_w(sax_state * docinfo, xmlChar * name, xmlChar ** attrs)
{
	if (docinfo->state == IGNORE) {
		docinfo->ignore_depth++;
		return;
	}

	if (!xmlStrcmp(name, "if")) {
		if (!test_form_conditional(docinfo, attrs)) {
			docinfo->saved_state = docinfo->state;
			docinfo->state = IGNORE;
			docinfo->ignore_depth = 0;
		}
	} else if (!xmlStrcmp(name, "ConfigToken")) {
		initialize_field(&(docinfo->field), attrs, docinfo->vars);
		docinfo->currfield = &(docinfo->field);
		docinfo->state = IN_TOKEN;
		docinfo->heading_depth++;
		if (docinfo->currfield->type == ROLEMATRIX)
			process_role_matrix(docinfo);
		else {
			if (docinfo->currfield->storexml)
				xml_process_formtoken(docinfo->xmldoc, docinfo->currfield);
			if (docinfo->currfield->type != CHECKBOX)
				store_text_var(docinfo->vars, docinfo->currfield);
		}
	} else if (!xmlStrcmp(name, "option")) {
		if (docinfo->currfield->type == CHECKBOX) {
			char *tmp;
			if ((tmp = get_token((aucVarField *) attrs, "method")))
				process_option_method(docinfo, tmp);
			else {
				if (docinfo->currfield->storexml)
					xml_cond_process_formtoken(docinfo->xmldoc, docinfo->currfield,
											   get_token((aucVarField *) attrs, "name"));
				store_boolean_var(docinfo->vars, docinfo->currfield,
								  get_token((aucVarField *) attrs, "name"));
			}
		}
	} else if (!xmlStrcmp(name, "attribute")) {
		initialize_field(&(docinfo->field_attr), attrs, docinfo->vars);
		docinfo->currfield = &(docinfo->field_attr);
		docinfo->is_attribute = 1;
		xml_process_formattr(docinfo->xmldoc, docinfo->field.xmlpath,
							 docinfo->field.name, docinfo->currfield);
		store_text_var(docinfo->vars, docinfo->currfield);
	}
}

void
end_el_w(sax_state * docinfo, xmlChar * name)
{
	if (docinfo->state == IGNORE) {
		if (docinfo->ignore_depth == 0)
			docinfo->state = docinfo->saved_state;
		else
			docinfo->ignore_depth--;
		return;
	}

	if (!xmlStrcmp(name, "ConfigToken")) {
		docinfo->heading_depth--;
		if (docinfo->heading_depth == 0)
			docinfo->state = START_FORM;
	} else if (!xmlStrcmp(name, "attribute")) {
		docinfo->currfield = &(docinfo->field);
		docinfo->is_attribute = 0;
	}
}

void
start_doc(sax_state * docinfo)
{
	char *tmp;

	tmp = get_token(docinfo->vars, "form_action");
	fprintf(docinfo->output, "<form method=\"POST\" action=\"%s\">\n\n", (tmp) ? tmp : "");

	docinfo->state = START_FORM;
	docinfo->ignore_depth = 0;
	docinfo->saved_state = 0;
	docinfo->heading_depth = 0;
	docinfo->is_attribute = 0;
	docinfo->cellcount = 0;
}

void
end_doc(sax_state * docinfo)
{
	table_end(docinfo->output);
	docinfo->state = FINISH_FORM;
}

void
start_doc_w(sax_state * docinfo)
{
	docinfo->state = START_FORM;
	docinfo->ignore_depth = 0;
	docinfo->saved_state = 0;
	docinfo->heading_depth = 0;
	docinfo->is_attribute = 0;
	docinfo->cellcount = 0;
}

void
end_doc_w(sax_state * docinfo)
{
	docinfo->state = FINISH_FORM;
}

char *
get_form_input(fieldInfo * field)
{
	char *values[2];

	switch (field->handler) {
	case FIXED:
		values[0] = strdup(field->defaultvalue);
		break;
	case GETVERSION:
		values[0] = strdup(DB_VERSION);
		break;
	default:
		aucHtmlReadForm(field->name, values, 0);
		break;
	}

	return values[0];
}

void
xml_process_formattr(xmlDocPtr doc, char *ppath, char *pname, fieldInfo * attrfield)
{
	char *value;
	xmlNodePtr tree;

	tree = xml_add_node(doc, pname, ppath, 0);
	value = get_form_input(attrfield);
	
	xmlSetProp(tree, attrfield->name, value);
	
	free(value);
}

void
xml_process_formtoken(xmlDocPtr doc, fieldInfo * field)
{
	char *value;
	xmlNodePtr tree;

	switch (field->type) {
	case CHECKBOX:
		xml_del_node(doc, field->name, field->xmlpath);
		break;
	case LIST:
		break;
	default:
		tree = xml_add_node(doc, field->name, field->xmlpath, 0);

		value = get_form_input(field);

		if (field->hascontent)
			xmlNodeSetContent(tree, value);

		free(value);
		break;
	}
}

void
xml_cond_process_formtoken(xmlDocPtr doc, fieldInfo * field, char *name)
{
	xmlNodePtr tree;

	if (cgiFormSuccess == cgiFormCheckboxSingle(name)) {
		tree = xml_add_node(doc, field->name, field->xmlpath, 1);
		xmlNodeSetContent(tree, name);
	}
}

void
store_text_var(aucVarField * vars, fieldInfo * field)
{
	char *value;

	if (!strlen(field->name))
		return;
	value = get_form_input(field);
	insert_var(vars, field->name, value);
}

void
store_boolean_var(aucVarField * vars, fieldInfo * field, char *name)
{
	if (!field->printable) {
		insert_var(vars, name, "true");
		return;
	}

	if (cgiFormSuccess == cgiFormCheckboxSingle(name))
		insert_var(vars, name, "true");
	else
		insert_var(vars, name, "false");
}

xmlNodePtr
xml_add_node(xmlDocPtr doc, char *name, char *path, int allowdups)
{
	xmlXPathContextPtr ctxt;
	xmlXPathObjectPtr obj;
	xmlNodePtr retval, par;
	char buf[1024];
	char *marker;


	if (xmlDocGetRootElement(doc) && !allowdups) {
		ctxt = xmlXPathNewContext(doc);
		snprintf(buf, sizeof(buf), "/child::%s%s%s", path, (strlen(path)) ? "/" : "", name);
		obj = xmlXPathEval(buf, ctxt);
		xmlXPathFreeContext(ctxt);
		if (obj && (obj->type == 1) && (obj->nodesetval->nodeNr > 0)) {
			retval = obj->nodesetval->nodeTab[0];
			xmlXPathFreeObject(obj);
			return retval;
		}
		xmlXPathFreeObject(obj);
	}

	if (strlen(path)) {
		if (xmlDocGetRootElement(doc)) {
			ctxt = xmlXPathNewContext(doc);
			snprintf(buf, sizeof(buf), "/child::%s", path);
			obj = xmlXPathEval(buf, ctxt);
			xmlXPathFreeContext(ctxt);
			if (obj && (obj->type == 1) && (obj->nodesetval->nodeNr > 0)) {
				retval = xmlNewChild(obj->nodesetval->nodeTab[0], NULL, name, NULL);
				xmlXPathFreeObject(obj);
				return retval;
			}
			xmlXPathFreeObject(obj);
		}

		if ((marker = strrchr(path, '/'))) {
			strncpy(buf, path, marker - path);
			buf[marker - path] = '\0';
			par = xml_add_node(doc, marker + 1, buf, 0);
		} else
			par = xml_add_node(doc, path, "", 0);

		retval = xmlNewChild(par, NULL, name, NULL);
		return retval;
	} else
		return (doc->children = xmlNewDocNode(doc, NULL, name, NULL));
}

xmlNodePtr
xml_find_node(xmlDocPtr doc, char *name, char *path)
{
	xmlXPathContextPtr ctxt;
	xmlXPathObjectPtr obj;
	char buf[1024];
	xmlNodePtr retval = NULL;

	ctxt = xmlXPathNewContext(doc);
	snprintf(buf, sizeof(buf), "/child::%s%s%s", path, (strlen(path)) ? "/" : "", name);
	obj = xmlXPathEval(buf, ctxt);
	if (obj && (obj->type == 1) && obj->nodesetval->nodeNr)
		retval = obj->nodesetval->nodeTab[0];

	xmlXPathFreeContext(ctxt);
	xmlXPathFreeObject(obj);
	return retval;
}

void
xml_del_node(xmlDocPtr doc, char *name, char *path)
{
	xmlXPathContextPtr ctxt;
	xmlXPathObjectPtr obj;
	char buf[1024];
	int x;

	ctxt = xmlXPathNewContext(doc);
	snprintf(buf, sizeof(buf), "/child::%s%s%s", path, (strlen(path)) ? "/" : "", name);
	obj = xmlXPathEval(buf, ctxt);
	if (obj && (obj->type == 1)) {
		for (x = 0; x < obj->nodesetval->nodeNr; x++) {
			xmlNodePtr node = obj->nodesetval->nodeTab[x];
			xmlUnlinkNode(node);
			xmlFreeNode(node);
		}
	}
	xmlXPathFreeContext(ctxt);
	xmlXPathFreeObject(obj);
}

void
xml_print_node(xmlDocPtr doc, char *name, char *path, FILE * output)
{
	xmlXPathContextPtr ctxt;
	xmlXPathObjectPtr obj;
	char buf[1024];
	int x;

	ctxt = xmlXPathNewContext(doc);
	snprintf(buf, sizeof(buf), "/child::%s%s%s", path, (strlen(path)) ? "/" : "", name);
	obj = xmlXPathEval(buf, ctxt);
	if (obj && (obj->type == 1)) {
		fprintf(output, "<table cellpadding=0 cellspacing=5 border=0><tr valign=top><td><small>\n");
		for (x = 0; x < obj->nodesetval->nodeNr; x++) {
			xmlNodePtr node = obj->nodesetval->nodeTab[x];
			if (x == (obj->nodesetval->nodeNr / 2 + 1))
				fprintf(output, "</small></td><td><small>\n");
			fprintf(output, "&nbsp; %s<br>\n", xmlNodeGetContent(node));
		}
		fprintf(output, "</small></td></tr></table>\n\n");
	}
	xmlXPathFreeContext(ctxt);
	xmlXPathFreeObject(obj);
}

void
xml_print_path_options(xmlDocPtr doc, char *path, FILE * output, char *defaultval)
{
	xmlXPathContextPtr ctxt;
	xmlXPathObjectPtr obj;
	int x;
	char *tmp;

	ctxt = xmlXPathNewContext(doc);
	obj = xmlXPathEval(path, ctxt);
	if (obj && (obj->type == 1)) {
		for (x = 0; x < obj->nodesetval->nodeNr; x++) {
			xmlNodePtr node = obj->nodesetval->nodeTab[x];
			tmp = xmlNodeGetContent(node);
			if (x == (obj->nodesetval->nodeNr - 1))
				fprintf(output, "<option%s>%s", (!strcmp(tmp, defaultval)) ? " selected" : "", tmp);
			else
				fprintf(output, "<option%s>%s</option>\n",
						(!strcmp(tmp, defaultval)) ? " selected" : "", tmp);
		}
	}
	xmlXPathFreeContext(ctxt);
	xmlXPathFreeObject(obj);
}

void
insert_var(aucVarField * vars, char *name, char *value)
{
	int x;

	if (!name || !strlen(name))
		return;

	for (x = 0; vars[x].name; x++) {
		if (!strcmp(vars[x].name, name)) {
			vars[x].value = strdup(value);
			return;
		}
	}

	vars[x].name = strdup(name);
	vars[x].value = strdup(value);
	vars[x + 1].name = NULL;
	vars[x + 1].value = NULL;

	return;
}

int
test_form_conditional(sax_state * docinfo, xmlChar ** attrs)
{
	char *tmp, *testval = NULL, *value;
	int comparison = 0;

	if ((tmp = get_token((aucVarField *) attrs, "variable")))
		testval = get_token(docinfo->vars, tmp);
	else if ((tmp = get_token((aucVarField *) attrs, "xpath")))
		testval = xml_xpath_value(tmp, docinfo->xmldoc);

	if (!testval)
		return 0;

	if ((value = get_token((aucVarField *) attrs, "isvalue")))
		comparison = !strcmp(testval, value);
	else if ((value = get_token((aucVarField *) attrs, "isnotvalue")))
		comparison = strcmp(testval, value);
	return comparison;
}

char *
get_token(aucVarField * vars, char *name)
{
	int x;

	if (!vars)
		return NULL;

	for (x = 0; vars[x].name; x++) {
		if (!strcmp(vars[x].name, name))
			return vars[x].value;
	}
	return NULL;
}

xmlEntityPtr
get_entity(sax_state * docinfo, xmlChar * name)
{
	return xmlGetPredefinedEntity(name);
}

void
p_warning(sax_state * docinfo, char *msg, ...)
{
	va_list args;
	char buf[1024];

	va_start(args, msg);
	vsnprintf(buf, sizeof(buf), msg, args);
	va_end(args);
	fprintf(docinfo->output, "<p><b>Warning:</b> %s<br>\n", buf);
	fprintf(docinfo->output, "Please report this error to the system administrator.</p>\n\n");
}

void
p_error(sax_state * docinfo, char *msg, ...)
{
	va_list args;
	char buf[1024];

	va_start(args, msg);
	vsnprintf(buf, sizeof(buf), msg, args);
	va_end(args);
	fprintf(docinfo->output, "<p><b>Error:</b> %s<br>\n", buf);
	fprintf(docinfo->output, "Please report this error to the system administrator.</p>\n\n");
}

void
p_fatal_error(sax_state * docinfo, char *msg, ...)
{
	va_list args;
	char buf[1024];

	va_start(args, msg);
	vsnprintf(buf, sizeof(buf), msg, args);
	va_end(args);
	fprintf(docinfo->output, "<p><b>Fatal Error:</b> %s<br>\n", buf);
	fprintf(docinfo->output, "Please report this error to the system administrator.</p>\n\n");
}

void
end_field_row(sax_state * docinfo)
{
	if (docinfo->currfield->printable) {
		if (docinfo->currfield->ptitle == 0) {
			docinfo->cellcount++;
			row_start(docinfo->output, docinfo->cellcount & 0x1);
			cell_blank(docinfo->output);
			docinfo->currfield->ptitle = 1;
		}
		if (docinfo->currfield->pfield == 0) {
			cell_field_start(docinfo->output, docinfo->currfield, docinfo->xmldoc);
			docinfo->currfield->pfield = 1;
		}
		if (docinfo->currfield->pfield == 1) {
			cell_field_end(docinfo->output, docinfo->currfield);
			docinfo->currfield->pfield = 2;
		}
		if (docinfo->currfield->pdesc == 0) {
			cell_start_long(docinfo->output, 38);
			fprintf(docinfo->output, "&nbsp;\n");
			cell_end(docinfo->output);
			docinfo->currfield->pdesc = 1;
		}
		row_end(docinfo->output);
		docinfo->currfield->printable = 0;
	} else if (docinfo->currfield->type == HIDDEN)
		fprintf(docinfo->output, "<input type=\"hidden\" name=\"%s\" value=\"%s\">\n\n",
				docinfo->currfield->name, docinfo->currfield->defaultvalue);
}

void
xml_tokenattr_default(fieldInfo * field, char *ppath, char *pname, xmlDocPtr doc)
{
	xmlXPathContextPtr ctxt;
	xmlXPathObjectPtr obj;
	char buf[1024], *tmp;

	if (xmlDocGetRootElement(doc) && (field->handler == INLINE)) {
		ctxt = xmlXPathNewContext(doc);
		snprintf(buf, sizeof(buf), "/child::%s%s%s", ppath, (strlen(ppath)) ? "/" : "", pname);
		obj = xmlXPathEval(buf, ctxt);

		if (obj && (obj->type == 1) && (obj->nodesetval->nodeNr > 0))
			if ((tmp = xmlGetProp(obj->nodesetval->nodeTab[0], field->name)))
				snprintf(field->defaultvalue, sizeof(field->defaultvalue), "%s", tmp);

		xmlXPathFreeContext(ctxt);
		xmlXPathFreeObject(obj);
	}
}

void
xml_token_default(fieldInfo * field, xmlDocPtr doc)
{
	xmlXPathContextPtr ctxt;
	xmlXPathObjectPtr obj;
	char buf[1024], *tmp;

	if (xmlDocGetRootElement(doc) && (field->handler == INLINE)) {
		ctxt = xmlXPathNewContext(doc);
		snprintf(buf, sizeof(buf), "/child::%s%s%s",
				 field->xmlpath, (strlen(field->xmlpath)) ? "/" : "", field->name);
		obj = xmlXPathEval(buf, ctxt);

		if (obj && (obj->type == 1) && (obj->nodesetval->nodeNr > 0)) {
			if ((tmp = xmlNodeGetContent(obj->nodesetval->nodeTab[0]))) {
				snprintf(field->defaultvalue, sizeof(field->defaultvalue), "%s", tmp);
			} else
				strcpy(field->defaultvalue, "");
			field->xmlnode = obj->nodesetval->nodeTab[0];
		}

		xmlXPathFreeContext(ctxt);
		xmlXPathFreeObject(obj);
	}
}

int
xml_token_exist(fieldInfo * field, char *content, xmlDocPtr doc)
{
	xmlXPathContextPtr ctxt;
	xmlXPathObjectPtr obj;
	char buf[1024];
	int retval = 0, x;

	if (xmlDocGetRootElement(doc) && (field->handler == INLINE)) {
		ctxt = xmlXPathNewContext(doc);
		snprintf(buf, sizeof(buf), "/child::%s%s%s",
				 field->xmlpath, (strlen(field->xmlpath)) ? "/" : "", field->name);
		obj = xmlXPathEval(buf, ctxt);

		if (obj && (obj->type == 1)) {
			if (content)
				for (x = 0; x < obj->nodesetval->nodeNr; x++) {
					if (!strcmp(xmlNodeGetContent(obj->nodesetval->nodeTab[x]), content))
						retval = 1;
			} else if (obj->nodesetval->nodeNr > 0)
				retval = 1;
		}

		xmlXPathFreeContext(ctxt);
		xmlXPathFreeObject(obj);
	}
	return retval;
}

char *
xml_get_content(xmlNodePtr node)
{
	if (!node)
		return NULL;

	switch (node->type) {
	case XML_ELEMENT_NODE:
		return xmlNodeGetContent(node);
		break;
	case XML_ATTRIBUTE_NODE:
		return ((xmlAttrPtr) node)->children->content;
		break;
	case XML_TEXT_NODE:
		return node->content;
		break;
	default:
		return NULL;
		break;
	}
	return NULL;
}

char *
xml_xpath_value(char *xpath, xmlDocPtr doc)
{
	xmlXPathContextPtr ctxt;
	xmlXPathObjectPtr obj;
	char *retval = NULL;

	if (xmlDocGetRootElement(doc)) {
		ctxt = xmlXPathNewContext(doc);
		obj = xmlXPathEval(xpath, ctxt);

		if (obj && (obj->type == 1) && (obj->nodesetval->nodeNr > 0))
			retval = xml_get_content(obj->nodesetval->nodeTab[0]);

		xmlXPathFreeContext(ctxt);
		xmlXPathFreeObject(obj);
	}
	return retval;
}

void
process_option_method(sax_state * docinfo, char *method)
{
	if (!strcmp(method, "listplugins")) {
		aucPluginInfoPtr *plugins;
		plugins = aucPluginQueryAll(aucDataDir);
		if (!plugins)
			return;
		for (; *plugins; plugins++) {
			if (docinfo->currfield->storexml)
				xml_cond_process_formtoken(docinfo->xmldoc, docinfo->currfield, (*plugins)->name);
			store_boolean_var(docinfo->vars, docinfo->currfield, (*plugins)->name);
		}
	}
}

void
initialize_field(fieldInfo * field, xmlChar ** attrs, aucVarField * vars)
{
	char *tmp;

	field->ptitle = field->pfield = field->pdesc = field->printable = 0;
	strcpy(field->name, "");
	field->handler = FIXED;
	field->type = TEXT;
	field->length = 30;
	field->width = 50;
	field->height = 5;
	field->maxlen = 0;
	field->hascontent = 0;
	field->storexml = 0;
	field->xmlnode = NULL;
	strcpy(field->xmlpath, "");
	strcpy(field->defaultvalue, "");
	strcpy(field->link, "");

	if ((tmp = get_token((aucVarField *) attrs, "name")))
		snprintf(field->name, sizeof(field->name), "%s", tmp);
	if ((tmp = get_token((aucVarField *) attrs, "handler"))) {
		field->handler = str2handler(tmp);
		field->hascontent = 1;
	}
	if ((tmp = get_token((aucVarField *) attrs, "type"))) {
		field->type = str2type(tmp);
		field->hascontent = 1;
	}
	if ((tmp = get_token((aucVarField *) attrs, "length")))
		field->length = strtoul(tmp, NULL, 10);
	if ((tmp = get_token((aucVarField *) attrs, "width")))
		field->width = strtoul(tmp, NULL, 10);
	if ((tmp = get_token((aucVarField *) attrs, "height")))
		field->height = strtoul(tmp, NULL, 10);
	if ((tmp = get_token((aucVarField *) attrs, "maxlen")))
		field->maxlen = strtoul(tmp, NULL, 10);
	if ((tmp = get_token((aucVarField *) attrs, "link")))
		snprintf(field->link, sizeof(field->link), "%s", tmp);
	if ((tmp = get_token((aucVarField *) attrs, "parentpath"))) {
		snprintf(field->xmlpath, sizeof(field->xmlpath), "%s", tmp);
		field->storexml = 1;
	}
	if ((tmp = get_token((aucVarField *) attrs, "default")))
		snprintf(field->defaultvalue, sizeof(field->defaultvalue), "%s", tmp);

	if ((field->handler == INLINE) && (field->type != HIDDEN) && (field->type != ROLEMATRIX))
		field->printable = 1;

	if ((tmp = get_token((aucVarField *) attrs, "value")))
		if ((tmp = get_token(vars, tmp)))
			snprintf(field->defaultvalue, sizeof(field->defaultvalue), "%s", tmp);
	if ((tmp = get_token(vars, field->name)))
		snprintf(field->defaultvalue, sizeof(field->defaultvalue), "%s", tmp);
}

handlerEnum
str2handler(char *str)
{
	if (!strcasecmp(str, "fixed"))
		return FIXED;
	if (!strcasecmp(str, "inline"))
		return INLINE;
	if (!strcasecmp(str, "getversion"))
		return GETVERSION;
	return FIXED;
}

typeEnum
str2type(char *str)
{
	if (!strcasecmp(str, "text"))
		return TEXT;
	if (!strcasecmp(str, "textarea"))
		return TEXTAREA;
	if (!strcasecmp(str, "radio"))
		return RADIO;
	if (!strcasecmp(str, "checkbox"))
		return CHECKBOX;
	if (!strcasecmp(str, "dropdown"))
		return DROPDOWN;
	if (!strcasecmp(str, "list"))
		return LIST;
	if (!strcasecmp(str, "password"))
		return PASSWORD;
	if (!strcasecmp(str, "submit"))
		return SUBMIT;
	if (!strcasecmp(str, "label"))
		return LABEL;
	if (!strcasecmp(str, "hidden"))
		return HIDDEN;
	if (!strcasecmp(str, "rolematrix"))
		return ROLEMATRIX;
	return TEXT;
}

void
table_start(FILE * output, aucVarField * vars)
{
	fprintf(output, "<table cellpadding=4 cellspacing=0 border=0 width=100%%>\n");
	fprintf(output, "<tr bgcolor=#%s><td colspan=40>", aucTrimColor);
	fprintf(output, "<img src=\"%s/transpix.gif\" alt=\"\" width=1 height=1>", aucGraphDir);
	fprintf(output, "</td></tr>\n\n");
}

void
table_end(FILE * output)
{
	fprintf(output, "<tr bgcolor=#%s><td colspan=40>", aucTrimColor);
	fprintf(output, "<img src=\"%s/transpix.gif\" alt=\"\" width=1 height=1>", aucGraphDir);
	fprintf(output, "</td></tr>\n\n");
	fprintf(output, "</form></table>\n\n");
}

void
row_start(FILE * output, int shaded)
{
	fprintf(output, "<tr valign=middle bgcolor=#%s>\n\n", (shaded) ? "EEEEEE" : "FFFFFF");
}

void
row_end(FILE * output)
{
	fprintf(output, "</tr>\n\n");
}

void
cell_start(FILE * output)
{
	fprintf(output, "<td>\n");
}

void
cell_start_long(FILE * output, int span)
{
	fprintf(output, "<td colspan=%d>\n", span);
}

void
cell_end(FILE * output)
{
	fprintf(output, "</td>\n\n");
}

void
cell_blank(FILE * output)
{
	cell_start(output);
	fprintf(output, "&nbsp;\n");
	cell_end(output);
}

void
cell_field_start(FILE * output, fieldInfo * field, xmlDocPtr doc)
{
	char extrafields[128], buf[128];

	strcpy(extrafields, "");

	if (field->type != TEXTAREA)
		cell_start(output);

	switch (field->type) {
	case TEXT:
	case PASSWORD:
		if (field->maxlen) {
			sprintf(buf, " maxlength=\"%d\"", field->maxlen);
			strcat(extrafields, buf);
		}
		fprintf(output, "<input type=\"%s\" name=\"%s\" size=\"%d\" value=\"%s\"%s>\n",
				(field->type == PASSWORD) ? "password" : "text",
				field->name, field->length, field->defaultvalue, extrafields);
		break;
	case DROPDOWN:
		fprintf(output, "<select name=\"%s\">\n", field->name);
		break;
	case LIST:
		fprintf(output, "<p>\n");
		if (field->storexml)
			xml_print_node(doc, field->name, field->xmlpath, output);
		fprintf(output, "</p>\n\n");
		break;
	case SUBMIT:
		if (strlen(field->name)) {
			sprintf(buf, " name=\"%s\"", field->name);
			strcat(extrafields, buf);
		}
		fprintf(output, "<input type=\"submit\" value=\"%s\"%s>\n\n",
				field->defaultvalue, extrafields);
		break;
	case TEXTAREA:
		cell_start_long(output, 39);
		fprintf(output, "<textarea name=\"%s\" cols=%d rows=%d wrap=soft>\n",
				field->name, field->width, field->height);
		fprintf(output, "%s", field->defaultvalue);
		fprintf(output, "</textarea>\n\n");
		field->pdesc = 1;
		break;
	default:
		break;
	}
}

void
cell_field_end(FILE * output, fieldInfo * field)
{
	switch (field->type) {
	case DROPDOWN:
		fprintf(output, "</select>\n");
		break;
	default:
		break;
	}
	cell_end(output);
}

void
cell_field_start_option(FILE * output, fieldInfo * field, xmlChar ** attrs, xmlDocPtr doc)
{
	char *tmp;
	int checked = 0;

	switch (field->type) {
	case CHECKBOX:
		if ((tmp = get_token((aucVarField *) attrs, "name"))) {
			if (doc && field->storexml)
				checked = xml_token_exist(field, tmp, doc);
			fprintf(output, "<input type=\"checkbox\" name=\"%s\"%s>\n",
					tmp, (checked) ? " checked" : "");
		} else if ((tmp = get_token((aucVarField *) attrs, "method"))) {
			if (!strcmp(tmp, "listplugins")) {
				aucPluginInfoPtr *plugins;
				plugins = aucPluginQueryAll(aucDataDir);
				if (!plugins) {
					fprintf(output, "<i>%s</i>\n\n", _("None available"));
					break;
				}
				for (; *plugins; plugins++) {
					checked = 0;
					if (doc && field->storexml)
						checked = xml_token_exist(field, (*plugins)->name, doc);
					fprintf(output, "<input type=\"checkbox\" name=\"%s\"%s>\n",
							(*plugins)->name, (checked) ? " checked" : "");
					fprintf(output, "%s\n", (*plugins)->description);
					if (*(plugins + 1))
						fprintf(output, "<br>\n");
				}
			}
		}
		break;
	case DROPDOWN:
		if ((tmp = get_token((aucVarField *) attrs, "name")))
			fprintf(output, "<option value=\"%s\"%s>", tmp,
					(!strcmp(tmp, field->defaultvalue)) ? " selected" : "");
		else if ((tmp = get_token((aucVarField *) attrs, "method"))) {
			if (!strcmp(tmp, "xpath")) {
				if ((tmp = get_token((aucVarField *) attrs, "path")))
					xml_print_path_options(doc, tmp, output, field->defaultvalue);
			} else if (!strcmp(tmp, "count")) {
				int start = 0, end = 1, x;
				if ((tmp = get_token((aucVarField *) attrs, "end"))) {
					if (strlen(tmp) > 5)
						tmp = xml_xpath_value(tmp, doc);
					end = strtoul(tmp, NULL, 10);
				}
				if ((tmp = get_token((aucVarField *) attrs, "start"))) {
					if (strlen(tmp) > 5)
						tmp = xml_xpath_value(tmp, doc);
					start = strtoul(tmp, NULL, 10);
				}
				for (x = start; x <= end; x++) {
					if (x == end) {
						if (x == strtol(field->defaultvalue, NULL, 10))
							fprintf(output, "<option selected>%d", x);
						else
							fprintf(output, "<option>%d", x);
					} else {
						if (x == strtol(field->defaultvalue, NULL, 10))
							fprintf(output, "<option selected>%d</option>\n", x);
						else
							fprintf(output, "<option>%d</option>\n", x);
					}
				}
			}
		} else
			fprintf(output, "<option>");
		break;
	case RADIO:
		if ((tmp = get_token((aucVarField *) attrs, "name"))) {
			if (!strcmp(tmp, field->defaultvalue))
				checked = 1;
			fprintf(output, "<input type=\"radio\" name=\"%s\" value=\"%s\"%s>\n",
					field->name, tmp, (checked) ? " checked" : "");
		}
		break;
	default:
		break;
	}
}

void
cell_field_end_option(FILE * output, fieldInfo * field)
{
	switch (field->type) {
	case CHECKBOX:
	case RADIO:
		fprintf(output, "<br>\n");
		break;
	case DROPDOWN:
		fprintf(output, "</option>\n");
		break;
	default:
		break;
	}
}

/* void form_submit(FILE * output, char * str)
{
}
*/
