/*
 * Authenticated User Community
 * Copyright (c) 2000 David C. Moore
 *
 * You may distribute and/or use for any purpose modified or unmodified
 * copies of this software if you preserve the copyright notice above.
 */
 
/**
 * @file pages.c
 *
 * Routines governing the URL organization of the AUC CGI script.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <unistd.h>
#include <dirent.h>

#include <auc/cgic.h>
#include <auc/auc.h>
#include <auc/html.h>
#include <auc/database.h>
#include <auc/auth.h>
#include <auc/files.h>
#include <auc/email.h>
#include <auc/error.h>

#include "pages.h"

/**
 * When AUC parses an input URL, the contents of this string are removed from the left
 * of the URL.  It determines which part of the URL should not be processed by AUC (such
 * as "/cgi-bin/auc.cgi/").  Autogenerated by aucPagesInit() at runtime.
 */
char *aucPagesRoot = "";

static aucPagePtr pages[256];
static aucPagePtr *lastpage;

int read_prefs(void);
int runpage(FILE *, aucPagePtr, char *);
int page_not_found(FILE * output, char *extrapath);
int page_permission_denied(FILE * output, char *message);
int do_login(void);
char * get_AUC_root(void);

/**
 * Initializes the URL processing system.
 *
 * This function is executed at AUC startup.
 */

int
aucPagesInit(void)
{
	*pages = NULL;
	lastpage = pages;

	aucPagesRoot = get_AUC_root();

	return 0;
}

/**
 * Registers a new "virtual" web page with AUC.
 *
 * AUC maintains a heirarchy of virtual pages that can be accessed with
 * URLs relative to the AUC cgi script.  This function is used to add a page
 * to AUCs internal heirarchy, and must be done at runtime for that page to
 * be available.  It is common for plugins to execute this function at initialization
 * time for each web page they wish to make available.
 * 
 * @param parentdir A pointer to the virtual directory where the page will be available.  NULL
 *    for the root directory of AUC.  aucPagesRegisterSubDir() can be used to create subdirectories.
 * @param name The filename of the virtual page.
 * @param type The page type. One of PAGE_DIRFUNC or PAGE_FILE.
 * @param drawfunc A function pointer to the function that outputs the contents of the page.
 * @param reqpriv The access permissions necessary to access the page.  One of "sysop", "normal",
 *    "guest", or NULL.  NULL will allow access without logging on.
 * @param flags Any flags, bitwise OR'ed.  PA_LOCAL will cause the page to be accessable only to
 *    those users with matching UNIX shell accounts.
 * @return Always 0
 */
 
int
aucPagesRegister(aucPagePtr parentdir, char *name,
				long unsigned int type, aucPageDrawFunc drawfunc, char *reqpriv,
				long unsigned int flags)
{
	aucPagePtr **currlast = &lastpage;

	if (parentdir)
		currlast = &parentdir->lastpage;

	**currlast = (aucPagePtr) malloc(sizeof(aucPage));

	(**currlast)->name = name;
	(**currlast)->type = type;
	(**currlast)->pages = (**currlast)->lastpage = NULL;
	(**currlast)->drawfunc = drawfunc;
	(**currlast)->defaultpage = NULL;
	(**currlast)->reqpriv = reqpriv;
	(**currlast)->flags = flags;

	(*currlast)++;
	**currlast = NULL;
	return 0;
}

/**
 * Creates a new virtual subdirectory in the AUC file heirarchy.
 *
 * Creates a virtual subdirectory in which other pages can be registered using
 * aucPagesRegister().
 *
 * @param parentdir A pointer to the virtual directory where the directory will be available.
 *    NULL for the root directory of AUC.
 * @param name The filename of the subdirectory.
 * @param defaultpage A string containing the name of the page (within this subdirectory) that
 *    will be loaded if this subdirectory's URL is typed explicitly.
 * @param reqpriv The access permissions necessary to access the page.  One of "sysop", "normal",
 *    "guest", or NULL.  NULL will allow access without logging on.
 * @param flags Any flags, bitwise OR'ed.  PA_LOCAL will cause the page to be accessable only to
 *    those users with matching UNIX shell accounts.
 * @return A pointer to the internal structure representing the created subdirectory.  This pointer
 *    can be used as input to further aucPagesRegisterSubDir() or aucPagesRegister() calls.
 */

aucPagePtr
aucPagesRegisterSubDir(aucPagePtr parentdir, char *name,
				  char *defaultpage, char *reqpriv, long unsigned int flags)
{
	aucPagePtr **currlast = &lastpage;

	if (parentdir)
		currlast = &parentdir->lastpage;

	**currlast = (aucPagePtr) malloc(sizeof(aucPage));

	(**currlast)->name = name;
	(**currlast)->type = PAGE_DIR;
	(**currlast)->pages = (**currlast)->lastpage =
		(aucPagePtr *) malloc(32 * sizeof(aucPagePtr));
	*(**currlast)->lastpage = NULL;
	(**currlast)->drawfunc = NULL;
	(**currlast)->defaultpage = defaultpage;
	(**currlast)->reqpriv = reqpriv;
	(**currlast)->flags = flags;

	(*currlast)++;
	**currlast = NULL;
	return *((*currlast) - 1);
}

int
pagematch(char *request, char *page, int last)
{
	char *loc;

	if (last && (loc = strstr(request, "-POST")))
		*loc = '\0';

	return (!strcmp(request, page));
}

void
db_failed(FILE * output)
{
	cgiHeaderContentType("text/html");
	aucHtmlSlimHead("%s", _("Error"));

	fprintf(output, "<p><big><b>%s:</b> %s</big></p>\n\n", _("Error"),
			_("AUC could not connect to the SQL database.  Please try again.  If this \
		problem persists, contact the system administrator."));
	aucHtmlSlimFoot();
}


/**
 * The URL processing function of AUC.
 *
 * This function takes a given "virtual" URL, relative to the URL of the AUC cgi script
 * itself, and outputs the desired page.  Any necessary authentication is performed by
 * either processing an existing cookie, or displaying the login page.  This function is
 * called by aucMain() and should not need to be called anywhere else.
 *
 * @param output The stream where any HTML output should be printed.
 * @param path The path, relative to the AUC cgi script, that the user has requested.
 * @return 0 on success, -1 on failure.
 */
 
int
aucPagesRun(FILE * output, char *path)
{
	aucPagePtr *currdir = pages;
	char *currdefault = "home";
	char buf[256];
	char inpath[256], *a, *b;
	int last = 0;

	if (aucFirstRun)
		currdefault = "firstrun";
	else if (aucDBConnectFailed) {
		db_failed(output);
		return -1;
	}

	if (!strlen(path)) {
		snprintf(buf, sizeof(buf), "%s/", aucPagesRoot);
		for (a = (buf + strlen(buf) - 2); a >= buf; a--) {
			if (*a == '/') {
				cgiHeaderLocation(a + 1);
				return -1;
			}
		}
		cgiHeaderLocation(buf);
		return -1;
	}

	snprintf(inpath, sizeof(inpath), "%s", path);
	a = inpath + 1;
	while (!last) {
		for (b = a; (*b != '\0') && (*b != '/'); b++);
		if (*b == '\0')
			last = 1;
		*b = '\0';
		if (last && !strlen(a)) {
			cgiHeaderLocation(currdefault);
			return -1;
		}
		for (; *currdir && !pagematch(a, (*currdir)->name, last); currdir++);
		if (!*currdir) {
			page_not_found(output, inpath);
			return -1;
		}
		switch ((*currdir)->type) {
		case PAGE_FILE:
			if (last)
				return runpage(output, *currdir, "");
			else {
				page_not_found(output, inpath);
				return -1;
			}
			break;
		case PAGE_DIRFUNC:
			if (last) {
				snprintf(buf, sizeof(buf), "%s/", a);
				cgiHeaderLocation(buf);
				return -1;
			} else
				return runpage(output, *currdir, b + 1);
			break;
		case PAGE_DIR:
			if (last) {
				snprintf(buf, sizeof(buf), "%s/%s", a, (*currdir)->defaultpage);
				cgiHeaderLocation(buf);
				return -1;
			} else {
				a = b + 1;
				currdir = (*currdir)->pages;
				currdefault = (*currdir)->defaultpage;
			}
			break;
		}
	}
	return -1;
}

int
do_login(void)
{
	int result;
	char buffer[256];

	result = aucAuthCheck(aucUserCache);
	if (result == AUTH_VALID) {

		aucAuthenticated = 1;
		read_prefs();

		if (!strcmp(cgiPathInfo, "/logincheck")) {
			aucAuthScramblePasswd(buffer, sizeof(buffer), aucUserCache->passwd);
			cgiSendCookie("uname", aucUserCache->uname, aucPagesRoot, 0);
			cgiSendCookie("key", buffer, aucPagesRoot, 0);
		} else if (!strcmp(cgiPathInfo, "/logoff")) {
			if (strlen(aucUserCache->homedir)) {
				snprintf(buffer, sizeof(buffer), "%s/%s/tmp", aucUserCache->homedir, aucUserDir);
				aucFileRemoveDir(buffer, 1);
			}
			cgiSendCookie("uname", "", aucPagesRoot, 1);
			cgiSendCookie("key", "", aucPagesRoot, 1);
			aucEmailClShutdown();
			aucAuthClearCache(aucUserCache->uname);
			aucAuthenticated = 0;
		}

		if (aucUserCache->uid >= 0) {
			setgid(getegid());
			setuid(geteuid());
			setegid(aucUserCache->gid);
			seteuid(aucUserCache->uid);
		}

		if (strlen(aucUserCache->homedir)) {
			snprintf(buffer, sizeof(buffer), "%s/%s", aucUserCache->homedir, aucUserDir);
			aucCheckDir(buffer, 0700, aucUserCache->uid, aucUserCache->gid);
		}
		return 0;

	} else if (result == AUTH_INVALID) {

		cgiHeaderContentType("text/html");
		aucHtmlSlimHead("Invalid Login");
		aucAuthLoginPage(_("<font color=\"CC0000\">Sorry, your username \
         or password was incorrect, please try again.</font>"));
		aucHtmlSlimFoot();
		return -1;

	} else if (result == AUTH_EXPIRED) {

		cgiHeaderContentType("text/html");
		aucHtmlSlimHead(_("Expired Login"));
		aucAuthLoginPage(_("<font color=\"CC0000\">Your login has expired. \
         Please confirm your username and password.</font>"));
		aucHtmlSlimFoot();
		return -1;

	} else if (result == AUTH_DBERROR) {

		cgiHeaderContentType("text/html");
		aucHtmlSlimHead("Database Error");
		aucAuthLoginPage(mysql_error(db));
		aucHtmlSlimFoot();
		return -1;

	} else {

		cgiHeaderContentType("text/html");
		aucHtmlSlimHead(_("Login"));
		aucAuthLoginPage(_("Please login to the system with your username and password."));
		aucHtmlSlimFoot();
		return -1;

	}
}

int
runpage(FILE * output, aucPagePtr currpage, char *path)
{
	int retval;
	uid_t saved;

	if (!currpage->reqpriv) {
		saved = getuid();
		setuid(geteuid());
		seteuid(saved);
		(*(currpage->drawfunc)) (output, path);
		return -1;
	} else if (aucFirstRun) {
		page_permission_denied(output, "");
		return -1;
	}

	if (do_login())
		return -1;

	if ((currpage->flags & PA_LOCAL) && !aucUserCache->local) {
		page_permission_denied(output, _("You must request a local server account in order to gain \
         access to this feature."));
		return 0;
	}
	if (aucAuthPermsof(currpage->reqpriv)) {
		retval = (*(currpage->drawfunc)) (output, path);
		if (retval == -2)
			page_permission_denied(output, "");
	} else {
		page_permission_denied(output, "");
		return 0;
	}
	return 0;
}

char *
get_AUC_root(void)
{
	char buf[256];
	char *a;
	int hack = 0;

	snprintf(buf, sizeof(buf), "%s", cgiRequestURI);

	/* cut off any GET query string that may follow the URL */
	if ((a = strchr(buf, '?')))
		*a = '\0';

	if (strlen(cgiPathInfo) > strlen(buf))
		return strdup(buf);

	a = buf;
	while ((a = strchr(a, '%'))) {
		hack += 2;
		a++;
	}

	*(buf + strlen(buf) - strlen(cgiPathInfo) - hack) = '\0';
	return strdup(buf);
}

/**
 * Reads a user's preferences from the database.
 *
 * Reads the current users preferences from the SQL database (start page
 * columns, file view setting, and pending error messages) and sets some
 * global variables accordingly.
 *
 * @Return Always 0.
 */

int
read_prefs(void)
{
	MYSQL_RES *res = NULL;
	MYSQL_ROW row;
	char buf[1024];

	snprintf(buf, sizeof(buf), "SELECT cols, fileexpert, commanderror \
      FROM prefs WHERE uname=\"%s\"", aucUserCache->uname);
	if (mysql_query(db, buf))
		aucError(mysql_error(db));
	else {
		res = mysql_store_result(db);
		if ((row = mysql_fetch_row(res))) {
			aucNodeColumns = strtoul(row[0], NULL, 10);
			if (!strcmp(row[1], "Expert"))
				aucFileExpert = 1;
			if (row[2] && strlen(row[2]))
				aucErrorAppendStr(row[2]);
		} else {
			snprintf(buf, sizeof(buf), "INSERT INTO prefs (uname) VALUES \
            (\"%s\")", aucUserCache->uname);
			if (mysql_query(db, buf))
				aucError(mysql_error(db));
		}
		mysql_free_result(res);
	}
	return 0;
}

int
aucPagePrefs(FILE * output, char *extrapath)
{
	MYSQL_RES *res = NULL;
	MYSQL_ROW row;
	char buf[8192], *fields[32], *grp, *list;
	char *widths[] = { "400", "600", "800", "1000", NULL };
	char *cols[] = { "2", "3", "4", "5", NULL };
	char *expert[] = { "Regular", "Expert", NULL };
	int x, num;

	for (x = 0; x < 32; x++)
		fields[x] = NULL;

	if (!strcmp(cgiRequestMethod, "GET")) {
		snprintf(buf, sizeof(buf), "SELECT title, fnamecfg, mnamecfg, cols, \
         fileexpert FROM prefs WHERE uname=\"%s\"", aucUserCache->uname);
		if (mysql_query(db, buf))
			aucError(mysql_error(db));
		else {
			res = mysql_store_result(db);
			if ((row = mysql_fetch_row(res))) {
				for (x = 0, num = mysql_num_fields(res); x < num; x++)
					fields[x] = (char *) strdup(row[x]);
				if (!strlen(fields[0]) && !strlen(fields[1]) && !strlen(fields[2]))
					fields[0] = fields[1] = fields[2] = NULL;
			}
			mysql_free_result(res);
		}
		snprintf(buf, sizeof(buf), "SELECT email, homepage, message FROM \
         directory WHERE uname=\"%s\"", aucUserCache->uname);
		if (mysql_query(db, buf))
			aucError(mysql_error(db));
		else {
			res = mysql_store_result(db);
			if ((row = mysql_fetch_row(res)))
				for (x = 0, num = mysql_num_fields(res); x < num; x++)
					fields[x + 5] = (char *) strdup(row[x]);
			mysql_free_result(res);
		}
		snprintf(buf, sizeof(buf), "SELECT students FROM \
         registrations WHERE uname=\"%s\"", aucUserCache->uname);
		if (mysql_query(db, buf))
			aucError(mysql_error(db));
		else {
			res = mysql_store_result(db);
			if ((row = mysql_fetch_row(res)))
				if (row[0])
					fields[8] = (char *) strdup(row[0]);
			mysql_free_result(res);
		}
	} else {
		int new = 0;
		aucHtmlReadForm("name-title name-fname name-mname cols expert \
            email homepage message students", fields, STR_ESCAPE);
		snprintf(buf, sizeof(buf), "REPLACE INTO prefs \
         (uname, title, fnamecfg, mnamecfg, cols, fileexpert) \
         VALUES (\"%s\", \"%s\", \"%s\", \"%s\", \"%s\", \"%s\")", aucUserCache->uname, fields[0], fields[1], fields[2], fields[3], fields[4]);
		if (mysql_query(db, buf))
			aucError(mysql_error(db));

		snprintf(buf, sizeof(buf), "SELECT COUNT(*) FROM directory \
         WHERE uname=\"%s\"", aucUserCache->uname);
		if (mysql_query(db, buf))
			aucError(mysql_error(db));
		else {
			res = mysql_store_result(db);
			row = mysql_fetch_row(res);
			if (!strcmp(row[0], "0"))
				new = 1;
			mysql_free_result(res);
		}
		if (new)
			snprintf(buf, sizeof(buf), "INSERT INTO directory \
            (uname, email, homepage, message) VALUES (\"%s\", \"%s\", \"%s\", \"%s\")", aucUserCache->uname, fields[5], fields[6], fields[7]);
		else
			snprintf(buf, sizeof(buf), "UPDATE directory \
            SET email=\"%s\", homepage=\"%s\", message=\"%s\" WHERE uname=\"%s\"", fields[5], fields[6], fields[7], aucUserCache->uname);
		if (mysql_query(db, buf))
			aucError(mysql_error(db));

		snprintf(buf, sizeof(buf), "UPDATE registrations \
         SET students=\"%s\" WHERE uname=\"%s\"", fields[8], aucUserCache->uname);
		mysql_query(db, buf);

		if (aucChPasswdStatus >= 0) {
			cgiHeaderLocation("home");
			return 0;
		}
	}
	cgiStaticContentType("text/html");

	aucHtmlSlimHead(_("%s's Preferences"), aucUserCache->fullname);

	if (aucChPasswdStatus >= 0) {
		aucGetFullPath("prefs-mess.html", aucDataDir, buf, sizeof(buf));
		aucHtmlEchoFile(output, buf, 0);
	} else {
		fprintf(output, "<p><b>%s:</b> <font color=#FF0000>\n", ("Error"));
		switch (aucChPasswdStatus) {
		case -1:
			fprintf(output, "%s", _("Your old password was incorrect."));
			break;
		case -2:
			fprintf(output, "%s", _("Your new passwords do not match."));
			break;
		case -3:
			fprintf(output, "%s", _("Your new password is indentical to your old password."));
			break;
		case -4:
			fprintf(output, "%s", _("There was an internal error, and your password \
was not changed.  Please try again."));
			break;
		}
		fprintf(output, "</font></p>\n\n");
	}

	aucDrawTableForm(output, "POST", "prefs");
	aucDrawChooseName(output, "name", aucUserCache->uname,
					  fields[0], fields[1], fields[2], _("Preferred Name Formatting"));
	if (!fields[3])
		fields[3] = "3";
	aucDrawDropdown(output, "cols", cols, widths, fields[3], _("Start Page Width (pixels)"));
	aucDrawDropdown(output, "expert", expert, expert, fields[4], _("File View"));

	list = strdup(aucDirectoryGrps);
	while ((grp = strsep(&list, ","))) {
		if (!strcmp(aucUserCache->role, grp)) {
			fprintf(output, "<tr valign=top><td>&nbsp;</td><td>\n");
			fprintf(output, "<p>%s</p>\n",
					_("The following information will appear in your directory listing."));
			fprintf(output, "</td></tr>\n\n");
			aucDrawTextbox(output, "email", 35, 64, fields[5], _("Preferred E-Mail Address"));
			aucDrawTextbox(output, "homepage", 35, 128, fields[6], _("WWW Homepage"));
			aucDrawTextarea(output, "message", 4, 50, fields[7], _("Greeting Text"));
			break;
		}
	}

	if (fields[8]) {
		fprintf(output, "<tr valign=top><td>&nbsp;</td><td>\n");
		fprintf(output, "<p>%s</p>\n",
				_
				("Enter the <i>usernames</i> of students whose course list you would like to see."));
		fprintf(output, "</td></tr>\n\n");
		aucDrawTextarea(output, "students", 3, 9, fields[8], _("Monitored Students"));
	}

	if (aucUserCache->authtype & (AUTH_TYPE_PAM | AUTH_TYPE_DBPASSWD)) {
		fprintf(output, "<tr valign=top><td>&nbsp;</td><td>\n");
		fprintf(output, "<p>%s</p>\n",
				_("Complete the three fields below only if you wish to change your password."));
		fprintf(output, "</td></tr>\n\n");
		aucDrawPassword(output, "oldpass", 16, 16, NULL, _("Old Password"));
		aucDrawPassword(output, "newpass", 16, 16, NULL, _("New Password"));
		aucDrawPassword(output, "newpass2", 16, 16, NULL, _("New Password (again)"));

	}

	aucDrawSubmit(output, _("Save Preferences"));
	aucDrawTableFormEnd(output);

	aucHtmlSlimFoot();

	for (x = 0; x < 32; x++)
		if (fields[x])
			free(fields[x]);
	return 0;
}

int
aucPageLoginCheck(FILE * output, char *extrapath)
{
	char url[256];

	if (cgiFormSuccess == cgiFormStringNoNewlines("url", url, sizeof(url)))
		cgiHeaderLocation(url);
	else
		cgiHeaderLocation("home");
	return 0;
}


int
aucPageLogoff(FILE * output, char *extrapath)
{
	cgiHeaderContentType("text/html");
	aucHtmlSlimHead("%s", _("Logged Off"));
	fprintf(output, "<p>%s</p>\n", _("Thank you for logging off."));
	fprintf(output, "<p><a href=\"%s/home\">%s</a></p>\n", aucPagesRoot, _("Log Back On"));
	aucHtmlSlimFoot();
	return 0;
}


int
aucPageStart(FILE * output, char *extrapath)
{
	aucNodelettListPtr preflist;

	cgiHeaderContentType("text/html");
	aucHtmlSlimHead(_("%s's Start Page"), aucUserCache->fullname);

	preflist = aucNodelettGetUserPrefList(aucUserCache->uname);
	if (aucNodelettVerifyPrefList(preflist, aucUserCache->role)) {
		aucError("User %s has unknown role %s", aucUserCache->uname, aucUserCache->role);
		fprintf(output, "<p>%s, \"%s\", %s</p>\n\n",
				_("Your user role"), aucUserCache->role,
				_("is not known to the system.  Please report this error to the system administrator."));
	} else
		aucNodelettDrawList(output, preflist);
	aucNodelettFreeList(preflist);

	aucHtmlSlimFoot();
	return 0;
}

int
page_not_found(FILE * output, char *extrapath)
{
	cgiHeaderStatus(404, "Not Found");
	cgiHeaderContentType("text/html");
	aucHtmlSlimHead("%s", _("404 Not Found"));
	fprintf(output, "<p>%s</p>\n\n", _("Sorry, the requested URL was not found on this server"));
	fprintf(output, "<p>%s<br>\n", cgiPathInfo);
	fprintf(output, "%s</p>\n", cgiQueryString);
	aucHtmlSlimFoot();
	return 0;
}

int
page_permission_denied(FILE * output, char *message)
{
	cgiHeaderContentType("text/html");

	aucHtmlSlimHead("%s", _("Permission Denied"));

	fprintf(output, "<p>%s.  %s\n", _("Sorry, but this area is not accessible to you"), message);
	fprintf(output, "%s</p>\n\n", 
			_("If this message is in error, please contact the system administrator."));

	aucHtmlSlimFoot();
	return 0;
}

int
aucPageAbout(FILE * output, char *extrapath)
{
	cgiHeaderContentType("text/html");

	aucHtmlSlimHead("%s", _("About AUC"));

	fprintf(output, "<h1>%s</h1>\n", _("AUC - Authenticated User Community"));
	fprintf(output, "<blockquote>%s <b>%s</b>\n", _("version"), VERSION);
	fprintf(output, "%s <a href=\"mailto:dcm@acm.org\">David Moore</a><br>\n", _("by"));
	fprintf(output, "%s: <a href=\"http://auc.sourceforge.net\">http://auc.sourceforge.net</a>\n",
			_("Available online at"));
	fprintf(output, "</blockquote>\n\n");

	fprintf(output, "<dl><dt><b>%s</b></dt>\n", _("Contributors:"));
	fprintf(output, "<dd>Ivan Askwith, Jesse Kovach, Ronen Mukamel, Paolo Prandini, \
Benjy Thomas, Javier Aviles, Helen Chuang, Jose L. Redrejo, \
Jens Kristian Sgaard, Theron J. Lewis, Ben Greslick, Carmen Garcia, \
Victoriano Giralt, Pascal Buch, Mark Forrer, Rachael Hurd, Daina Klimanis, \
Edward Ward, Peter Hammond, Christopher Lloyd, Ben Krefting, Evan Case</dd>\n");
	fprintf(output, "</dl>\n");
	fprintf(output, "<dl><dt><b>%s</b></dt>\n", _("Originated from the CAPonline Project by:"));
	fprintf(output, "<dd>Ben Krefting, David Moore, %s Ivan Askwith</dd>\n", _("and"));
	fprintf(output, "</dl>\n");

	fprintf(output, "<p><small>%s</small></p><hr>\n\n",
				_("AUC is free software; you can redistribute it and/or modify \
it under the terms of the GNU General Public License as published by \
the Free Software Foundation; either version 2 of the License, or \
(at your option) any later version."));

	fprintf(output, "<h2>%s:</h2>\n", _("Powered by"));
	fprintf(output, "<table cellspacing=30 border=0>\n");


	fprintf(output, "<tr><td><p><b>MySQL v%s</b></p>\n", mysql_get_client_info());

	if (db) {
		fprintf(output, "<p><i>%s %s<br>\n", _("Connected to"), mysql_get_host_info(db));
		fprintf(output, "%s %s</i></p>\n", _("Server version"), mysql_get_server_info(db));
	}
	
	fprintf(output, "</td>\n\n");

	fprintf(output, "<td><img src=\"%s/powered_by_mysql.gif\" alt=\"\"></td></tr>", aucGraphDir);

	fprintf(output, "<tr><td colspan=2><p><b>CGIC v1.05</b></p><p><small>\n");

	fprintf(output, "%s\n", _("cgic, copyright 1996 by Thomas Boutell. Permission is granted to use cgic  \
in any application, commercial or noncommercial, at no cost. HOWEVER, \
this copyright paragraph must appear on a \"credits\" page accessible in \
the public online and offline documentation of the program. Modified \
versions of the cgic library should not be distributed without the \
attachment of a clear statement regarding the author of the \
modifications, and this notice may in no case be removed. \
Modifications may also be submitted to the author for inclusion \
in the main cgic distribution."));

	fprintf(output, "</small></p></td></tr>\n\n");

	fprintf(output, "<tr><td colspan=2><p><b>cgiupl v1.4</b></p><p>\n");

	fprintf(output, "%s (C) 1998  James Bourne &lt;jbourne@hardrock.org&gt;</p>\n", _("Copyright"));

	fprintf(output, "<p><small>%s</small></p>\n\n",
			_("This program is free software; you can redistribute it and/or modify \
it under the terms of the GNU General Public License as published by \
the Free Software Foundation; either version 2 of the License, or \
(at your option) any later version."));

	fprintf(output, "</td></tr>\n\n");

	fprintf(output, "<tr><td colspan=2><p><b>IMAP4rev1/c-client</b></p><p>\n");

	fprintf(output, "%s (C) 1998  Mark Crispin &lt;MRC@CAC.Washington.EDU&gt;</p>\n", _("Copyright"));
	fprintf(output, "</td></tr>\n\n");


	fprintf(output, "<tr><td colspan=2><p><b>The GNOME Project</b></p><p>\n");

	fprintf(output, "AUC contains graphical icons made available under the GPL by the\n");
	fprintf(output, "<a href=\"http://www.gnome.org\">GNOME</a> Project.</p>\n");
	fprintf(output, "<p>&copy; 2000 GNOME</p>\n");
	fprintf(output, "</td></tr>\n\n");

	fprintf(output, "</table>\n");

	aucHtmlSlimFoot();
	return 0;

}

