/*
 * Authenticated User Community
 * Copyright (c) 2000 David C. Moore
 *
 * You may distribute and/or use for any purpose modified or unmodified
 * copies of this software if you preserve the copyright notice above.
 *
 * util.c
 *
 * Generic utility functions used throughout AUC
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <regex.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <dirent.h>
#include <errno.h>

#include <auc/auc.h>
#include "util.h"

extern void aucComposeHref(char *email, FILE * output);

#define REGEXP_EMAIL 	"^[^@/]+@[^@/]+$"
#define REGEXP_WWW_INC	"^[^@]+(\\.edu|\\.com|\\.net|\\.org|\\.gov|\\.mil)($|/.*$)"
#define REGEXP_WWW		"^http://.*$"

static int regex_init_done = 0;
static regex_t www_preg;
static regex_t www_inc_preg;
static regex_t email_preg;

char *
aucGetFileExt(char *name)
{
	char *end;

	if (!name)
		return NULL;

	end = name + strlen(name);
	while ((end != name) && (*end != '.'))
		end--;
	if (*end == '.')
		return end + 1;
	else
		return name + strlen(name);
}

int
aucCheckDir(char *dirname, int perm, int uid, int gid)
{
	DIR *dirptr;
	char dirbuf[512], *ptr;
	int need = 1;

	strncpy(dirbuf, dirname, sizeof(dirbuf));
	if ((dirptr = opendir(dirname))) {
		closedir(dirptr);
		return 1;
	} else
		do {
			if (!mkdir(dirbuf, perm)) {
				if (uid != 0)
					chown(dirbuf, uid, gid);
				chmod(dirbuf, perm);
				need--;
				if (need)
					dirbuf[strlen(dirbuf)] = '/';
			} else if (errno == ENOENT) {
				for (ptr = &dirbuf[strlen(dirbuf) - 1]; *ptr != '/'; ptr--);
				*ptr = '\0';
				need++;
			} else
				return 0;

			if (!need)
				return 1;
		} while (errno == ENOENT);
	return 0;
}

void
aucPrintLinkText(FILE * output, char *text)
{
	char *a;

	for (a = text; *a != '\0'; a++)
		switch (*a) {
		case ' ':
			fprintf(output, "%%20");
			break;
		case '/':
			fprintf(output, "%%2F");
			break;
		case '"':
			fprintf(output, "%%22");
			break;
		case '<':
			fprintf(output, "%%3C");
			break;
		case '>':
			fprintf(output, "%%3E");
			break;
		default:
			fputc(*a, output);
			break;
		};
}

void
init_regex(void)
{
	if (!regex_init_done) {
		regcomp(&www_preg, REGEXP_WWW, REG_EXTENDED | REG_ICASE | REG_NOSUB);
		regcomp(&www_inc_preg, REGEXP_WWW_INC, REG_EXTENDED | REG_ICASE | REG_NOSUB);
		regcomp(&email_preg, REGEXP_EMAIL, REG_EXTENDED | REG_ICASE | REG_NOSUB);
		regex_init_done = 1;
	}
}

void
print_chars(char *input, FILE * output)
{
	char *a;

	for (a = input; *a; a++) {
		switch (*a) {
		case '>':
			fprintf(output, "&gt;");
			break;
		case '<':
			fprintf(output, "&lt;");
			break;
		case '"':
			fprintf(output, "&quot;");
			break;
		case '&':
			fprintf(output, "&amp;");
			break;
		default:
			fputc(*a, output);
			break;
		}
	}
}

int
print_word(char *input, FILE * output, unsigned int func_flags)
{
	char oldchar;
	int len, closelink = 1;
	char *subendpos, subendch = 0, *pretext = NULL;
	char *start;

	init_regex();

	len = strcspn(input, " \t\n\r");
	oldchar = *(input + len);
	*(input + len) = '\0';

	if (*input == '(') {
		subendpos = strchr(input, (subendch = ')'));
		pretext = "(";
	} else if (*input == '<') {
		subendpos = strchr(input, (subendch = '>'));
		pretext = "&lt;";
	} else
		subendpos = NULL;

	if (subendpos) {
		*subendpos = '\0';
		fprintf(output, "%s", pretext);
		print_word(input + 1, output, func_flags);
		*subendpos = subendch;
		start = subendpos;
		closelink = 0;
	} else {

		if (func_flags & TEXT_NOLINKS)
			closelink = 0;
		else if (!regexec(&www_preg, input, 0, NULL, 0)) {
			fprintf(output, "<a href=\"");
			print_chars(input, output);
			fprintf(output, "\">");
		} else if (!regexec(&www_inc_preg, input, 0, NULL, 0)) {
			fprintf(output, "<a href=\"http://");
			print_chars(input, output);
			fprintf(output, "\">");
		} else if (!regexec(&email_preg, input, 0, NULL, 0)) {
			aucComposeHref(input, output);
		} else
			closelink = 0;
		start = input;
	}

	print_chars(start, output);

	if (closelink)
		fprintf(output, "</a>");

	*(input + len) = oldchar;
	return len;
}

void
aucPrintText(char *input, FILE * output, unsigned int func_flags)
{
	char *a = input;
	int didword = 1;

	while (*a) {
		int inc = 1;

		switch (*a) {
		case ' ':
			if ((func_flags & TEXT_BSPACES) || didword)
				fputc(*a, output);
			else
				fprintf(output, "&nbsp;");
			break;
		case '\t':
			if (func_flags & TEXT_BSPACES)
				fputc(*a, output);
			else
				fprintf(output, "&nbsp;&nbsp;&nbsp;&nbsp;");
			break;
		case '\n':
			if (func_flags & TEXT_NOBREAKS)
				fputc(*a, output);
			else
				fprintf(output, "<br>\n");
			break;
		case '\r':
			if (func_flags & TEXT_NOBREAKS)
				fputc(*a, output);
			break;
		default:
			inc = print_word(a, output, func_flags);
			break;
		}

		if (inc > 1)
			didword = 1;
		else
			didword = 0;

		a += inc;
	}
}


char *
aucLinkText(char *link)
{
	char *text, *a, *b;

	text = (char *) malloc((strlen(link) + 20) * sizeof(char));
	strcpy(text, "");
	for (a = link; *a != '\0'; a++)
		switch (*a) {
		case ' ':
			strcat(text, "%20");
			break;
		default:
			for (b = text; *b; b++);
			*b = *a;
			*(b + 1) = '\0';
			break;
		}
	return text;
}

void getlanguage(char * str, size_t len)
{
        char * lang;
        int langlen;
        
        if ((lang = getenv("LANGUAGE")) && strlen(lang)) {
                snprintf(str, len, "%s", lang);
                langlen = strcspn(str, ", \n\r\t");
                *(str+langlen) = '\0';
        } else
                snprintf(str, len, "en");
}

char *
aucSmallCaps(char *str, int size)
{
	char buf[1024], smallcaps[1024], *a;
	int prev = 0, open = 0;

	getlanguage(buf, sizeof(buf));
	if (!strncmp(buf, "zh", 2))
		return (char *) strdup(str);

	strcpy(smallcaps, "");
	for (a = str; *a; a++) {
		if ((*a <= 'Z') && (*a >= 'A') && (prev != size)) {
			if (open) {
				strcat(smallcaps, "</font>");
				open--;
			}
			snprintf(buf, sizeof(buf), "<font size=%+d>%c", size, *a);
			open++, prev = size;
		} else if ((*a <= 'z') && (*a >= 'a') && (prev != (size - 1))) {
			if (open) {
				strcat(smallcaps, "</font>");
				open--;
			}
			snprintf(buf, sizeof(buf), "<font size=%+d>%c", size - 1, *a - 'a' + 'A');
			open++, prev = (size - 1);
		} else if ((*a <= 'z') && (*a >= 'a'))
			snprintf(buf, sizeof(buf), "%c", *a - 'a' + 'A');
		else
			snprintf(buf, sizeof(buf), "%c", *a);
		strcat(smallcaps, buf);
	}
	if (open) {
		strcat(smallcaps, "</font>");
		open--;
	}
	return (char *) strdup(smallcaps);
}

int
aucPaginate(FILE * output, char *sort, size_t len, char *name, int recs,
	int maxlines, int pageno)
{
	int x;

	fprintf(output, "%s: \n", _("Page"));
	if (sort)
		snprintf(sort, len, "LIMIT %d,%d", (pageno - 1) * maxlines, maxlines);
	for (x = 1; (x * maxlines - maxlines) < recs; x++)
		if (pageno == x)
			fprintf(output, "<b>%d</b>\n", x);
		else if (strchr(name, '?'))
			fprintf(output, "<a href=\"%s&page=%d\">%d</a>\n", name, x, x);
		else
			fprintf(output, "<a href=\"%s?page=%d\">%d</a>\n", name, x, x);
	fprintf(output, "\n");
	return pageno;
}

int
aucGetFullPath(char * name, char * root, char * path, size_t len)
{
	struct stat sbuf;
	char buf[64];
	
	getlanguage(buf, sizeof(buf));

	snprintf(path, len, "%s/%s.%s", root, name, buf);
	if (!stat(path, &sbuf))
		return 0;
	
	snprintf(path, len, "%s/%s", root, name);
	return stat(path, &sbuf);
}

int savedu = -1;
int savedg = -1;

int
aucSuRoot()
{
	if ((geteuid() == 0) && (getegid() == 0))
		return 0;
		
	if (getegid() != 0) {
		if (getgid() != 0)
			return -1;
		savedg = getegid();
		setegid(0);
	}
	
	if (geteuid() != 0) {
		if (getuid() != 0)
			return -1;
		savedu = geteuid();
		seteuid(0);
	}
	
	return 0;
}

int
aucSuUser()
{
	if ((geteuid() != 0) && (getegid() != 0)) {
		savedu = savedg = -1;
		return 0;
	}
		
	if ((getegid() == 0) && (savedg > -1)) {
		setgid(0);
		setegid(savedg);
	}
	
	if ((geteuid() == 0) && (savedu > -1)) {
		setuid(0);
		seteuid(savedu);
	}
	
	savedg = savedu = -1;
	
	return 0;
}


void
aucSanitizeUsername(char * uname)
{
	char * c;
	
	for (c=uname; *c; ) {
		if (isspace(*c) || !isprint(*c))
			memmove(c, c+1, strlen(c));
		else {
			*c = tolower(*c);
			c++;
		}
	}
}
