/*
 * Authenticated User Community
 * Copyright (c) 2000 David C. Moore
 *
 * You may distribute and/or use for any purpose modified or unmodified
 * copies of this software if you preserve the copyright notice above.
 *
 * aucemail.c
 *
 * The AUC email retrieval agent
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <time.h>
#include <unistd.h>
#include <fcntl.h>
#include <malloc.h>
#include <string.h>
#include <errno.h>
#include <auc/sock.h>
#include "aucemail.h"
#include "client.h"
#include "emailutil.h"

int create_socket(char *name);
int wait_for_connection(int sock);
int close_socket(int sock, char *name);
int process_command(int sock, int comm, char **args);

#define FUNC_COUNT 17

aucEmailAgentFunc funcs[] = {
	aucEmailInit,
	aucEmailClose,
	aucEmailMoveTo,
	aucEmailOptions,
	aucEmailFirstNew,
	aucEmailSetAnswered,
	aucEmailRemoveFolder,
	aucEmailNewFolder,
	aucEmailEmptyTrash,
	aucEmailInject,
	aucEmailCompose,
	aucEmailPrintMessage,
	aucEmailHeader,
	aucEmailFooter,
	aucEmailViewFolders,
	aucEmailPrintList,
	aucEmailMailstat
};

int
main(int argc, char **argv)
{
	int sock;
	char *auth = NULL;

	if (argc != 2) {
		aucError("aucemail: Wrong number of arguments");
		return -1;
	}

	setlocale(LC_ALL, "en_US");
	bindtextdomain(PACKAGE, LOCALEDIR);
	textdomain(PACKAGE);
	bind_textdomain_codeset(PACKAGE, "UTF-8");

	sock = create_socket(argv[1]);
	if (sock == -1) {
		aucError("aucemail: Could not create socket: %s", strerror(errno));
		aucError("aucemail: terminating");
		return -1;
	}

	while (!wait_for_connection(sock)) {
		int conn, command = 0;
		char **args;

		conn = accept(sock, NULL, NULL);
		command = aucSockRecvCommand(conn, "aucemail", &auth, &args);
		if (command != -1) {
			if (funcs[command] == aucEmailClose)
				break;
			process_command(conn, command, args);
		} else
			aucError("aucemail: error in receiving command");

		aucSockFreeArgs(args);
		close(conn);

	}

	aucEmailClose(stderr, NULL);
	free(auth);
	close_socket(sock, argv[1]);
	return 0;
}

int
process_command(int sock, int comm, char **args)
{
	FILE *output;

	output = fdopen(sock, "w");

	if (comm < FUNC_COUNT)
		funcs[comm] (output, args);
	else
		aucError("aucemail: invalid function: %d", comm);
	fclose(output);
	return 0;
}

int
wait_for_connection(int sock)
{
	fd_set rfds;
	struct timeval tv;

	FD_ZERO(&rfds);
	FD_SET(sock, &rfds);

	tv.tv_sec = 300;
	tv.tv_usec = 0;

	if (select(sock + 1, &rfds, NULL, NULL, &tv))
		return 0;
	else
		return -1;

}

int
close_socket(int sock, char *name)
{
	char buf[128];

	snprintf(buf, sizeof(buf), "/tmp/aucemail/%s", name);
	unlink(buf);
	close(sock);
	return 0;
}

int
create_socket(char *name)
{
	int sock;
	struct sockaddr_un sockaddr;

	sock = socket(PF_UNIX, SOCK_STREAM, 0);
	if (sock == -1)
		return -1;

/*	fcntl(sock, F_SETFL, O_NONBLOCK); */

	sockaddr.sun_family = AF_UNIX;
	snprintf(sockaddr.sun_path, sizeof(sockaddr.sun_path), "/tmp/aucemail/%s", name);
	unlink(sockaddr.sun_path);

	if (bind(sock, (struct sockaddr *) &sockaddr, sizeof(sockaddr)) == -1) {
		close(sock);
		return -1;
	}

	if (listen(sock, 5) == -1) {
		close(sock);
		unlink(sockaddr.sun_path);
		return -1;
	}

	return sock;
}
