/*
 * Authenticated User Community
 * Copyright (c) 2000 David C. Moore
 *
 * You may distribute and/or use for any purpose modified or unmodified
 * copies of this software if you preserve the copyright notice above.
 *
 * client.c
 *
 * Low-level email manipulation routines for the web-based email
 * client
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <signal.h>
#include <string.h>
#include <pwd.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <fcntl.h>
#include <errno.h>
#include <netdb.h>
#include <iconv.h>
#include <auc/sock.h>
#include <auc/util.h>
#include "emailutil.h"
#include "client.h"
#include "mail.h"
#include "osdep.h"
#include "rfc822.h"
#include "smtp.h"
#include "nntp.h"
#include "misc.h"

struct options
{
	char personal[64], sigfile[64], domain[32], fcc[32], *hostlist[5];
	int newwindow, pagelength;
	char postponed[32], trashfolder[32], savedfolder[32];
	char uname[65], homedir[128], userdir[16];
	int uid, gid, local;
	char graphdir[64], pagesroot[64], etcroot[32];
	char imapuname[65], imappasswd[32];
};

struct mailstats
{
	int num, new;
	char mailbox[128], liststr[256], selected_box[128], imaphost[64];
	BODY *cur_body;
	FILE *outfile;
	int alert, nolist, foundtrash, foundfcc, foundpostponed;
	int statsonly, printable, printerrors, firstnew;
	int usingimap, loginfailed;
};


struct mailstats ms;
struct options user;
MAILSTREAM *inbox = NIL;


void printaddresses(FILE *, ADDRESS *);
void printaddresserrors(FILE *, ADDRESS *);
void printmimepart(FILE *, unsigned long, BODY *, char *, char *);
void printplaintext(FILE *, char *, unsigned long, unsigned long);
int writeattachedfile(BODY * body, void *, unsigned long, char *, size_t, char *, size_t, char *,
					  size_t);
void printheaders(FILE *, ENVELOPE *, unsigned long);
void printmimeenc(FILE *, unsigned short);
void showattachedimage(BODY * body, FILE *, void *, unsigned long);
void linkattachedfile(BODY * body, FILE *, void *, unsigned long);
void addtostringlist(STRINGLIST **, char *);
void printheadertext(FILE *, char *, unsigned long);
char *generatemessid(void);
int parseheaderforreply(char **, char **, char **, char **, char **, unsigned long);
int parseheaderforforward(char **, char **, unsigned long);
void addresscat(char **, ADDRESS *, int);
int findlongestline(char *, unsigned long, int);
void printreplytext(FILE *, char *, char *, unsigned long);
void printfolderstats(FILE *, MAILSTREAM *, char *, unsigned long);
void printnavcommands(FILE *, unsigned long);
void dropcase(char *);
int badaddresschars(char *buf);
PART *encodeattachments(void);
int compattachedfiles();
int print_raw(FILE * outfile, unsigned long uid);

int
aucEmailInit(FILE * output, char **args)
{
	char newmailbox[512];
	char buf[1024];
	int rshtimeout = 0, debug = 0;

	ms.outfile = 0;
	debug = atoi(args[9]);

	if (!inbox) {
		mail_link(&imapdriver);	/* link in the imap driver */
		mail_link(&nntpdriver);	/* link in the nntp driver */
		mail_link(&pop3driver);	/* link in the pop3 driver */
		mail_link(&mhdriver);	/* link in the mh driver */
		mail_link(&mxdriver);	/* link in the mx driver */
		mail_link(&mbxdriver);	/* link in the mbx driver */
		mail_link(&tenexdriver);	/* link in the tenex driver */
		mail_link(&mtxdriver);	/* link in the mtx driver */
		mail_link(&mmdfdriver);	/* link in the mmdf driver */
		mail_link(&unixdriver);	/* link in the unix driver */
		mail_link(&newsdriver);	/* link in the news driver */
		mail_link(&philedriver);	/* link in the phile driver */
		mail_link(&dummydriver);	/* link in the dummy driver */
		auth_link(&auth_log);	/* link in the log authenticator */

		ms.new = ms.num = -1;
		ms.cur_body = NULL;
		ms.alert = 0;
		ms.nolist = 0;
		ms.foundtrash = 0;
		ms.foundfcc = 0;
		ms.foundpostponed = 0;
		ms.statsonly = 0;
		ms.printable = 0;
		ms.printerrors = 1;
		ms.firstnew = -1;
		ms.usingimap = 0;
		ms.loginfailed = 0;


		snprintf(ms.imaphost, sizeof(ms.imaphost), "%s", args[0]);
		if (strlen(ms.imaphost)) {
			ms.usingimap = 1;
			snprintf(buf, sizeof(buf), "{%s}INBOX", ms.imaphost);
			mail_parameters(NIL, SET_SYSINBOX, buf);
			snprintf(ms.liststr, sizeof(ms.liststr), "{%s}%s", ms.imaphost, args[2]);
		} else {
			mail_parameters(NIL, SET_HOMEDIR, args[8]);
			mail_parameters(NIL, SET_SYSINBOX, args[1]);
			snprintf(ms.liststr, sizeof(ms.liststr), "%s", args[2]);
		}
		if (ms.liststr[strlen(ms.liststr) - 1] != '/')
			strcat(ms.liststr, "/");

		user.hostlist[0] = strdup(args[3]);
		user.hostlist[1] = NULL;

		mail_parameters(NIL, SET_RSHTIMEOUT, &rshtimeout);
		mail_parameters(NIL, SET_LOCKEACCESERROR, NIL);

		snprintf(user.imapuname, sizeof(user.imapuname), "%s", args[6]);
		snprintf(user.imappasswd, sizeof(user.imappasswd), "%s", args[7]);

		if (ms.usingimap && (args[4][0] != '{'))
			snprintf(newmailbox, sizeof(newmailbox), "{%s}%s", ms.imaphost, args[4]);
		else
			snprintf(newmailbox, sizeof(newmailbox), "%s", args[4]);

		inbox = mail_open(inbox, newmailbox, (debug)?OP_DEBUG:NIL);

		if (inbox) {
			snprintf(user.trashfolder, sizeof(user.trashfolder), "%s", args[5]);
			ms.nolist = 1;
			mail_list(inbox, ms.liststr, args[5]);
			ms.nolist = 0;
			snprintf(buf, sizeof(buf), "%s%s", ms.liststr, args[5]);

			if (!ms.foundtrash)
				mail_create(inbox, buf);
		}

	} else {

		if (ms.usingimap && (args[4][0] != '{'))
			snprintf(newmailbox, sizeof(newmailbox), "{%s}%s", ms.imaphost, args[4]);
		else
			snprintf(newmailbox, sizeof(newmailbox), "%s", args[4]);

		if (strcmp(newmailbox, ms.mailbox)) {
			inbox = mail_open(inbox, newmailbox, (debug)?OP_DEBUG:NIL);
			ms.new = ms.num = -1;
			ms.firstnew = -1;
		}
	}


	if (inbox) {
		ms.num = inbox->nmsgs;
		snprintf(ms.mailbox, sizeof(ms.mailbox), "%s", newmailbox);
		snprintf(ms.selected_box, sizeof(ms.selected_box), "%s", newmailbox);
		aucSockSendReply(0, output);
	} else {
		aucError(user.uname, "%s %s", _("could not open"), newmailbox);
		if (ms.loginfailed)
			aucSockSendReply(-2, output);
		else
			aucSockSendReply(-1, output);
	}
	return 0;
}

int
aucEmailClose(FILE * output, char **args)
{
	ms.outfile = 0;
	if (inbox) {
		mail_gc(inbox, GC_ELT && GC_ENV && GC_TEXTS);
		inbox = mail_close_full(inbox, CL_EXPUNGE);
	}
	return 0;
}

int
aucEmailMoveTo(FILE * output, char **args)
{
	ms.outfile = 0;

	if (T == mail_copy_full(inbox, args[1], args[0], CP_UID | CP_MOVE)) {
		mail_expunge(inbox);
		mail_status(inbox, ms.mailbox, SA_UNSEEN);
		aucSockSendReply(0, output);
	} else
		aucSockSendReply(-1, output);
	return 0;
}

int
aucEmailOptions(FILE * output, char **args)
{
	char buf[256];

	ms.outfile = 0;

	snprintf(user.personal, sizeof(user.personal), "%s", args[0]);
	snprintf(user.domain, sizeof(user.domain), "%s", args[1]);
	snprintf(user.fcc, sizeof(user.fcc), "%s", args[2]);
	snprintf(user.sigfile, sizeof(user.sigfile), "%s", args[3]);
//   snprintf(user.imapemail, sizeof(user.imapemail), "%s", args[4]);
	snprintf(buf, sizeof(buf), "%s", args[5]);
	user.pagelength = atoi(buf);

	snprintf(buf, sizeof(buf), "%s", args[6]);
	dropcase(buf);
	if (*buf == 'y')
		user.newwindow = 1;
	else
		user.newwindow = 0;

	snprintf(user.postponed, sizeof(user.postponed), "%s", args[7]);
	snprintf(user.uname, sizeof(user.uname), "%s", args[8]);
	snprintf(user.homedir, sizeof(user.homedir), "%s", args[9]);
	snprintf(user.userdir, sizeof(user.userdir), "%s", args[10]);
	snprintf(buf, sizeof(buf), "%s", args[11]);
	user.uid = atoi(buf);
	snprintf(buf, sizeof(buf), "%s", args[12]);
	user.gid = atoi(buf);
	snprintf(buf, sizeof(buf), "%s", args[13]);
	user.local = atoi(buf);

	snprintf(user.graphdir, sizeof(user.graphdir), "%s", args[14]);
	snprintf(user.savedfolder, sizeof(user.savedfolder), "%s", args[15]);
	snprintf(user.pagesroot, sizeof(user.pagesroot), "%s", args[16]);
	snprintf(user.etcroot, sizeof(user.etcroot), "%s", args[17]);


	aucSockSendReply(0, output);
	return 0;
}

int
aucEmailFirstNew(FILE * output, char **args)
{
	SEARCHPGM *pgm;

	ms.outfile = 0;

	pgm = mail_newsearchpgm();
	pgm->unseen = 1;
	mail_search_full(inbox, NIL, pgm, SE_UID | SE_FREE | SE_NOPREFETCH);
	aucSockSendReply(ms.firstnew, output);
	return 0;
}

int
aucEmailSetAnswered(FILE * output, char **args)
{
	ms.outfile = 0;

	mail_setflag_full(inbox, args[0], "\\ANSWERED", ST_UID);
	aucSockSendReply(0, output);
	return 0;
}

int
aucEmailRemoveFolder(FILE * output, char **args)
{
	ms.outfile = 0;

	if (T == mail_delete(ms.usingimap ? inbox : NIL, args[0])) {
		if (output)
			aucSockSendReply(0, output);
		else
			return 0;
	} else {
		if (output)
			aucSockSendReply(-1, output);
		else
			return -1;
	}
	return 0;
}

int
aucEmailMailstat(FILE * output, char **args)
{
	char *a;

	if ((a = strchr(ms.mailbox, '}')))
		a++;
	else
		a = ms.mailbox;

	if (!strcmp(args[0], a)) {
		aucSockSendReply(ms.num, output);
		aucSockSendReply(ms.new, output);
	} else {
		aucSockSendReply(-1, output);
		aucSockSendReply(-1, output);
	}

	return 0;
}

int
aucEmailNewFolder(FILE * output, char **args)
{
	char buffer[256];

	ms.outfile = 0;

	snprintf(buffer, sizeof(buffer), "%s%s", ms.liststr, args[0]);

	if (T == mail_create(ms.usingimap ? inbox : NIL, buffer)) {
		if (output)
			aucSockSendReply(0, output);
		else
			return 0;
	} else {
		if (output)
			aucSockSendReply(-1, output);
		else
			return -1;
	}
	return 0;
}

int
aucEmailEmptyTrash(FILE * output, char **args)
{
	char buffer[256];

	ms.outfile = 0;

	snprintf(buffer, sizeof(buffer), "%s%s", ms.liststr, user.trashfolder);
	mail_delete(ms.usingimap ? inbox : NIL, buffer);
	mail_create(ms.usingimap ? inbox : NIL, buffer);

	aucSockSendReply(0, output);
	return 0;
}

int
aucEmailInject(FILE * output, char **args)
{
	ENVELOPE *env;
	BODY *body, *curbody;
	SENDSTREAM *outstream;
	FILE *outfile;
	char *tmp, buffer[128], *to, *cc, *subj, *text, *inreplyto = NULL, *textnew = NULL;
	ADDRESS *to_lst = NIL, *cc_lst = NIL, *from_lst = NIL;
	long textnewlen;
	char tmpfilename[32], savefolder[256];
	char fromemail[64], date[64];
	unsigned long func_flags;
	int fd;

	ms.outfile = output;

	func_flags = strtoul(args[5], NULL, 10);

	if (func_flags & QN_POSTPONE)
		snprintf(savefolder, sizeof(savefolder), "%s%s", ms.liststr, user.postponed);
	else
		snprintf(savefolder, sizeof(savefolder), "%s%s", ms.liststr, user.fcc);

	fprintf(output, "<p>%s...</p>\n", _("Parsing Message"));

	to = strdup(args[0]);
	cc = strdup(args[1]);
	subj = strdup(args[2]);
	text = strdup(args[3]);

	if (strlen(args[4]))
		inreplyto = strdup(args[4]);

	ms.printerrors = 1;

	env = mail_newenvelope();
	body = mail_newbody();

	snprintf(fromemail, sizeof(fromemail), "%s@%s",
			 (ms.usingimap) ? user.imapuname : user.uname, user.domain);

	if (badaddresschars(user.personal))
		snprintf(buffer, sizeof(buffer), "\"%s\" <%s>", user.personal, fromemail);
	else
		snprintf(buffer, sizeof(buffer), "%s <%s>", user.personal, fromemail);

	rfc822_parse_adrlist(&to_lst, to, user.domain);
	rfc822_parse_adrlist(&cc_lst, cc, user.domain);
	rfc822_parse_adrlist(&from_lst, buffer, user.domain);

	if (ms.alert)
		return 0;

	env->return_path = from_lst;
	rfc822_date(date);
	env->date = strdup(date);
	env->from = from_lst;
	env->sender = from_lst;
	if (strlen(subj) > 0)
		env->subject = subj;
	env->to = to_lst;
	env->cc = cc_lst;
	env->message_id = (char *) generatemessid();
	if (inreplyto)
		env->in_reply_to = inreplyto;

	if (compattachedfiles()) {
		body->type = TYPEMULTIPART;
		body->encoding = ENC7BIT;
		body->nested.part = mail_newbody_part();
		curbody = &(body->nested.part->body);
		body->nested.part->next = encodeattachments();
	} else
		curbody = body;

	curbody->type = TYPETEXT;
	curbody->encoding = ENC7BIT;

	/* FIXME: html option for compose */
	curbody->subtype = strdup("PLAIN");

	strcrlfcpy(&textnew, &textnewlen, text, (long) strlen(text));

	curbody->contents.text.data = textnew;
	curbody->contents.text.size = textnewlen;

	snprintf(tmpfilename, sizeof(tmpfilename), "/tmp/aucemailXXXXXX");
	fd = mkstemp(tmpfilename);
	outfile = fdopen(fd, "w");
	tmp = (char *) malloc((strlen(to) + strlen(cc) + 8192) * sizeof(char));
	if (T == rfc822_output(tmp, env, body, (soutr_t) fprintf, outfile, 0))
		fprintf(output, "<p>%s %s...</p>\n", _("Saving Message in"), savefolder);
	else {
		fprintf(output, "<p><b>%s:</b> %s</p>\n", _("Error"), _("Message content is invalid"));
		fclose(outfile);
		return 0;
	}
	free(tmp);
	fclose(outfile);

	if (func_flags & QN_POSTPONE) {
		ms.nolist = 1;
		mail_list(inbox, ms.liststr, user.postponed);
		ms.nolist = 0;

		if (!ms.foundpostponed)
			mail_create(inbox, savefolder);
	} else {
		ms.nolist = 1;
		mail_list(inbox, ms.liststr, user.fcc);
		ms.nolist = 0;

		if (!ms.foundfcc)
			mail_create(inbox, savefolder);
	}

	if ((outfile = fopen(tmpfilename, "r"))) {
		STRING messtext;
		void *textstr, *a;
		unsigned long buflen, textlen = 0, len = 0;

		a = textstr = malloc(1024);
		len = 1024;
		while ((buflen = fread(a, 1, 1024, outfile))) {
			textstr = realloc(textstr, len + 1024);
			len += 1024;
			textlen += buflen;
			a = textstr + textlen;
		}

		INIT(&messtext, mail_string, textstr, textlen);

		if (T != mail_append(inbox, savefolder, &messtext))
			fprintf(output, "<p><b>%s</b></p>\n", _("Error saving message in folder"));

		free(textstr);
		fclose(outfile);
	}

	unlink(tmpfilename);

	if (!(func_flags & QN_POSTPONE)) {
		fprintf(output, "<p>%s...</p>\n", _("Sending Message"));
		if ((outstream = smtp_open(user.hostlist, NIL))) {
			if (smtp_mail(outstream, "MAIL", env, body))
				fprintf(output, "<p><b>%s</b></p>\n", _("Success!"));
			else {
				fprintf(output, "<p><b>%s:</b> %s</p>\n", _("Error"), outstream->reply);
				printaddresserrors(output, env->to);
				printaddresserrors(output, env->cc);
				smtp_close(outstream);
				return 0;
			}
			smtp_close(outstream);
		} else {
			fprintf(output, "<p><b>%s</b></p>\n", _("Error opening SMTP connection"));
			return 0;
		}
	}

	ms.printerrors = 0;

	return 0;
}

int
compattachedfiles(void)
{
	char buf[1024];
	struct dirent *dir_ent;
	DIR *dir;
	struct stat finfo;
	int result = 0;

	snprintf(buf, sizeof(buf), "%s/%s/attach", user.homedir, user.userdir);
	if ((dir = opendir(buf))) {
		while ((dir_ent = readdir(dir))) {
			snprintf(buf, sizeof(buf), "%s/%s/attach/%s", user.homedir, user.userdir,
					 dir_ent->d_name);
			if (!stat(buf, &finfo) && S_ISREG(finfo.st_mode)) {
				result = 1;
				break;
			}
		}
		closedir(dir);
	}
	return result;
}

PART *
encodeattachments(void)
{
	char buf[1024];
	struct dirent *dir_ent;
	DIR *dir;
	struct stat finfo;
	PART *result = NIL, **curpart;
	FILE *infile;
	char *data = NULL, *bufdata, *mimetype;
	unsigned long datalen = 0, bufsize;

	curpart = &result;

	snprintf(buf, sizeof(buf), "%s/%s/attach", user.homedir, user.userdir);
	if ((dir = opendir(buf))) {
		while ((dir_ent = readdir(dir))) {
			snprintf(buf, sizeof(buf), "%s/%s/attach/%s", user.homedir, user.userdir,
					 dir_ent->d_name);
			if (!stat(buf, &finfo) && S_ISREG(finfo.st_mode) && (infile = fopen(buf, "r"))) {
				*curpart = mail_newbody_part();

				mimetype = aucMimeGetType(aucGetFileExt(dir_ent->d_name), user.etcroot);
				if (!mimetype)
					mimetype = "application/octet-stream";

				(*curpart)->body.type = aucMimeType2Int(mimetype);
				(*curpart)->body.subtype = strdup(aucMimeType2Subtype(mimetype));
				(*curpart)->body.encoding = ENCBINARY;
				(*curpart)->body.id = (char *) generatemessid();
				(*curpart)->body.description = strdup("");
				(*curpart)->body.parameter = mail_newbody_parameter();
				(*curpart)->body.parameter->attribute = strdup("name");
				(*curpart)->body.parameter->value = strdup(dir_ent->d_name);

				(*curpart)->body.disposition.type = strdup("attachment");
				(*curpart)->body.disposition.parameter = mail_newbody_parameter();
				(*curpart)->body.disposition.parameter->attribute = strdup("filename");
				(*curpart)->body.disposition.parameter->value = strdup(dir_ent->d_name);

				datalen = 0;
				bufdata = data = (char *) malloc(1024);
				while ((bufsize = fread(bufdata, 1, 1024, infile))) {
					datalen += bufsize;
					data = realloc(data, datalen + 1024);
					bufdata = data + datalen;
				}
				fclose(infile);

				(*curpart)->body.contents.offset = 0;
				(*curpart)->body.contents.text.data = data;
				(*curpart)->body.contents.text.size = datalen;
				(*curpart)->body.size.bytes = datalen;

				(*curpart)->next = NIL;

				curpart = &((*curpart)->next);
			}
		}
		closedir(dir);
		return result;
	} else
		return NIL;
}

int
aucEmailCompose(FILE * outfile, char **args)
{
	FILE *sig;
	char buf[512], *messtext = NULL, *newto = NULL, *cc = NULL, *subject = NULL, *replystr =
		NULL, *inreplyto = NULL;
	unsigned long len;
	int offset = 2;
	DIR *dir;
	struct dirent *dir_ent;
	unsigned long func_flags;
	long messid;

	ms.outfile = outfile;

	messid = strtol(args[1], NULL, 10);
	func_flags = strtoul(args[2], NULL, 10);

	/* Clean out the temporary directory used for file attachments */
	snprintf(buf, sizeof(buf), "%s/%s/attach", user.homedir, user.userdir);
	if ((dir = opendir(buf))) {
		while ((dir_ent = readdir(dir))) {
			snprintf(buf, sizeof(buf), "%s/%s/attach/%s", user.homedir, user.userdir,
					 dir_ent->d_name);
			if (strcmp(dir_ent->d_name, ".") && strcmp(dir_ent->d_name, ".."))
				unlink(buf);
		}
		closedir(dir);
	} else
		aucCheckDir(buf, 0700, user.uid, user.gid);

	if (func_flags & QN_FORWARD) {
		offset = 0;
		parseheaderforforward(&replystr, &subject, messid);
		newto = strdup(args[0]);
	} else if (messid >= 0)
		parseheaderforreply(&replystr, &newto, &cc, &subject, &inreplyto, messid);
	else
		newto = strdup(args[0]);

	fprintf(outfile, "<form action=inject method=post>\n");

	if (messid >= 0) {
		fprintf(outfile, "<input type=hidden name=ansbox value=\"%s\">\n", ms.mailbox);
		fprintf(outfile, "<input type=hidden name=ansmesg value=\"%ld\">\n", messid);
	}

	if (inreplyto) {
		fprintf(outfile, "<input type=hidden name=inreplyto value=\"");
		aucPrintText(inreplyto, outfile, TEXT_BSPACES | TEXT_NOBREAKS | TEXT_NOLINKS);
		fprintf(outfile, "\">\n");
	}

	fprintf(outfile, "<table border=0 cellpadding=2><tr bgcolor=#666666><td>\n");
	fprintf(outfile, "<table bgcolor=#FFFFFF cellspacing=0 cellpadding=2 border=0>\n");

	fprintf(outfile, "<tr><td width=80><b>%s:</b></td><td>\n", _("To"));
	if (newto) {
		fprintf(outfile, "<input type=text name=to size=55 value=\"");
		aucPrintText(newto, outfile, TEXT_BSPACES | TEXT_NOBREAKS | TEXT_NOLINKS);
		fprintf(outfile, "\"></td></tr>\n");
	} else
		fprintf(outfile, "<input type=text name=to size=55 value=\"\"></td></tr>\n");

	fprintf(outfile, "<tr><td width=80><b>%s:</b></td><td>\n", _("cc"));

	if (cc) {
		fprintf(outfile, "<input type=text name=cc size=55 value=\"");
		aucPrintText(cc, outfile, TEXT_BSPACES | TEXT_NOBREAKS | TEXT_NOLINKS);
		fprintf(outfile, "\"></td></tr>\n");
	} else
		fprintf(outfile, "<input type=text name=cc size=55 value=\"\"></td></tr>\n");

	fprintf(outfile, "<tr><td width=80><b>%s:</b></td><td>\n", _("Subject"));
	if (subject) {
		fprintf(outfile, "<input type=text name=subject size=55 value=\"");
		aucPrintText(subject, outfile, TEXT_BSPACES | TEXT_NOBREAKS | TEXT_NOLINKS);
		fprintf(outfile, "\"></td></tr>\n");
	} else
		fprintf(outfile, "<input type=text name=subject size=55></td></tr>\n");

	fprintf(outfile, "<tr><td width=80><b>%s:</b></td><td>\n", _("Attachments"));
	fprintf(outfile,
			"<a href=\"javascript: newWindow('attach','attach','640')\"><b>%s</b></a></td></tr>\n",
			_("Click here to attach files to this message"));

	fprintf(outfile, "</table></td></tr></table>\n");


	if ((!(func_flags & QN_REPLYFULL) && !(func_flags & QN_FORWARD)) || (messid < 0))
		fprintf(outfile, "<p><textarea name=text rows=20 cols=74 wrap=hard>");
	else {
		switch (ms.cur_body->type) {
		case TYPETEXT:
			if (!strcmp(lcase(ms.cur_body->subtype), "plain"))
				messtext = (char *) mail_fetchtext_full(inbox, messid, &len, FT_UID);
			break;
		case TYPEMULTIPART:
			if (ms.cur_body->nested.part->body.type == TYPETEXT
				&& !strcmp(lcase(ms.cur_body->nested.part->body.subtype), "plain"))
				messtext = (char *) mail_fetchbody_full(inbox, messid, "1", &len, FT_UID);
			break;
		default:
			break;
		}
		fprintf(outfile, "<p><textarea name=text rows=20 cols=%d wrap=hard>",
				findlongestline(messtext, len, offset));
	}

	if ((func_flags & QN_REPLYFULL) && messtext) {
		fprintf(outfile, replystr);
		printreplytext(outfile, "> ", messtext, len);
	}

	if (*user.sigfile == '/')
		snprintf(buf, sizeof(buf), "%s", user.sigfile);
	else
		snprintf(buf, sizeof(buf), "%s/%s", user.homedir, user.sigfile);

	if ((sig = fopen(buf, "r"))) {
		fprintf(outfile, "\n\n");
		while (fgets(buf, sizeof(buf), sig))
			fprintf(outfile, buf);
		fclose(sig);
	}

	if ((func_flags & QN_FORWARD) && messtext) {
		fprintf(outfile, replystr);
		printreplytext(outfile, "", messtext, len);
	}

	fprintf(outfile, "</textarea></p>\n");
	fprintf(outfile, "<input type=submit value=\"%s\">\n", _("Send Message"));
	fprintf(outfile, "<!--input type=submit name=postpone value=\"%s\"--></form>\n", _("Postpone"));
	return 0;
}

int
print_header_box(FILE * outfile, unsigned long uid)
{
	ENVELOPE *mess_env;
	MESSAGECACHE *mess_cache;

	mess_env = mail_fetchstructure_full(inbox, uid, &ms.cur_body, FT_UID);
	mess_cache = mail_elt(inbox, mail_msgno(inbox, uid));
	if (!mess_cache->seen)
		ms.new--;
	printheaders(outfile, mess_env, uid);
	return 0;
}

int
aucEmailPrintMessage(FILE * outfile, char **args)
{
	char buf[256];
	DIR *dir;
	struct dirent *dir_ent;
	unsigned long uid, func_flags;

	ms.outfile = outfile;

	uid = strtoul(args[0], NULL, 10);
	func_flags = strtoul(args[1], NULL, 10);

	if (func_flags & QN_PRINTABLE)
		ms.printable = 1;
	else
		ms.printable = 0;

	if (!(func_flags & QN_RAWTEXT) && !(func_flags & QN_PRINTABLE))
		printnavcommands(outfile, uid);

	if (func_flags & QN_RAWTEXT) {
		print_raw(outfile, uid);
		return 0;
	} else
		print_header_box(outfile, uid);

	snprintf(buf, sizeof(buf), "%s/%s/tmp", user.homedir, user.userdir);
	if ((dir = opendir(buf))) {
		while ((dir_ent = readdir(dir))) {
			snprintf(buf, sizeof(buf), "%s/%s/tmp/%s", user.homedir, user.userdir, dir_ent->d_name);
			if (strcmp(dir_ent->d_name, ".") && strcmp(dir_ent->d_name, ".."))
				unlink(buf);
		}
		closedir(dir);
	} else
		aucCheckDir(buf, 0700, user.uid, user.gid);

	printmimepart(outfile, uid, ms.cur_body, "", "");

	if (!(func_flags & QN_RAWTEXT) && !(func_flags & QN_PRINTABLE)) {
		unsigned long msgno;
		msgno = mail_msgno(inbox, uid);
		fprintf(outfile, "<form action=\"move\" method=post>\n");
		if (msgno < inbox->nmsgs)
			fprintf(outfile, "<input type=hidden name=next value=\"%ld\">\n",
					mail_uid(inbox, msgno + 1));
		fprintf(outfile, "<input type=hidden name=mess value=\"%ld\">\n", uid);
	}

	if ((ms.cur_body->size.lines > 12) && !(func_flags & QN_RAWTEXT)
		&& !(func_flags & QN_PRINTABLE)) printnavcommands(outfile, uid);

	return 0;
}

int
print_raw(FILE * outfile, unsigned long uid)
{
	char *str;
	unsigned long len;
	MESSAGECACHE *mess_cache;

	ms.outfile = outfile;

	str = mail_fetchbody_full(inbox, uid, "", &len, FT_UID);
	mess_cache = mail_elt(inbox, mail_msgno(inbox, uid));
	if (!mess_cache->seen)
		ms.new--;
	printplaintext(outfile, str, len, TEXT_NOLINKS);
	return 0;
}

int
folder_dropdown(FILE * outfile, char *tdargs, char *action, char *caption)
{
	if (inbox) {
		fprintf(outfile, "<form action=\"%s\" method=\"get\">\n", action);
		fprintf(outfile, "<td %s>\n", tdargs);
		fprintf(outfile, "%s <select name=mailbox>\n", caption);
		fprintf(outfile, "<option value=\"INBOX\">%s</option>\n", _("INBOX"));
		mail_list(inbox, ms.liststr, "*");
		fprintf(outfile, "</select><input type=submit value=\"%s\">\n", _("go"));
		fprintf(outfile, "</td>\n");
		fprintf(outfile, "</form>\n\n");
	} else {
		fprintf(outfile, "<td %s>&nbsp;</td>\n\n", tdargs);
	}
	return 0;
}

int
aucEmailHeader(FILE * outfile, char **args)
{
	int pageno = 1;
	char buf[256];
	int func_flags, page;

	ms.outfile = outfile;

	func_flags = strtoul(args[0], NULL, 10);
	page = strtoul(args[1], NULL, 10);

	fprintf(outfile, "%s: <b>%s@%s</b>\n\n", _("Your email address is"),
			(ms.usingimap) ? user.imapuname : user.uname, user.domain);

	fprintf(outfile,
			"<table border=0 cellpadding=2 cellspacing=0 bgcolor=#000000 width=100%%><tr><td>\n");
	fprintf(outfile,
			"<table border=0 cellspacing=0 cellpadding=2 width=100%%><tr bgcolor=#CCCCCC>\n\n");

	fprintf(outfile,
			"<td>&nbsp;<a %s href=\"compose\"><img src=\"%s/fileops/email-compose.gif\" border=0 align=absmiddle alt=\"\"> <b>%s</b></a></td>\n",
			user.newwindow ? "target=\"_blank\"" : "", user.graphdir, _("Compose Message"));
	fprintf(outfile,
			"<td>&nbsp;<a href=\"view\"><img src=\"%s/fileops/email.gif\" border=0 align=absmiddle alt=\"\"> <b>%s</b></a></td>\n",
			user.graphdir, _("INBOX"));
	fprintf(outfile, "<td>&nbsp;<a href=\"newfolder\"><b>%s</b></a></td>\n", _("New Folder"));
	fprintf(outfile,
			"<td>&nbsp;<a href=\"folders\"><img src=\"%s/fileops/folder.gif\" border=0 align=absmiddle alt=\"\"> <b>%s</b></a></td>\n",
			user.graphdir, _("Folders"));
	fprintf(outfile,
			"<td>&nbsp;<a href=\"emptrash\"><img src=\"%s/fileops/trash.gif\" border=0 align=absmiddle alt=\"\"> <b>%s</b></a></td>\n",
			user.graphdir, _("Empty Trash"));
	fprintf(outfile,
			"<td>&nbsp;<a href=\"options\"><img src=\"%s/fileops/prefs.gif\" border=0 align=absmiddle alt=\"\"> <b>%s</b></a></td>\n",
			user.graphdir, _("Preferences"));

	fprintf(outfile, "</tr>\n\n");

	if (ms.new == -1)
		mail_status(inbox, ms.mailbox, SA_UNSEEN);
	else
		mail_ping(inbox);

/*   if (inbox && (ms.new == -1)) {
      SEARCHPGM *pgm;
      ms.num = inbox->nmsgs;
      ms.new = 0;
      pgm = mail_newsearchpgm();
      pgm->unseen = 1;
      mail_search_full(inbox, NIL, pgm, SE_UID | SE_FREE | SE_NOPREFETCH);
		mail_free_searchpgm(&pgm);
   }
*/

	fprintf(outfile,
			"<tr><td colspan=6 bgcolor=#CCCCCC width=100%%><table cellpadding=0 cellspacing=0 border=0 width=100%%>\n");
	if (func_flags & QN_COMPOSE) {
		fprintf(outfile,
				"<tr bgcolor=#CCCCCC><td colspan=3><font face=\"Arial,Helvetica\"><b>%s</b></font></td>",
				_("Compose Message"));
		fprintf(outfile, "<td colspan=3 align=right>&nbsp;</td></tr>\n");
	} else if (func_flags & QN_REPLY) {
		fprintf(outfile,
				"<tr bgcolor=#CCCCCC><td colspan=3><font face=\"Arial,Helvetica\"><b>%s</b></font></td>",
				_("Reply"));
		fprintf(outfile, "<td colspan=3 align=right>&nbsp;</td></tr>\n");
	} else if (func_flags & QN_OPTIONS) {
		fprintf(outfile,
				"<tr bgcolor=#CCCCCC><td colspan=3><font face=\"Arial,Helvetica\"><b>%s</b></font></td>",
				_("Email Preferences"));
		fprintf(outfile, "<td colspan=3 align=right>&nbsp;</td></tr>\n");
	} else if (func_flags & QN_FOLDERS) {
		fprintf(outfile,
				"<tr bgcolor=#CCCCCC><td colspan=3><font face=\"Arial,Helvetica\"><b>%s</b></font></td>",
				_("Folders"));
		fprintf(outfile, "<td colspan=3 align=right>&nbsp;</td></tr>\n");
	} else if ((func_flags & QN_MAILBOX) || (func_flags & QN_SINGLE)) {
		char action[256];
		fprintf(outfile,
				"<tr bgcolor=#CCCCCC><td colspan=3><font face=\"Arial,Helvetica\"><b>%s</b> (%s:%d %s:%d)</font>\n",
				_(ms.mailbox), _("mail"), ms.num, _("new"), ms.new);
		if (func_flags & QN_MAILBOX) {
			fprintf(outfile, "<br><font size=-1><a href=\"view?mailbox=");
			aucPrintLinkText(outfile, ms.mailbox);
			if (!(func_flags & QN_FORWARDSORT))
				fprintf(outfile, "&sort=forward\">%s</a>: %s</font>\n\n", _("Sort"),
						_("Newest First"));
			else
				fprintf(outfile, "\">%s</a>: %s</font>\n\n", _("Sort"), _("Oldest First"));

			if (user.pagelength && (ms.num > user.pagelength)) {
				fprintf(outfile, "&nbsp;&nbsp;&nbsp;<font size=-1>[\n");
				snprintf(buf, sizeof(buf), "view?mailbox=%s", aucLinkText(ms.mailbox));
				pageno = aucPaginate(outfile, NULL, 0, buf, ms.num, user.pagelength, page);
				fprintf(outfile, "]</font>\n\n");
			}
		}
		fprintf(outfile, "</td>\n");
		snprintf(action, sizeof(action), "view");
		snprintf(buf, sizeof(buf),
				 "&nbsp;<img src=\"%s/fileops/goto.gif\" border=0 align=absmiddle alt=\"\"> %s",
				 user.graphdir, _("Goto Folder:"));
		folder_dropdown(outfile, "align=right colspan=3", action, buf);
		fprintf(outfile, "</tr>\n");
	}
	fprintf(outfile, "</table></td></tr>\n");

	fprintf(outfile, "</table></td></tr></table>\n\n");

	return 0;
}

int
aucEmailFooter(FILE * outfile, char **args)
{
	int func_flags;

	ms.outfile = outfile;

	func_flags = strtoul(args[0], NULL, 10);

	fprintf(outfile,
			"<table border=0 cellpadding=2 cellspacing=0 bgcolor=#000000 width=100%%><tr><td>\n");
	fprintf(outfile, "<table border=0 cellspacing=0 cellpadding=2 width=100%%>\n\n");

	if (inbox && ((QN_MAILBOX | QN_SINGLE) & func_flags) && (ms.num > 0)) {
		fprintf(outfile, "<tr bgcolor=#CCCCCC><td colspan=2 align=right>\n");
		fprintf(outfile, "<input type=hidden name=source value=\"%s\">\n", ms.mailbox);
		fprintf(outfile,
				"&nbsp;<img src=\"%s/fileops/move-email.gif\" border=0 align=absmiddle alt=\"\">\n",
				user.graphdir);
		if (func_flags & QN_SINGLE) {
			fprintf(outfile, "%s: <select name=\"dest\">\n", _("Move current message to folder"));
			snprintf(ms.selected_box, sizeof(ms.selected_box), "%s%s", ms.liststr,
					 user.savedfolder);
		} else {
			fprintf(outfile, "%s: <select name=\"dest\">\n", _("Move selected messages to folder"));
			snprintf(ms.selected_box, sizeof(ms.selected_box), "%s%s", ms.liststr,
					 user.trashfolder);
		}
		fprintf(outfile, "<option value=\"INBOX\">%s</option>\n", _("INBOX"));
		mail_list(inbox, ms.liststr, "*");
		snprintf(ms.selected_box, sizeof(ms.selected_box), "%s", ms.mailbox);
		fprintf(outfile, "</select><input type=submit value=%s>\n", _("move"));
		fprintf(outfile, "</td></tr></form>\n\n");
	} else
		fprintf(outfile, "<tr bgcolor=#CCCCCC><td>&nbsp;</td></tr>\n");

	fprintf(outfile, "</table></td></tr></table>\n\n");

	return 0;
}

int
aucEmailViewFolders(FILE * output, char **args)
{

	ms.outfile = output;

	fprintf(output, "<dl><dt><b>%s:</b></dt>\n", _("Inboxes"));
	fprintf(output, "<dd><table border=0 cellspacing=2>\n");
	fprintf(output, "<tr bgcolor=#CCCCCC><td width=200><b>%s</b></td><td width=50><b>%s</b></td>\n",
			_("Name"), _("Mail"));
	fprintf(output, "<td width=50><b>%s</b></td><td width=200><b>%s</b></td></tr>\n",
			_("New"), _("Options"));

	if (ms.mailbox[0] == '{')
		printfolderstats(output, inbox, ms.mailbox, QN_NODELETE);
	else
		printfolderstats(output, inbox, "INBOX", QN_NODELETE);

	fprintf(output, "</table></dd></dl>\n");
	fprintf(output, "<dl><dt><b>%s %s:</b></dt>\n", _("Folders in"), ms.liststr);
	fprintf(output, "<dd><table border=0>\n");
	fprintf(output, "<tr bgcolor=#CCCCCC><td width=200><b>%s</b></td><td width=50><b>%s</b></td>\n",
			_("Name"), _("Mail"));
	fprintf(output, "<td width=50><b>%s</b></td><td width=200><b>%s</b></td></tr>\n",
			_("New"), _("Options"));
	ms.statsonly = 1;
	mail_list((ms.liststr[0] == '{') ? inbox : NIL, ms.liststr, "*");
	ms.statsonly = 0;
	fprintf(output, "</table></dd></dl>\n");
	ms.new = -1;

	return 0;
}

int
aucEmailPrintList(FILE * outfile, char **args)
{
	ENVELOPE *mess_env;
	MESSAGECACHE *mess_elt;
	int y, x, reversesort = 1;
	time_t secs;
	char from[40], datestr[7], subject[256], subjtext[256], flags[3];
	int page;
	int func_flags;

	ms.outfile = outfile;

	page = atoi(args[0]);
	func_flags = atoi(args[1]);

	if (func_flags & QN_FORWARDSORT)
		reversesort = 0;

	fprintf(outfile, "<form action=\"move\" method=post>\n");
	fprintf(outfile, "<pre>");
	if (inbox->nmsgs)
		fprintf(outfile,
				"<code>&nbsp;&nbsp;&nbsp;&nbsp;</code><font face=\"Arial,Helvetica\" size=-1>v %s</font>\n",
				_("Delete/Move"));
	for (y = (page - 1) * user.pagelength;
		 (y < inbox->nmsgs) && ((user.pagelength == 0) ? 1 : (y < page * user.pagelength)); y++) {
		if (reversesort)
			x = inbox->nmsgs - y;
		else
			x = y + 1;

		mess_env = mail_fetchenvelope(inbox, x);
		if (mess_env->from) {
			if (mess_env->from->personal)
				snprintf(from, sizeof(from), "%s", mess_env->from->personal);
			else
				snprintf(from, sizeof(from), "%s@%s", mess_env->from->mailbox,
						 mess_env->from->host);
		} else
			snprintf(from, sizeof(from), "%s", _("no from address"));

		mess_elt = mail_elt(inbox, x);
		if (mess_elt->seen && (func_flags & QN_NEW))
			continue;
		secs = mail_longdate(mess_elt);
		strftime(datestr, sizeof(datestr), _("%b %d"), localtime(&secs));
		if (mess_env->subject)
			snprintf(subjtext, sizeof(subjtext), "%.60s", mess_env->subject);
		else
			snprintf(subjtext, sizeof(subjtext), "%s", _("no subject"));
//      aucText2html(subjtext, sizeof(subjtext), NULL, TEXT_NOLINKS);
		if (subjtext[0] == '\0')
			strcpy(subjtext, _("no subject"));
		snprintf(subject, sizeof(subject), "<a href=\"view?mailbox=%s&UID=%ld\">%s</a>",
				 aucLinkText(ms.mailbox), mail_uid(inbox, x), subjtext);
		strcpy(flags, "  ");
		if (!mess_elt->seen) {
//         ms.new++;
			flags[0] = 'N';
		}
		if (mess_elt->answered)
			flags[1] = 'A';
		fprintf(outfile,
				"%3d.<input type=checkbox name=mess value=\"%ld\">%-2s %-6s %-25.25s %-60s\n", x,
				mail_uid(inbox, x), flags, datestr, from, subject);

	}
	fprintf(outfile, "</pre>");

	if ((ms.num == 0) && !(func_flags & QN_NEW))
		fprintf(outfile, "<i>%s</i>\n", _("no messages in folder"));
	else if ((ms.new == 0) && (func_flags & QN_NEW))
		fprintf(outfile, "<i>%s</i>\n", _("no new messages in folder"));

	return 0;
}

void
dropcase(char *str)
{
	char *a;

	for (a = str; *a; a++)
		if ((*a <= 'Z') && (*a >= 'A'))
			*a -= 'A' - 'a';
}

void
printreplytext(FILE * outfile, char *lead, char *text, unsigned long len)
{
	if (text && lead) {
		fprintf(outfile, lead);
		while (len && *text) {
			fputc(*text, outfile);
			if (*text == '\n')
				fprintf(outfile, lead);
			len--;
			text++;
		}
	}
}

int
findlongestline(char *text, unsigned long len, int offset)
{
	int longest = 74, max = 80, c;
	char *mark;

	longest -= offset;
	max -= offset;
	if (text) {
		mark = text;
		while (*mark && len) {
			for (c = 0; *mark != '\n' && *mark != '\0'; mark++, len--)
				c++;
			if (*mark != '\0')
				mark++, len--;
			if (c > longest && c < max)
				longest = c;
			else if (c >= max)
				return (max + offset);
		}
	}
	return (longest + offset);
}

int
parseheaderforforward(char **replystr, char **subject, unsigned long uid)
{
	ENVELOPE *mess_env;
	unsigned long len;
	STRINGLIST *list, **curr;
	char *headers;
	char buf[128];

	mess_env = mail_fetchstructure_full(inbox, uid, &ms.cur_body, FT_UID);
	*subject = *replystr = NULL;
	snprintf(buf, sizeof(buf), "\n---------- %s ----------\n", _("Forwarded message"));
	*replystr = strdup(buf);
	curr = &list;
	addtostringlist(curr, "Resent-Date");
	curr = &(*curr)->next;
	addtostringlist(curr, "Resent-From");
	curr = &(*curr)->next;
	addtostringlist(curr, "Resent-To");
	curr = &(*curr)->next;
	addtostringlist(curr, "Resent-cc");
	curr = &(*curr)->next;
	addtostringlist(curr, "Resent-Subject");
	curr = &(*curr)->next;
	addtostringlist(curr, "From");
	curr = &(*curr)->next;
	addtostringlist(curr, "Reply-to");
	curr = &(*curr)->next;
	addtostringlist(curr, "Date");
	curr = &(*curr)->next;
	addtostringlist(curr, "cc");
	curr = &(*curr)->next;
	addtostringlist(curr, "To");
	curr = &(*curr)->next;
	addtostringlist(curr, "Subject");

	headers = mail_fetchheader_full(inbox, uid, list, &len, FT_UID);

	*replystr = (char *) realloc(*replystr, len + strlen(*replystr) + 1);
	strncat(*replystr, headers, len);

	if (mess_env->subject) {
		if ((strlen(mess_env->subject) >= 5)
			&& !strncmp(mess_env->subject + strlen(mess_env->subject) - 5, "(fwd)", 5)) {
			*subject = malloc(1 + strlen(mess_env->subject));
			sprintf(*subject, "%s", mess_env->subject);
		} else {
			*subject = malloc(7 + strlen(mess_env->subject));
			sprintf(*subject, _("%s (fwd)"), mess_env->subject);
		}
	} else
		*subject = strdup(_("no subject (fwd)"));

	return (0);
}

void
printfolderstats(FILE * output, MAILSTREAM * stream, char *mailbox, unsigned long func_flags)
{
	long result;

	if (strncmp(mailbox, ms.liststr, strlen(ms.liststr)) == 0) {
		fprintf(output, "<tr><td><a href=\"view?mailbox=");
		if (ms.usingimap)
			aucPrintLinkText(output, &mailbox[strlen(ms.imaphost) + 2]);
		else
			aucPrintLinkText(output, mailbox);
		fprintf(output, "\">%s</a></td>\n", &mailbox[strlen(ms.liststr)]);
	} else {
		fprintf(output, "<tr><td><a href=\"view?mailbox=");
		aucPrintLinkText(output, mailbox);
		fprintf(output, "\">%s</a></td>\n", mailbox);
	}

	result = mail_status(NIL, mailbox, SA_MESSAGES | SA_UNSEEN);
	if (result && ms.num && ms.new)
		fprintf(output, "<td>%d</td><td>%d</td>\n", ms.num, ms.new);
	else if (result && ms.num)
		fprintf(output, "<td>%d</td><td></td>\n", ms.num);
	else
		fprintf(output, "<td></td><td></td>\n");

	fprintf(output, "<td>\n");
	if (!(func_flags & QN_NODELETE)) {
		fprintf(output, "<a href=\"delfolder?mailbox=");
		aucPrintLinkText(output, mailbox);
		fprintf(output, "\">%s</a>\n", _("Delete"));
	}
	fprintf(output, "</td>\n");
	fprintf(output, "</tr>\n");
}

int
parseheaderforreply(char **replystr, char **to, char **cc, char **subject, char **inreplyto,
					unsigned long uid)
{
	ENVELOPE *mess_env;

	mess_env = mail_fetchstructure_full(inbox, uid, &ms.cur_body, FT_UID);
	if (mess_env->reply_to == NIL)
		return (1);
	*to = *cc = *subject = *replystr = *inreplyto = NULL;
	*replystr = strdup(_("On "));
	if (mess_env->date) {
		*replystr = (char *) realloc(*replystr, strlen(*replystr) + strlen(mess_env->date) + 1);
		strcat(*replystr, mess_env->date);
	}
	*replystr = (char *) realloc(*replystr, 2 + strlen(*replystr) + 1);
	strcat(*replystr, ", ");
	if (mess_env->from->personal) {
		*replystr =
			(char *) realloc(*replystr, strlen(*replystr) + strlen(mess_env->from->personal) + 1);
		strcat(*replystr, mess_env->from->personal);
	} else {
		*replystr =
			(char *) realloc(*replystr,
							 strlen(*replystr) + 1 + strlen(mess_env->from->host) +
							 strlen(mess_env->from->mailbox) + 1);
		strcat(*replystr, mess_env->from->mailbox);
		strcat(*replystr, "@");
		strcat(*replystr, mess_env->from->host);
	}
	*replystr = (char *) realloc(*replystr, strlen(_("wrote"))+ 4 + strlen(*replystr) + 1);
	strcat(*replystr, " ");
	strcat(*replystr, _("wrote"));
	strcat(*replystr, ":\n\n");

	addresscat(to, mess_env->reply_to, 0);
	
	/* make sure reply_to and to differ before including the to address in CC */
	if (mess_env->to && mess_env->reply_to && (
			strcmp(mess_env->to->mailbox, mess_env->reply_to->mailbox) ||
		 	strcmp(mess_env->to->host, mess_env->reply_to->host)))
		addresscat(cc, mess_env->to, QN_NOSELF);
		
	addresscat(cc, mess_env->cc, QN_NOSELF);
	if (mess_env->subject) {
		if ((strlen(mess_env->subject) >= 3) && !strncmp(mess_env->subject, "Re:", 3)) {
			*subject = malloc(1 + strlen(mess_env->subject));
			sprintf(*subject, "%s", mess_env->subject);
		} else {
			*subject = malloc(5 + strlen(mess_env->subject));
			sprintf(*subject, _("Re: %s"), mess_env->subject);
		}
	} else
		*subject = strdup(_("Re: your mail"));

	if (mess_env->message_id) {
		*inreplyto = malloc(1 + strlen(mess_env->message_id));
		sprintf(*inreplyto, "%s", mess_env->message_id);
	}

	return 0;
}

void
addresscat(char **str, ADDRESS * addr, int func_flags)
{
	char buf[256];

	if (!addr)
		return;
	if (addr->mailbox && addr->host) {
		if ((func_flags & QN_NOSELF) && !strcmp(addr->mailbox, user.uname)
			&& !strcmp(addr->host, user.domain)) {
			addresscat(str, addr->next, func_flags);
			return;
		}
		if (*str) {
			*str = realloc(*str, 2 + strlen(*str) + 1);
			strcat(*str, ", ");
		} else {
			*str = malloc(sizeof(char));
			strcpy(*str, "");
		}
		if (addr->personal) {
			snprintf(buf, sizeof(buf), "%s", addr->personal);
			if (badaddresschars(buf))
				snprintf(buf, sizeof(buf), "\"%s\"", addr->personal);
			*str = realloc(*str, 2 + strlen(buf) + strlen(*str) + 1);
			strcat(*str, buf);
			strcat(*str, " <");
		}
		if (addr->mailbox) {
			*str = realloc(*str, strlen(addr->mailbox) + strlen(*str) + 1);
			strcat(*str, addr->mailbox);
		}
		if (addr->host) {
			*str = realloc(*str, 1 + strlen(addr->host) + strlen(*str) + 1);
			strcat(*str, "@");
			strcat(*str, addr->host);
		}
		if (addr->personal) {
			*str = realloc(*str, 1 + strlen(*str) + 1);
			strcat(*str, ">");
		}
	}
	addresscat(str, addr->next, func_flags);
}

int
badaddresschars(char *buf)
{
	char *a;
	for (a = buf; *a != '\0'; a++)
		if (!(((*a >= 'a') && (*a <= 'z')) || ((*a >= 'A') && (*a <= 'Z'))
			  || (*a == '.') || (*a == ' ') || (*a == '-')))
			return 1;
	return 0;
}

void
printheadertext(FILE * output, char *text, unsigned long len)
{
	char value[1024], *p, *v, *a;
	int buf;

	a = text;
	while (len) {
		for (p = a, buf = 0; *a != ':' && len; len--, a++, buf++);
		if (!len)
			break;
		fprintf(output, "<tr valign=top><td width=80><b>%.*s</b></td>\n", buf + 1, p);
		for (v = ++a, len--, buf = 0;
			 len && !(*a != ' ' && (*(a - 1) == '\015' || *(a - 1) == '\012')); len--, a++, buf++);
		strncpy(value, v, buf);
		value[buf] = '\0';

		fprintf(output, "<td>");
		aucPrintText(value, output, 0);
		fprintf(output, "</td></tr>\n");
	}
}

void
addtostringlist(STRINGLIST ** list, char *str)
{
	*list = mail_newstringlist();
	(*list)->text.data = str;
	(*list)->text.size = strlen(str);
	(*list)->next = NULL;
}

unsigned short idcount = 0;

char *
generatemessid(void)
{
	char *id, hostname[128];
	time_t now;
	struct tm *now_x;

	gethostname(hostname, sizeof(hostname));
	now = time(NULL);
	now_x = localtime(&now);
	id = (char *) malloc(128 * sizeof(char));

	sprintf(id, "<aucMail.%02d%02d%02d%02d%02d%02d-%02d.%d@%.50s>",
			now_x->tm_year, now_x->tm_mon + 1,
			now_x->tm_mday, now_x->tm_hour, now_x->tm_min, now_x->tm_sec,
			idcount, getpid(), hostname);
	idcount++;
	return (id);
}

void
printnavcommands(FILE * outfile, unsigned long uid)
{
	char buffer[256];
	unsigned long msgno;

	msgno = mail_msgno(inbox, uid);
	fprintf(outfile, "<table cellspacing=0 cellpadding=2 border=0 width=100%%><tr>\n");
	fprintf(outfile, "<td width=60%%>\n");

	if (msgno > 1) {
		strcpy(buffer, "");
		mail_fetchsubject(buffer, inbox, msgno - 1, sizeof(buffer) - 1);
		fprintf(outfile, "<b>%s:</b> <a href=\"view?mailbox=%s&UID=%ld\">",
				_("Previous"), ms.mailbox, mail_uid(inbox, msgno - 1));
		if (strlen(buffer) && !((strlen(buffer) == 1) && (*buffer == ' ')))
			aucPrintText(buffer, outfile, TEXT_NOLINKS);
		else
			fprintf(outfile, "%s", _("no subject"));
		fprintf(outfile, "</a>\n");
	} else
		fprintf(outfile, "&nbsp;");
	fprintf(outfile, "<br>\n");

	if (msgno < inbox->nmsgs) {
		strcpy(buffer, "");
		mail_fetchsubject(buffer, inbox, msgno + 1, sizeof(buffer) - 1);
		fprintf(outfile, "<b>%s:</b> <a href=\"view?mailbox=%s&UID=%ld\">",
				_("Next"), ms.mailbox, mail_uid(inbox, msgno + 1));
		if (strlen(buffer) && !((strlen(buffer) == 1) && (*buffer == ' ')))
			aucPrintText(buffer, outfile, TEXT_NOLINKS);
		else
			fprintf(outfile, "%s", _("no subject"));
		fprintf(outfile, "</a>\n");
	} else
		fprintf(outfile, "&nbsp;");

	fprintf(outfile, "</td><td width=20%%>\n");

	fprintf(outfile, "<b><a href=\"move?source=");
	aucPrintLinkText(outfile, ms.mailbox);
	fprintf(outfile, "&dest=mail/%s&mess=%ld", user.trashfolder, uid);
	if (msgno < inbox->nmsgs)
		fprintf(outfile, "&next=%ld\">%s</a></b><br>\n", mail_uid(inbox, msgno + 1), _("Delete"));
	else
		fprintf(outfile, "\">%s</a></b><br>\n", _("Delete"));

	fprintf(outfile, "<b><a %s href=\"compose?type=forward&mailbox=",
			user.newwindow ? "target=\"_blank\"" : "");
	aucPrintLinkText(outfile, ms.mailbox);
	fprintf(outfile, "&UID=%ld\">%s</a></b><br>\n", uid, _("Forward"));

	fprintf(outfile, "</td><td width=20%%>\n");

	fprintf(outfile, "<b><a %s href=\"compose?type=reply&mailbox=",
			user.newwindow ? "target=\"_blank\"" : "");
	aucPrintLinkText(outfile, ms.mailbox);
	fprintf(outfile, "&UID=%ld\">%s</a></b><br>\n", uid, _("Reply"));

	fprintf(outfile, "<b><a %s href=\"compose?type=replyfull&mailbox=",
			user.newwindow ? "target=\"_blank\"" : "");
	aucPrintLinkText(outfile, ms.mailbox);
	fprintf(outfile, "&UID=%ld\">%s</a></b><br>\n", uid, _("Reply w/ text"));

	fprintf(outfile, "</td></tr></table>\n");
}

void
printmimeenc(FILE * outfile, unsigned short encoding)
{
	switch (encoding) {
	case ENC7BIT:
		fprintf(outfile, "7 bit");
		break;
	case ENC8BIT:
		fprintf(outfile, "8 bit");
		break;
	case ENCBINARY:
		fprintf(outfile, "binary");
		break;
	case ENCBASE64:
		fprintf(outfile, "base-64 encoded");
		break;
	case ENCQUOTEDPRINTABLE:
		fprintf(outfile, "quoted-printable");
		break;
	default:
		fprintf(outfile, "unknown encoding scheme");
		break;
	}
}

void
printheaders(FILE * outfile, ENVELOPE * mess_env, unsigned long msgno)
{
	char subject[512], *headers;
	unsigned long len;
	STRINGLIST *list, **curr;

	fprintf(outfile,
			"<table width=100%% border=0 cellpadding=2 cellspacing=0><tr bgcolor=#666666><td>\n");
	fprintf(outfile,
			"<table width=100%% cellpadding=2 cellspacing=0 border=0 bgcolor=#FFFFFF><tr valign=top><td>\n");
	fprintf(outfile, "<table width=100%% cellspacing=0 cellpadding=1 border=0>\n");
	if (mess_env->date)
		fprintf(outfile, "<tr valign=top><td width=120><b>%s:</b></td><td width=90%%>%s</td></tr>\n",
				_("Date"), mess_env->date);
	if (mess_env->from) {
		fprintf(outfile, "<tr valign=top><td width=120><b>%s:</b></td><td width=90%%>\n", _("From"));
		printaddresses(outfile, mess_env->from);
		fprintf(outfile, "</td></tr>\n");
	}
	if (mess_env->to) {
		fprintf(outfile, "<tr valign=top><td width=120><b>%s:</b></td><td width=90%%>\n", _("To"));
		printaddresses(outfile, mess_env->to);
		fprintf(outfile, "</td></tr>\n");
	}
	if (mess_env->from && mess_env->reply_to &&
		(strcmp(mess_env->from->mailbox, mess_env->reply_to->mailbox) ||
		 strcmp(mess_env->from->host, mess_env->reply_to->host))) {
		fprintf(outfile, "<tr valign=top><td width=120><b>%s:</b></td><td>\n", _("Reply-to"));
		printaddresses(outfile, mess_env->reply_to);
		fprintf(outfile, "</td></tr>\n");
	}
	if (mess_env->cc) {
		fprintf(outfile, "<tr valign=top><td width=120><b>%s:</b></td><td>\n", _("cc"));
		printaddresses(outfile, mess_env->cc);
		fprintf(outfile, "</td></tr>\n");
	}
	if (mess_env->bcc) {
		fprintf(outfile, "<tr valign=top><td width=120><b>%s:</b></td><td>\n", _("bcc"));
		printaddresses(outfile, mess_env->bcc);
		fprintf(outfile, "</td></tr>\n");
	}

	curr = &list;
	addtostringlist(curr, "Resent-Date");
	curr = &(*curr)->next;
	addtostringlist(curr, "Resent-From");
	curr = &(*curr)->next;
	addtostringlist(curr, "Resent-To");
	curr = &(*curr)->next;
	addtostringlist(curr, "Resent-cc");
	curr = &(*curr)->next;
	addtostringlist(curr, "Resent-Subject");
/*   curr = &(*curr)->next;
   addtostringlist(curr, "Received"); */

	if (msgno) {
		headers = mail_fetchheader_full(inbox, msgno, list, &len, FT_UID);
		printheadertext(outfile, headers, len);
	}
	if (mess_env->subject && strlen(mess_env->subject) > 0) {
		snprintf(subject, sizeof(subject), "%s", mess_env->subject);
		fprintf(outfile, "<tr valign=top><td width=120><b>%s:</b></td><td>", _("Subject"));
		aucPrintText(subject, outfile, 0);
		fprintf(outfile, "</td></tr>\n");
	}
	fprintf(outfile, "</table></td>\n");
	if (msgno && !ms.printable) {
		fprintf(outfile,
				"<td><table width=150 cellpadding=0 cellspacing=0 border=0><tr valign=top><td><p align=right>\n");

		fprintf(outfile, "<a href=\"view?mailbox=");
		aucPrintLinkText(outfile, ms.mailbox);
		fprintf(outfile, "&UID=%ld&show=printable\">%s</a><br>\n", msgno, _("View Printable"));

		fprintf(outfile, "<a href=\"view?mailbox=");
		aucPrintLinkText(outfile, ms.mailbox);
		fprintf(outfile, "&UID=%ld&show=raw\">%s</a>\n", msgno, _("View Raw"));

		fprintf(outfile, "</p></td></tr></table></td>\n");
	}
	fprintf(outfile, "</tr></table></td></tr></table>\n");
/*   curr = &list;
   mail_free_stringlist(curr); */
}

void
printmimepart(FILE * outfile, unsigned long msgno, BODY * body, char *secstr, char *stem)
{
	PART *part;
	int x;
	unsigned long len;
	char buffer[128];
	void *data;

	switch (body->type) {
	case TYPEMULTIPART:
		for (x = 1, part = body->nested.part; part != NULL; part = part->next, x++) {
			if (strlen(stem))
				snprintf(buffer, sizeof(buffer), "%s.%d", stem, x);
			else
				snprintf(buffer, sizeof(buffer), "%d", x);
			printmimepart(outfile, msgno, &part->body, buffer, buffer);
		}
		return;
	case TYPEMESSAGE:
		if (body->description && strlen(body->description) && strcmp(body->description, "(null)"))
			fprintf(outfile,
					"<p><table bgcolor=#CCCCCC width=100%% border=0><tr><td><b>%s</td></tr></table></p>\n",
					body->description);
		
		/* If this isn't an RFC 822 encapsulated message, don't try to decode it */
		if (strcasecmp(body->subtype, "RFC822"))
			return;
			
		printheaders(outfile, body->nested.msg->env, 0);
		if (strlen(stem))
			snprintf(buffer, sizeof(buffer), "%s.%d", stem, 1);
		else
			snprintf(buffer, sizeof(buffer), "%d", 1);
		printmimepart(outfile, msgno, body->nested.msg->body, buffer, stem);
		return;
	default:
		break;
	}

	if (strlen(secstr))
		data = (void *) mail_fetchbody_full(inbox, msgno, secstr, &len, FT_UID);
	else
		data = (void *) mail_fetchtext_full(inbox, msgno, &len, FT_UID);

	switch (body->encoding) {
	case ENC7BIT:
	case ENC8BIT:
		break;
	case ENCBINARY:
		fprintf(outfile, "<p>FIXME: BINARY encoding of data</p>\n");
		data = NULL;
		break;
	case ENCBASE64:
		data = rfc822_base64((char *) data, len, &len);
		break;
	case ENCQUOTEDPRINTABLE:
		data = rfc822_qprint((char *) data, len, &len);
		break;
	default:
		fprintf(outfile, "<p>FIXME: unknown encoding of data</p>\n");
		data = NULL;
		break;
	}
	if (data) {
		PARAMETER *par;
		int converted = 0;

		switch (body->type) {
		case TYPETEXT:
			for (par = body->parameter; par; par = par->next) {
				iconv_t iconvp;
				int inleft = len, outleft = 2*len+1;
				char * outbuf = (char *) malloc(sizeof(char) * outleft);
				if (!strcmp(par->attribute, "CHARSET")) {
					converted = 1;
					iconvp = iconv_open("UTF-8", par->value);
					iconv(iconvp, (char **) &data, &inleft, &outbuf, &outleft);
					data = outbuf - (2*len+1 - outleft);
					len = 2*len+1 - outleft;
					iconv_close(iconvp);
					break;
				}
			}
			
			if (strcmp(lcase(body->subtype), "html") == 0)
				fwrite((char *) data, sizeof(char), len, outfile);
			else if (strcmp(lcase(body->subtype), "plain") == 0)
				printplaintext(outfile, (char *) data, len, 0);
			else
				linkattachedfile(body, outfile, data, len);
				
			if (converted)
				free(data);
				
			break;
		case TYPEAPPLICATION:
		case TYPEAUDIO:
		case TYPEVIDEO:
			linkattachedfile(body, outfile, data, len);
			break;
		case TYPEIMAGE:
			if (!strcmp(lcase(body->subtype), "gif") || !strcmp(lcase(body->subtype), "jpeg"))
				showattachedimage(body, outfile, data, len);
			else
				linkattachedfile(body, outfile, data, len);
			break;
		default:
			linkattachedfile(body, outfile, data, len);
			break;
		}
	}
}

void
linkattachedfile(BODY * body, FILE * outfile, void *data, unsigned long len)
{
	char name[128], url[128], loc[128];

	fprintf(outfile, "<ul><table bgcolor=#CCCCCC border=0 cellspacing=0>\n");
	fprintf(outfile, "<tr bgcolor=#000000><td><font color=#FFFFFF><b>%s:<b>\n",
			_("Attached file of type"));
	aucMimePrintType(outfile, body->type);
	fprintf(outfile, "/%s</font></td></tr><tr><td>\n", lcase(body->subtype));
	if (body->description && strlen(body->description) && strcmp(body->description, "(null)"))
		fprintf(outfile, "<b>%s:</b> %s<br>\n", _("Description"), body->description);
	fprintf(outfile, "<b>%s:</b>\n", _("Encoding"));
	printmimeenc(outfile, body->encoding);
	fprintf(outfile, "<br>\n");
	if (writeattachedfile(body, data, len, name, sizeof(name), url, sizeof(url), loc, sizeof(loc))) {
		fprintf(outfile, "%s:\n", _("View or download (shift-click) attached file"));
		fprintf(outfile, "<b><a href=\"%s\">%s</a></b>\n", url, name);
	} else
		fprintf(outfile, "<b>%s</b>\n", _("Error: Attached file could not be written"));
	fprintf(outfile, "</td></tr></table></ul>\n");
}

void
showattachedimage(BODY * body, FILE * outfile, void *data, unsigned long len)
{
	char name[128], url[128], loc[128];

	if (writeattachedfile(body, data, len, name, sizeof(name), url, sizeof(url), loc, sizeof(loc))) {
		fprintf(outfile, "<p align=center>");
		if (body->description && strlen(body->description) && strcmp(body->description, "(null)"))
			fprintf(outfile, "<b>%s</b> -\n", body->description);
		fprintf(outfile, "<a href=\"%s\">%s<br>\n", url, name);
		fprintf(outfile, "<img src=\"%s\" alt=\"%s\"></a></p>\n", url, name);
	} else
		fprintf(outfile, "<ul>%s</ul>\n", _("Error writing attached image"));
}

void
choosefilename(BODY * body, char *str, size_t len)
{
	char prefix[16], ext[16], *defext;

	if (body->type == TYPEIMAGE)
		strcpy(prefix, "image");
	else
		strcpy(prefix, "file");
	strcpy(ext, "");
	defext = aucMimeGetDefaultExt(aucMimeInt2Type(body->type), body->subtype, user.etcroot);
	if (defext)
		snprintf(ext, sizeof(ext), ".%s", defext);
	snprintf(str, len, "%s%%02d%s", prefix, ext);
}

int
writeattachedfile(BODY * body, void *data, unsigned long len, char *name, size_t n_len, char *url,
				  size_t u_len, char *loc, size_t l_len)
{
	PARAMETER *para;
	char buffer[n_len];
	int desc, x = 1, fixed = 1;
	int flags;

	flags = O_TRUNC | O_CREAT | O_WRONLY;
	strncpy(name, "", n_len);

	for (para = body->parameter; para != NULL; para = para->next)
		if (strcmp(lcase(para->attribute), "name") == 0)
			snprintf(name, n_len, "%s", para->value);
	if (strlen(name) == 0) {
		choosefilename(body, buffer, n_len);
		fixed = 0;
	}
	if (!fixed) {
		snprintf(name, n_len, buffer, x++);
		flags = flags | O_EXCL;
	}
	snprintf(url, u_len, "%s/homedir/.auc/tmp/%s", user.pagesroot, aucLinkText(name));
	snprintf(loc, l_len, "%s/%s/tmp/%s", user.homedir, user.userdir, name);
	for (desc = open(loc, flags, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP);
		 !fixed && (x < 100) && (desc == -1) && (errno == EEXIST);) {
		snprintf(name, n_len, buffer, x++);
		snprintf(url, u_len, "%s/homedir/.auc/tmp/%s", user.pagesroot, aucLinkText(name));
		snprintf(loc, l_len, "%s/%s/tmp/%s", user.homedir, user.userdir, name);
		desc = open(loc, flags, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP);
	}
	if (desc != -1) {
		write(desc, data, len);
		close(desc);
/*      snprintf(buf, sizeof(buf), "%s.header", loc);
      if (content=fopen(buf, "w")) {
         fprintf(content, "Content-type: ");
         aucPrintMimeType(content, body->type);
         fprintf(content, "/%s\n", body->subtype);
         fclose(content);
      } */
		return (1);
	} else {
		printf("%d", errno);
		return (0);
	}
}

void
printplaintext(FILE * outfile, char *a, unsigned long len, unsigned long flags)
{
	char *str;

	str = malloc((len + 1) * sizeof(char));
	strncpy(str, a, len);
	str[len] = '\0';
	fprintf(outfile, "<p><code>\n");
	aucPrintText(str, outfile, flags);
	fprintf(outfile, "</code></p>\n");
	free(str);
}

void
printaddresses(FILE * outfile, ADDRESS * address)
{
	ADDRESS *current;
	char buffer[512];
	char *addrtext;
	int notfirst = 0;

	current = address;
	while (current) {
		if (current->personal || current->mailbox || current->host) {
			if (notfirst)
				fprintf(outfile, ",\n");
			notfirst = 1;
			addrtext = (char *) strdup("");
			if (current->personal) {
				addrtext = (char *) realloc(addrtext,
											4 + strlen(addrtext) + strlen(current->personal) + 1);
				if (badaddresschars(current->personal)) {
					snprintf(buffer, sizeof(buffer), "\"%s\" <", current->personal);
					strcat(addrtext, buffer);
				} else {
					snprintf(buffer, sizeof(buffer), "%s <", current->personal);
					strcat(addrtext, buffer);
				}
			}
			if (current->mailbox) {
				addrtext = (char *) realloc(addrtext,
											strlen(addrtext) + strlen(current->mailbox) + 1);
				strcat(addrtext, current->mailbox);
				if (current->host) {
					addrtext = (char *) realloc(addrtext,
												1 + strlen(addrtext) + strlen(current->host) + 1);
					snprintf(buffer, sizeof(buffer), "@%s", current->host);
					strcat(addrtext, buffer);
				}
			}
			if (current->personal) {
				addrtext = (char *) realloc(addrtext, 1 + strlen(addrtext) + 1);
				strcat(addrtext, ">");
			}
			if (current->mailbox && current->host) {
				fprintf(outfile, "<a %s href=\"compose?to=",
						user.newwindow ? "target=\"_blank\"" : "");
				aucPrintLinkText(outfile, addrtext);
				fprintf(outfile, "\">");
				aucPrintText(addrtext, outfile, TEXT_NOLINKS);
				fprintf(outfile, "</a>");
			} else
				fprintf(outfile, "%s", addrtext);
			free(addrtext);
		}
		current = current->next;
	}
	fprintf(outfile, "\n\n");
}

void
printaddresserrors(FILE * outfile, ADDRESS * address)
{
	ADDRESS *current;
	char buffer[512];

	current = address;
	while (current) {
		if (current->error) {
			if (current->personal)
				snprintf(buffer, sizeof(buffer), "%s <%s@%s>", current->personal, current->mailbox,
						 current->host);
			else
				snprintf(buffer, sizeof(buffer), "%s@%s", current->mailbox, current->host);
			fprintf(outfile, "<p>");
			aucPrintText(buffer, outfile, TEXT_NOLINKS);
			fprintf(outfile, ": %s</p>\n", current->error);
		}
		current = current->next;
	}
}


/* Interfaces to C-client */


void
mm_searched(MAILSTREAM * stream, unsigned long number)
{
	if (ms.firstnew == -1)
		ms.firstnew = number;
}


void
mm_exists(MAILSTREAM * stream, unsigned long number)
{
	if (number > ms.num)
		ms.new += (number - ms.num);
	ms.num = number;
}


void
mm_expunged(MAILSTREAM * stream, unsigned long number)
{
}


void
mm_flags(MAILSTREAM * stream, unsigned long number)
{
}


void
mm_notify(MAILSTREAM * stream, char *string, long errflg)
{
	mm_log(string, errflg);
}


void
mm_list(MAILSTREAM * stream, int delimiter, char *mailbox, long attributes)
{
	char matchstr[256], selected[128];

	snprintf(matchstr, sizeof(matchstr), "%s%s", ms.liststr, user.trashfolder);
	if (strcmp(mailbox, matchstr) == 0)
		ms.foundtrash = 1;
	snprintf(matchstr, sizeof(matchstr), "%s%s", ms.liststr, user.fcc);
	if (strcmp(mailbox, matchstr) == 0)
		ms.foundfcc = 1;
	snprintf(matchstr, sizeof(matchstr), "%s%s", ms.liststr, user.postponed);
	if (strcmp(mailbox, matchstr) == 0)
		ms.foundpostponed = 1;

	if (ms.statsonly && (attributes & ~LATT_NOSELECT)) {
		if (strncmp(mailbox, ms.liststr, strlen(ms.liststr)) == 0)
			printfolderstats(ms.outfile, stream, mailbox, 0);

	} else if (!ms.nolist && (attributes & ~LATT_NOSELECT)) {
		snprintf(matchstr, sizeof(matchstr), "%s", ms.selected_box);
		if (strcmp(mailbox, matchstr) == 0)
			strcpy(selected, _("selected"));
		else
			strcpy(selected, "");
		if (strncmp(mailbox, ms.liststr, strlen(ms.liststr)) == 0) {
			if (ms.outfile)
				fprintf(ms.outfile, "<option value=\"%s\" %s>%s</option>\n",
						ms.usingimap ? &mailbox[strlen(ms.imaphost) + 2] : mailbox,
						selected, &mailbox[strlen(ms.liststr)]);
		}
	}
}


void
mm_lsub(MAILSTREAM * stream, int delimiter, char *mailbox, long attributes)
{
	putchar(' ');
	if (delimiter)
		putchar(delimiter);
	else
		fputs("NIL", stdout);
	putchar(' ');
	fputs(mailbox, stdout);
	if (attributes & LATT_NOINFERIORS)
		fputs(", no inferiors", stdout);
	if (attributes & LATT_NOSELECT)
		fputs(", no select", stdout);
	if (attributes & LATT_MARKED)
		fputs(", marked", stdout);
	if (attributes & LATT_UNMARKED)
		fputs(", unmarked", stdout);
	putchar('\n');
}


void
mm_status(MAILSTREAM * stream, char *mailbox, MAILSTATUS * status)
{
	if (status->flags & SA_MESSAGES)
		ms.num = status->messages;
	if (status->flags & SA_UNSEEN)
		ms.new = status->unseen;
}

void
mm_log(char *string, long errflg)
{
	FILE *output = ms.outfile;

	if (strlen(string) <= 1)
		return;

	switch ((short) errflg) {
	case NIL:
		aucError("aucemail: %s", string);
		break;
	case PARSE:
		ms.alert = 1;
		if (ms.printerrors && output)
			fprintf(output, "<b>%s:</b> %s\n", _("Parse Error"), string);
		aucError("aucemail: %s", string);
		break;
	case WARN:
		if (strncmp(string, "error in rsh", strlen("error in rsh")))
			aucError("aucemail: %s", string);
		break;
	case ERROR:
		if (output)
			fprintf(output, "<b>%s:</b> %s\n", _("Error"), string);
		aucError("aucemail: %s", string);
		break;
	}
}


void
mm_dlog(char *string)
{
	aucError("aucemail: %s", string);
}


void
mm_login(NETMBX * mb, char *username, char *pwd, long trial)
{
	snprintf(username, 16, "%s", user.imapuname);
	snprintf(pwd, 16, "%s", user.imappasswd);
	if (trial || !strlen(user.imappasswd))
		ms.loginfailed = 1;
}


void
mm_critical(MAILSTREAM * stream)
{
}


void
mm_nocritical(MAILSTREAM * stream)
{
}


long
mm_diskerror(MAILSTREAM * stream, long errcode, long serious)
{
	kill(getpid(), SIGSTOP);
	return (NIL);
}


void
mm_fatal(char *string)
{
	aucError("aucemail: FATAL ERROR: %s", string);
}

void
aucComposeHref(char *email, FILE * output)
{
	fprintf(output, "<a %s href=\"compose?to=%s\">",
			user.newwindow ? "target=\"_blank\"" : "", email);
}
